// -*-c++-*-

/***************************************************************************
                                 rccptest.cc 
                     RoboCup Client Parser testing program
                             -------------------
    begin                : 01-JUL-2002
    copyright            : (C) 2002 by Tom Howard
    email                : tomhoward@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#include "rccparser.h"
#include <stack>
#include <iostream>
#include <string>

#ifndef HAVE_LIBRCSSCLANGPARSER
class CLangParser
    : public rcss::Parser
{
    bool
    doParse( std::istream& )
    { return true; }
};
#else
#include <rcssserver/clangparser.h>
#include <rcssserver/clangmsgbuilder.h>
#include <rcssserver/clangmsg.h>
#endif

class TestParser
    : public rcc::Parser
{
public:
    TestParser( rcss::Parser& clang_parser )
        : rcc::Parser( clang_parser )
    {}
    
    virtual
    ~TestParser()
    {}

    std::string
    getString()
    { return ( M_stack.empty() ? "" : M_stack.top() ); }
    
    void
    clear()
    { while( !M_stack.empty() ) M_stack.pop(); }
    
private:
    std::stack< std::string > M_stack;

    std::string
    buildFlagName( bool close )
    { return doBuildFlagName( close ); }

    std::string
    buildGoalName( bool close )
    { return doBuildGoalName( close ); }

    void
    buildPlayerName( std::ostream& strm, bool close, bool long_goalie )
    { doBuildPlayerName( strm, close, long_goalie ); }
     
    bool
    isFieldObj( const std::string& str ) const
    {
        if( str.size() < 3 || str[ 0 ] != '(' || str[ 1 ] != '(' )
        {
            return false;
        }
        switch( str[ 2 ] )
        {
            case 'p':
            case 'P':
            case 'b':
            case 'B':
            case 'g':
            case 'G':
            case 'l':
            case 'f':
            case 'F':
                return true;
            default:
                return false;
        }
    }
    
    bool
    isLocation( const std::string& str ) const
    {
        if( str.size() == 2 )
        {
#ifdef HAVE_SSTREAM
            std::istringstream strm( str );
#else
            std::istrstream strm( str );
#endif
            int offset = -1;
            strm >> offset;
            if( offset != -1 )
                return true;
            else
                return false;
        }
        
        if( str.size() != 1 )
            return false;


        switch( str[ 0 ] )
        {
            case 'l':
            case 'r':
            case 't':
            case 'b':
            case 'p':
            case 'c':
            case 'g':
            case '0':
            case '5':
                return true;
            default:
                return false;
        }
    }

    virtual
    void
    doMsgParsed()
    {
    }

    virtual
    void
    doParsedCLang( const std::string& msg )
    {
#ifndef HAVE_LIBRCSSCLANGPARSER
        M_stack.push( msg );
#endif
    }

    virtual
    void
    doBuildFullState( int time )
    {
        std::ostringstream strm;
        std::stack< std::string > players;
        while( !M_stack.empty() 
               && M_stack.top().size() > 3
               && M_stack.top()[ 2 ] == 'p' )
        {
            players.push( M_stack.top() );
            M_stack.pop();
        }
        std::string ball;
        if( !M_stack.empty() )
        { ball = M_stack.top(); M_stack.pop(); }
        std::string score;
        if( !M_stack.empty() )
        { score = M_stack.top(); M_stack.pop(); }
        std::string arm;
        if( !M_stack.empty() )
        { arm = M_stack.top(); M_stack.pop(); }
        std::string counts;
        if( !M_stack.empty() )
        { counts = M_stack.top(); M_stack.pop(); }
        std::string vmode;
        if( !M_stack.empty() )
        { vmode = M_stack.top(); M_stack.pop(); }
        std::string pmode;
        if( !M_stack.empty() )
        { pmode = M_stack.top(); M_stack.pop(); }
        
        strm << "(fullstate " << time
             << " (pmode " << pmode << ")"
             << " (vmode " << vmode << ")"
             << " " << counts
             << " " << arm
             << " " << score
             << " " << ball;
        while( !players.empty() )
        {
            strm << " " << players.top();
            players.pop();
        }
        strm << ")";
        M_stack.push( strm.str() );
    }
 
    virtual
    void
    doBuildCounts( int kick, int dash, int turn, int katch,
                   int move, int turn_neck, int change_view, int say )
    {
        std::ostringstream strm;
        strm << "(count"
             << " " << kick
             << " " << dash
             << " " << turn
             << " " << katch
             << " " << move 
             << " " << turn_neck
             << " " << change_view
             << " " << say
             << ")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildCatchBallLeftPlayMode()
    {
        M_stack.push( "goalie_catch_ball_l" );
    }

    virtual
    void
    doBuildCatchBallRightPlayMode()
    {
        M_stack.push( "goalie_catch_ball_r" );
    }

    virtual
    void
    doBuildBeforeKickOffPlayMode()
    {
        M_stack.push( "before_kick_off" );
    }

    virtual
    void
    doBuildTimeOverPlayMode()
    {
        M_stack.push( "time_over" );
    }

    virtual
    void
    doBuildPlayOnPlayMode()
    {
        M_stack.push( "play_on" );
    }

    virtual
    void
    doBuildKickOffLeftPlayMode()
    {
        M_stack.push( "kick_off_l" );
    }

    virtual
    void
    doBuildKickOffRightPlayMode()
    {
        M_stack.push( "kick_off_r" );
    }
 
    virtual
    void
    doBuildKickInLeftPlayMode()
    {
        M_stack.push( "kick_in_l" );
    }
 
    virtual
    void
    doBuildKickInRightPlayMode()
    {
        M_stack.push( "kick_in_r" );
    }

    virtual
    void
    doBuildFreeKickLeftPlayMode()
    {
        M_stack.push( "free_kick_l" );
    }

    virtual
    void
    doBuildFreeKickRightPlayMode()
    {
        M_stack.push( "free_kick_r" );
    }

    virtual
    void
    doBuildCornerKickLeftPlayMode()
    {
        M_stack.push( "corner_kick_l" );
    }

    virtual
    void
    doBuildCornerKickRightPlayMode()
    {
        M_stack.push( "corner_kick_r" );
    }
 
    virtual
    void
    doBuildGoalKickLeftPlayMode()
    {
        M_stack.push( "goal_kick_l" );
    }

    virtual
    void
    doBuildGoalKickRightPlayMode()
    {
        M_stack.push( "goal_kick_r" );
    }
 
    virtual
    void
    doBuildAfterGoalLeftPlayMode()
    {
        M_stack.push( "goal_l" );
    }
 
    virtual
    void
    doBuildAfterGoalRightPlayMode()
    {
        M_stack.push( "goal_r" );
    }
 
    virtual
    void
    doBuildDropBallPlayMode()
    {
        M_stack.push( "drop_ball" );
    }
 
    virtual
    void
    doBuildOffSideLeftPlayMode()
    {
        M_stack.push( "offside_l" );
    }

    virtual
    void
    doBuildOffsideRightPlayMode()
    {
        M_stack.push( "offside_r" );
    }
 
    virtual
    void
    doBuildPenaltyKickLeftPlayMode()
    {
        M_stack.push( "penalty_kick_l" );
    }
 
    virtual
    void
    doBuildPenaltyKickRightPlayMode()
    {
        M_stack.push( "penalty_kick_r" );
    }
 
    virtual
    void
    doBuildFirstHalfOverPlayMode()
    {
        M_stack.push( "first_half_over" );
    }
 
    virtual
    void
    doBuildPausePlayMode()
    {
        M_stack.push( "pause" );
    }
 
    virtual
    void
    doBuildHumanPlayMode()
    {
        M_stack.push( "human_judge" );
    }
 
    virtual
    void
    doBuildFoulLeftPlayMode()
    {
        M_stack.push( "foul_l" );
    }

    virtual
    void
    doBuildFoulRightPlayMode()
    {
        M_stack.push( "foul_r" );
    }

    virtual
    void
    doBuildFoulChargeLeftPlayMode()
    {
        M_stack.push( "foul_charge_l" );
    }

    virtual
    void
    doBuildFoulChargeRightPlayMode()
    {
        M_stack.push( "foul_charge_r" );
    }

    virtual
    void
    doBuildFoulPushLeftPlayMode()
    {
        M_stack.push( "foul_push_l" );
    }

    virtual
    void
    doBuildFoulPushRightPlayMode()
    {
        M_stack.push( "foul_push_r" );
    }

    virtual
    void
    doBuildFoulMultipleAttackerLeftPlayMode()
    {
        M_stack.push( "foul_multiple_attack_l" );
    }

    virtual
    void
    doBuildFoulMultipleAttackerRightPlayMode()
    {
        M_stack.push( "foul_multiple_attack_r" );
    }

    virtual
    void
    doBuildFoulBallOutLeftPlayMode()
    {
        M_stack.push( "foul_ballout_l" );
    }

    virtual
    void
    doBuildFoulBallOutRightPlayMode()
    {
        M_stack.push( "foul_ballout_r" );
    }

    virtual
    void
    doBuildBackPassLeftPlayMode()
    {
        M_stack.push( "back_pass_l" );
    }

    virtual
    void
    doBuildBackPassRightPlayMode()
    {
        M_stack.push( "back_pass_r" );
    }

    virtual
    void
    doBuildFreeKickFaultLeftPlayMode()
    {
        M_stack.push( "free_kick_fault_l" );
    }

    virtual
    void
    doBuildFreeKickFaultRightPlayMode()
    {
        M_stack.push( "free_kick_fault_r" );
    }

    virtual
    void
    doBuildScore( int our, int opp )
    {
        std::ostringstream strm;
        strm << "(score " << our << " " << opp << ")";
        M_stack.push( strm.str() );
                        
    }

    virtual
    void
    doBuildBall( double x, double y, 
                 double delta_x, double delta_y )
    {
        std::ostringstream strm;
        strm << "((b) " << x << " " << y
             << " " << delta_x << " " << delta_y
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildPlayer( int unum,
                   double x, double y, 
                   double delta_x, double delta_y,
                   double orientation, double head_orientation,
                   int stamina, double effort, double recovery )
    {
    }

    virtual
    void
    doBuildPlayer( int unum, int type,
                   double pos_x, double pos_y,
                   double vel_x, double vel_y,
                   double body_dir, double head_dir,
                   double stamina, double effort,
                   double recovery )
    {        
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty() )
        { side = M_stack.top(); M_stack.pop(); }
        strm << "((p " << side << " " << unum << " " << type << ")" 
             << " " << pos_x << " " << pos_y
             << " " << vel_x << " " << vel_y
             << " " << body_dir << " " << head_dir
             << " (stamina " << stamina 
             << " " << effort 
             << " " << recovery 
             << "))";
        M_stack.push( strm.str() );
                
    }
       
      virtual
      void
      doBuildGoalie( int unum,
                     double pos_x, double pos_y,
                     double vel_x, double vel_y,
                     double body_dir, double head_dir,
                     double stamina, double effort,
                     double recovery )
    {
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty() )
        { side = M_stack.top(); M_stack.pop(); }
        strm << "((p " << side << " " << unum << " g)" 
             << " " << pos_x << " " << pos_y
             << " " << vel_x << " " << vel_y
             << " " << body_dir << " " << head_dir
             << " (stamina " << stamina 
             << " " << effort 
             << " " << recovery 
             << "))";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildLeftSide()
    {
        M_stack.push( "l" );
    }

    virtual
    void
    doBuildRightSide()
    {
        M_stack.push( "r" );
    }

    virtual
    void
    doBuildOppSide()
    {
        M_stack.push( "opp" );
    }

    virtual
    void
    doBuildOurSide()
    {
        M_stack.push( "our" );
    }

    virtual
    void
    doBuildNeutralSide()
    {
        M_stack.push( "n" );
    }

    virtual
    void
    doBuildSenseBody( int time, 
                      double stamina,
                      double effort,
                      double speed_mag,
                      double speed_head,
                      double head_angle,
                      int kick_count,
                      int dash_count,
                      int turn_count,
                      int say_count,
                      int turn_neck_count,
                      int catch_count,
                      int move_count,
                      int chg_view_count )
    {
        std::ostringstream strm;
        std::string tackle;
        std::string focus;
        std::string arm;
        std::string view_mode;
        if( !M_stack.empty() )
        {
            tackle = M_stack.top(); M_stack.pop();
            if( tackle.substr( 0, 7 ) != "(tackle" )
                {
                    view_mode = tackle;
                    tackle = "";
                }
            else
            { 
                if( !M_stack.empty() )
                { focus = M_stack.top(); M_stack.pop(); }
                if( !M_stack.empty() )
                { arm = M_stack.top(); M_stack.pop(); }
                if( !M_stack.empty() )
                { view_mode = M_stack.top(); M_stack.pop(); }
            }
        }
        
        strm << "(sense_body"
             << " " << time 
             << " (view_mode " << view_mode << ")"
             << " (stamina " << stamina << " " << effort << ")"
             << " (speed " << speed_mag << " " << speed_head << ")"
             << " (head_angle " << head_angle << ")"
             << " (kick " << kick_count << ")"
             << " (dash " << dash_count << ")"
             << " (turn " << turn_count << ")"
             << " (say " << say_count << ")"
             << " (turn_neck " << turn_neck_count << ")"
             << " (catch " << catch_count << ")"
             << " (move " << move_count << ")"
             << " (change_view " << chg_view_count << ")";
        if( !arm.empty() )
            strm << " " << arm;
        if( !focus.empty() )
            strm << " " << focus;
        if( !tackle.empty() )
            strm << " " << tackle;
        strm << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildArmState( int movable_in,
                     int expires_in,
                     double target_x,
                     double target_y,
                     int count )
    {
        std::ostringstream strm;
        strm << "(arm " 
             << " (movable " << movable_in << ")"
             << " (expires " << expires_in << ")"
             << " (target " << target_x << " " << target_y << ")"
             << " (count " << count << ")"
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildFocusState( int count )
    {
        std::ostringstream strm;
        strm << "(focus"
             << " (target none)"
             << " (count " << count << ")"
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildFocusState( int unum, int count )
    {
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty() )
        { side = M_stack.top(); M_stack.pop(); }
        strm << "(focus "
             << " (target " << side << " " << unum << ")"
             << " (count " << count << ")"
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildTackleState( int expires_in, int count )
    {
        std::ostringstream strm;
        strm << "(tackle"
             << " (expires " << expires_in << ")"
             << " (count " << count << ")"
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildViewMode()
    {
        std::ostringstream strm;
        std::string qual, width;
        if( !M_stack.empty() )
        { width = M_stack.top(); M_stack.pop(); }
        if( !M_stack.empty() )
        { qual = M_stack.top(); M_stack.pop(); }
        strm << qual << " " << width;
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildViewQualityHigh()
    {
        M_stack.push( "high" );
    }

    virtual
    void
    doBuildViewQualityLow()
    {
        M_stack.push( "low" );
    }

    virtual
    void
    doBuildViewWidthNarrow()
    {
        M_stack.push( "narrow" );
    }

    virtual
    void
    doBuildViewWidthNormal()
    {
        M_stack.push( "normal" );
    }

    virtual
    void
    doBuildViewWidthWide()
    {
        M_stack.push( "wide" );
    }

    virtual
    void
    doBuildServerParam ( double gwidth, 
                         double inertia_moment, 
                         double psize, 
                         double pdecay, 
                         double prand,
                         double pweight, 
                         double pspeed_max, 
                         double paccel_max, 
                         double stamina_max, 
                         double stamina_inc,
                         double recover_init, 
                         double recover_dthr, 
                         double recover_min, 
                         double recover_dec, 
                         double effort_init,
                         double effort_dthr, 
                         double effort_min, 
                         double effort_dec, 
                         double effort_ithr, 
                         double effort_inc,
                         double kick_rand, 
                         int team_actuator_noise, 
                         double prand_factor_l, 
                         double prand_factor_r, 
                         double kick_rand_factor_l,
                         double kick_rand_factor_r, 
                         double bsize, 
                         double bdecay, 
                         double brand, 
                         double bweight,
                         double bspeed_max, 
                         double baccel_max, 
                         double dprate, 
                         double kprate, 
                         double kmargin,
                         double ctlradius, 
                         double ctlradius_width, 
                         double maxp, 
                         double minp, 
                         double maxm,
                         double minm, 
                         double maxnm, 
                         double minnm, 
                         double maxn, 
                         double minn,
                         double visangle, 
                         double visdist, 
                         double windir, 
                         double winforce, 
                         double winang,
                         double winrand, 
                         double kickable_area, 
                         double catch_area_l, 
                         double catch_area_w, 
                         double catch_prob,
                         int goalie_max_moves, 
                         double ckmargin, 
                         double offside_area, 
                         int win_no, 
                         int win_random,
                         int say_cnt_max, 
                         int SayCoachMsgSize, 
                         int clang_win_size, 
                         int clang_define_win, 
                         int clang_meta_win,
                         int clang_advice_win, 
                         int clang_info_win, 
                         int clang_mess_delay, 
                         int clang_mess_per_cycle, 
                         int half_time,
                         int sim_st, 
                         int send_st, 
                         int recv_st, 
                         int sb_step, 
                         int lcm_st,
                         int SayMsgSize, 
                         int hear_max, 
                         int hear_inc, 
                         int hear_decay, 
                         int cban_cycle,
                         int slow_down_factor, 
                         int useoffside, 
                         int kickoffoffside, 
                         double offside_kick_margin, 
                         double audio_dist,
                         double dist_qstep, 
                         double land_qstep, 
                         double dir_qstep, 
                         double dist_qstep_l, 
                         double dist_qstep_r,
                         double land_qstep_l, 
                         double land_qstep_r, 
                         double dir_qstep_l, 
                         double dir_qstep_r, 
                         int CoachMode,
                         int CwRMode, 
                         int old_hear, 
                         int sv_st, 
                         int start_goal_l, 
                         int start_goal_r,
                         int fullstate_l, 
                         int fullstate_r, 
                         int drop_time )
    {
        std::ostringstream strm;
        strm << "(server_param"
             << " " << gwidth 
             << " " << inertia_moment 
             << " " << psize 
             << " " << pdecay 
             << " " << prand
             << " " << pweight 
             << " " << pspeed_max 
             << " " << paccel_max 
             << " " << stamina_max 
             << " " << stamina_inc
             << " " << recover_init 
             << " " << recover_dthr 
             << " " << recover_min 
             << " " << recover_dec 
             << " " << effort_init
             << " " << effort_dthr 
             << " " << effort_min 
             << " " << effort_dec 
             << " " << effort_ithr 
             << " " << effort_inc
             << " " << kick_rand 
             << " " << team_actuator_noise 
             << " " << prand_factor_l 
             << " " << prand_factor_r 
             << " " << kick_rand_factor_l
             << " " << kick_rand_factor_r 
             << " " << bsize 
             << " " << bdecay 
             << " " << brand 
             << " " << bweight
             << " " << bspeed_max 
             << " " << baccel_max 
             << " " << dprate 
             << " " << kprate 
             << " " << kmargin
             << " " << ctlradius 
             << " " << ctlradius_width 
             << " " << maxp 
             << " " << minp 
             << " " << maxm
             << " " << minm 
             << " " << maxnm 
             << " " << minnm 
             << " " << maxn 
             << " " << minn
             << " " << visangle 
             << " " << visdist 
             << " " << windir 
             << " " << winforce 
             << " " << winang
             << " " << winrand 
             << " " << kickable_area 
             << " " << catch_area_l 
             << " " << catch_area_w 
             << " " << catch_prob
             << " " << goalie_max_moves 
             << " " << ckmargin 
             << " " << offside_area 
             << " " << win_no 
             << " " << win_random
             << " " << say_cnt_max 
             << " " << SayCoachMsgSize 
             << " " << clang_win_size 
             << " " << clang_define_win 
             << " " << clang_meta_win
             << " " << clang_advice_win 
             << " " << clang_info_win 
             << " " << clang_mess_delay 
             << " " << clang_mess_per_cycle 
             << " " << half_time
             << " " << sim_st 
             << " " << send_st 
             << " " << recv_st 
             << " " << sb_step 
             << " " << lcm_st
             << " " << SayMsgSize 
             << " " << hear_max 
             << " " << hear_inc 
             << " " << hear_decay 
             << " " << cban_cycle
             << " " << slow_down_factor 
             << " " << useoffside 
             << " " << kickoffoffside 
             << " " << offside_kick_margin 
             << " " << audio_dist
             << " " << dist_qstep 
             << " " << land_qstep 
             << " " << dir_qstep 
             << " " << dist_qstep_l 
             << " " << dist_qstep_r
             << " " << land_qstep_l 
             << " " << land_qstep_r 
             << " " << dir_qstep_l 
             << " " << dir_qstep_r 
             << " " << CoachMode
             << " " << CwRMode 
             << " " << old_hear 
             << " " << sv_st 
             << " " << start_goal_l 
             << " " << start_goal_r
             << " " << fullstate_l 
             << " " << fullstate_r 
             << " " << drop_time 
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildServerParam()
    {
        std::ostringstream strm;
        strm << "(server_param";
        std::stack< std::string > params;
        while( !M_stack.empty() )
        {
            params.push( M_stack.top() );
            M_stack.pop();
        }

        while( !params.empty() )
        {
            strm << " " << params.top();
            params.pop();
        }
        strm << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildParam( const std::string& name,
                  int value )
    {
        std::ostringstream strm;
        strm << "(" << name << " " << value << ")";
        M_stack.push( strm.str() );
            
     }

    virtual
    void
    doBuildParam( const std::string& name,
                  double value )
    {
        std::ostringstream strm;
        strm << "(" << name << " " << value << ")";
        M_stack.push( strm.str() );
            
    }
      
    virtual
    void
    doBuildParam( const std::string& name,
                  const std::string& value )
    {
        std::ostringstream strm;
        strm << "(" << name << " \"" << value << "\")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildPlayerParam( int player_types,
                        int subs_max,
                        int pt_max,
                        double player_speed_max_delta_min,
                        double player_speed_max_delta_max,
                        double stamina_inc_max_delta_factor,
                        double player_decay_delta_min,
                        double player_decay_delta_max,
                        double inertia_moment_delta_factor,
                        double dash_power_rate_delta_min,
                        double dash_power_rate_delta_max,
                        double player_size_delta_factor,
                        double kickable_margin_delta_min,
                        double kickable_margin_delta_max,
                        double kick_rand_delta_factor,
                        double extra_stamina_delta_min,
                        double extra_stamina_delta_max,
                        double effort_max_delta_factor,
                        double effort_min_delta_factor )
    {
        std::ostringstream strm;
        strm << "(player_param"
             << " " << player_types
             << " " << subs_max
             << " " << pt_max
             << " " << player_speed_max_delta_min
             << " " << player_speed_max_delta_max
             << " " << stamina_inc_max_delta_factor
             << " " << player_decay_delta_min
             << " " << player_decay_delta_max
             << " " << inertia_moment_delta_factor
             << " " << dash_power_rate_delta_min
             << " " << dash_power_rate_delta_max
             << " " << player_size_delta_factor
             << " " << kickable_margin_delta_min
             << " " << kickable_margin_delta_max
             << " " << kick_rand_delta_factor
             << " " << extra_stamina_delta_min
             << " " << extra_stamina_delta_max
             << " " << effort_max_delta_factor
             << " " << effort_min_delta_factor
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildPlayerParam()
    {
        std::ostringstream strm;
        strm << "(player_param";
        std::stack< std::string > params;
        while( !M_stack.empty() )
        {
            params.push( M_stack.top() );
            M_stack.pop();
        }

        while( !params.empty() )
        {
            strm << " " << params.top();
            params.pop();
        }
        strm << ")";
        M_stack.push( strm.str() );
            
     }

    virtual
    void
    doBuildPlayerType ( int id,
                        double player_speed_max,
                        double stamina_inc_max,
                        double player_decay,
                        double inertia_moment,
                        double dash_power_rate,
                        double player_size,
                        double kickable_margin,
                        double kick_rand,
                        double extra_stamina,
                        double effort_max,
                        double effort_min )
    {
        std::ostringstream strm;
        strm << "(player_type"
             << " " << id
             << " " << player_speed_max
             << " " << stamina_inc_max
             << " " << player_decay
             << " " << inertia_moment
             << " " << dash_power_rate
             << " " << player_size
             << " " << kickable_margin
             << " " << kick_rand
             << " " << extra_stamina
             << " " << effort_max
             << " " << effort_min
             << ")";
        M_stack.push( strm.str() );
            
    }

    virtual
    void
    doBuildPlayerType()
    {
        std::ostringstream strm;
        strm << "(player_type";
        std::stack< std::string > params;
        while( !M_stack.empty() )
        {
            params.push( M_stack.top() );
            M_stack.pop();
        }

        while( !params.empty() )
        {
            strm << " " << params.top();
            params.pop();
        }
        strm << ")";
        M_stack.push( strm.str() );
            
    }
 
    virtual
    void
    doBuildVisual( int time )
    {
        std::ostringstream strm;
        strm << "(see " << time;
        std::stack< std::string > objects;
        while( !M_stack.empty()
               && isFieldObj( M_stack.top() ) )
        {
            objects.push( M_stack.top() );
            M_stack.pop();
        }

        while( !objects.empty() )
        {
            strm << " " << objects.top();
            objects.pop();
        }
        strm << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildGlobalVisual( int time )
    {
        std::ostringstream strm;
        strm << "(see_global " << time;
        std::stack< std::string > objects;
        while( !M_stack.empty()
               && isFieldObj( M_stack.top() ) )
        {
            objects.push( M_stack.top() );
            M_stack.pop();
        }

        while( !objects.empty() )
        {
            strm << " " << objects.top();
            objects.pop();
        }
        strm << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildLine( double dist, double dir )
    {
        std::ostringstream strm;
        strm << "((l ";
        if( !M_stack.empty() )
        {
            strm << M_stack.top();
            M_stack.pop();
        }
        strm << ") " << dist << " " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildLine( double dir )
    {
        std::ostringstream strm;
        strm << "((l ";
        if( !M_stack.empty() )
        {
            strm << M_stack.top();
            M_stack.pop();
        }
        strm << ") " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildTopLocation()
    {
        M_stack.push( "t" );
    }
 
    virtual
    void
    doBuildBottomLocation()
    {
        M_stack.push( "b" );
    }
 
    virtual
    void
    doBuildLeftLocation()
    {
        M_stack.push( "l" );
    }
 
    virtual
    void
    doBuildRightLocation()
    {
        M_stack.push( "r" );
    }
 
    virtual
    void
    doBuildCenterLocation()
    {
        M_stack.push( "c" );
    }
 
    virtual
    void
    doBuildPenaltyLocation()
    {
        M_stack.push( "p" );
    }
 
    virtual
    void
    doBuildGoalLocation()
    {
        M_stack.push( "g" );
    }

    virtual
    void
    doBuildFlagOffset( int offset )
    {
        std::ostringstream strm;
        strm << offset;
        M_stack.push( strm.str() );
                
    }
    
    virtual
    std::string
    doBuildFlagName( bool close )
    {
        if( close )
            return "(F)";
        else
        {
            std::string rval = "(f";
            std::stack< std::string > locs;
            while( !M_stack.empty() && isLocation( M_stack.top() ) )
            {
                locs.push( M_stack.top() );
                M_stack.pop();
            }
            while( !locs.empty() )
            {
                rval += " " + locs.top();
                locs.pop();
            }
            rval += ")";
            return rval;
        }
    }

    virtual
    std::string
    doBuildGoalName( bool close )
    {
        if( close )
            return "(G)";
        else
        {
            std::string rval = "(g";
            std::stack< std::string > locs;
            while( !M_stack.empty() && isLocation( M_stack.top() ) )
            {
                locs.push( M_stack.top() );
                M_stack.pop();
            }
            while( !locs.empty() )
            {
                rval += " " + locs.top();
                locs.pop();
            }
            rval += ")";
            return rval;
        }
    }

    virtual
    void
    doBuildFlag( bool close,
                 double dist, double dir,
                 double dist_chg, double dir_chg )
    {
        std::ostringstream strm;
        strm << "(" << buildFlagName( close )
             << " " << dist << " " << dir
             << " " << dist_chg << " " << dir_chg
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildFlag( bool close, double dist, double dir )
    {
        std::ostringstream strm;
        strm << "(" << buildFlagName( close )
             << " " << dist << " " << dir
             << ")";
        M_stack.push( strm.str() );
                    
    }

    virtual
    void
    doBuildFlag( bool close, double dir )
    {
        std::ostringstream strm;
        strm << "(" << buildFlagName( close )
             << " " << dir
             << ")";
        M_stack.push( strm.str() );
                    
    }

    virtual
    void
    doBuildGoal( bool close,
                 double dist, double dir,
                 double dist_chg, double dir_chg )
    {
        std::ostringstream strm;
        strm << "(" << buildGoalName( close )
             << " " << dist << " " << dir
             << " " << dist_chg << " " << dir_chg
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildGoal( bool close, double dist, double dir )
    {
        std::ostringstream strm;
        strm << "(" << buildGoalName( close )
             << " " << dist << " " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildGoal( bool close, double dir )
    {
        std::ostringstream strm;
        strm << "(" << buildGoalName( close )
             << " " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildGlobalGoal( double x, double y )
    {
        std::ostringstream strm;
        strm << "(" << buildGoalName( false )
             << " " << x << " " << y
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildBall( bool close,
                 double dist, double dir,
                 double dist_chg, double dir_chg )
    {
        std::ostringstream strm;
        if( close )
            strm << "((B) ";
        else
            strm << "((b) ";
        strm << dist << " " << dir
             << " " << dist_chg << " " << dir_chg
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildBall( bool close, double dist, double dir )
    {
        std::ostringstream strm;
        if( close )
            strm << "((B) ";
        else
            strm << "((b) ";
        strm << dist << " " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildBall( bool close, double dir )
    {
        std::ostringstream strm;
        if( close )
            strm << "((B) ";
        else
            strm << "((b) ";
        strm << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildGlobalBall( double x, double y,
                       double delta_x, double delta_y )
    {
        std::ostringstream strm;
        strm << "((b) " << x << " " << y
             << " " << delta_x << " " << delta_y
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildPlayerName( std::ostream& strm, bool close, bool )
    {
        if( close )
            strm << "(P)";
        else
        {
            std::string team_name;
            int unum = -1;
            bool goalie = false;
            if( !M_stack.empty() && M_stack.top() == "g" )
            {
                goalie = true;
                M_stack.pop();
            }
            if( !M_stack.empty() )
            {
                std::istringstream in( M_stack.top() );
                in >> unum;
                if( unum != -1 )
                    M_stack.pop();
            }
            if( !M_stack.empty() 
                && !M_stack.top().empty()
                && M_stack.top()[ 0 ] == '"' )
            {
                team_name = M_stack.top();
                M_stack.pop();
            }
            strm << "(p";
            if( !team_name.empty() )
                strm << " " << team_name;
            if( unum != -1 )
                strm << " " << unum;
            if( goalie )
            {
                strm << " goalie";
            }
            strm << ")";
        }
    }

    virtual
    void
    doBuildPlayer( bool close,
                   double dist, double dir,
                   double dist_chg, double dir_chg,
                   double orientation, double head_orientation )
    {
        std::ostringstream strm;
        strm << "(";
        buildPlayerName( strm, close, true );
        strm << " " << dist << " " << dir
             << " " << dist_chg << " " << dir_chg
             << " " << orientation << " " << head_orientation
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildPlayer( bool close,
                   double dist, double dir,
                   double dist_chg, double dir_chg )
    {
        std::ostringstream strm;
        strm << "(";
        buildPlayerName( strm, close, false );
        strm << " " << dist << " " << dir
             << " " << dist_chg << " " << dir_chg
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildPlayer( bool close, double dist, double dir )
    {
        std::ostringstream strm;
        strm << "(";
        buildPlayerName( strm, close, false );
        strm << " " << dist << " " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildPlayer( bool close, double dir )
    {
        std::ostringstream strm;
        strm << "(";
        buildPlayerName( strm, close, false );
        strm << " " << dir
             << ")";
        M_stack.push( strm.str() );
                
    }
    
    virtual
    void
    doBuildGlobalPlayer( double x, double y,
                         double delta_x, double delta_y,
                         double orientation, double head_orientation )
    {
        std::ostringstream strm;
        strm << "(";
        buildPlayerName( strm, false, true );
        strm << " " << x << " " << y
             << " " << delta_x << " " << delta_y
             << " " << orientation << " " << head_orientation
             << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildTeamName( const std::string& name )
    {
        std::ostringstream strm;
        strm << "\"" << name << "\"";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildUNum( int unum )
    {
        std::ostringstream strm;
        strm << unum;
        M_stack.push( strm.str() );
                        
    }

    virtual
    void
    doBuildGoalie()
    {
        M_stack.push( "g" );
    }

    virtual
    void
    doBuildScore( int time, int our, int opp )
    {
        std::ostringstream strm;
        strm << "(score " << time
             << " " << our << " " << opp << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildInit( int unum )
    {
        std::ostringstream strm;
        std::string play_mode;
        std::string side;
        if( !M_stack.empty() )
        {
            play_mode = M_stack.top();
            M_stack.pop();
        }
        if( !M_stack.empty() )
        {
            side = M_stack.top();
            M_stack.pop();
        }
        strm << "(init " << side << " " << unum << " "
             << play_mode << ")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildInit()
    {
        std::ostringstream strm;
        std::string play_mode;
        std::string side;
        if( !M_stack.empty() )
        {
            play_mode = M_stack.top();
            M_stack.pop();
        }
        if( !M_stack.empty() )
        {
            side = M_stack.top();
            M_stack.pop();
        }
        strm << "(reconnect " << side << " "
             << play_mode << ")";
        M_stack.push( strm.str() );
        
    }
    
    virtual
    void
    doBuildCoachInit()
    {
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty() )
        {
            side = M_stack.top();
            M_stack.pop();
        }
        strm << "(init " << side << " "
             << "ok)";
        M_stack.push( strm.str() );
        
    }
    
    virtual
    void
    doBuildCoachAudio( int time, const std::string& msg )
    {
        std::ostringstream strm;
        std::string name;
        if( !M_stack.empty() )
        {
            if( M_stack.top() == "l" )
                name = "online_coach_left";
            else if( M_stack.top() == "r" )
                name = "online_coach_right";
            else
                name = "coach";
        }
        strm << "(hear " << time << " " << name
             << " \"" << msg << "\")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildRefAudio( int time )
    {
        std::ostringstream strm;
        std::string play_mode;
        if( !M_stack.empty() )
        {
            play_mode = M_stack.top();
            M_stack.pop();
        }
        strm << "(hear " << time << " referee "
             << play_mode << ")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildGoalRefAudio( int time, int score )
    {
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty() )
        {
            side = M_stack.top();
            M_stack.pop();
        }
        strm << "(hear " << time << " referee goal_"
             << side << "_" << score << ")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildUnknownRefAudio( int time, const std::string& message )
    {
        std::ostringstream strm;
        strm << "(hear " << time << " referee "
             << message << ")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildPlayerAudio( int time, double dir, const std::string& msg )
    {
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty()
            && M_stack.top().size() > 0
            && M_stack.top()[ 0 ] == 'o')
        { side = M_stack.top(); M_stack.pop(); }
        strm << "(hear " << time << " " << dir;
        if( !side.empty() )
            strm << " " << side;
        strm << " \"" << msg << "\")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildPlayerAudio( int time, const std::string& msg )
    {
        std::ostringstream strm;    
        strm << "(hear " << time << " ";
        buildPlayerName( strm, false, false );
        strm << " \"" << msg << "\")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildPlayerAudio( int time, double dir, int unum,
                        const std::string& msg )
    {
        std::ostringstream strm;
        std::string side;
        if( !M_stack.empty()
            && M_stack.top().size() > 0
            && M_stack.top()[ 0 ] == 'o')
        { side = M_stack.top(); M_stack.pop(); }
        strm << "(hear " << time << " " << dir
             << " " << side << " " << unum
             << " \"" << msg << "\")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildSelfAudio( int time, const std::string& msg )
    {
        std::ostringstream strm;
        strm << "(hear " << time << " self \"" << msg << "\")";
        M_stack.push( strm.str() );
        
    }

    virtual
    void
    doBuildSubstitution( int unum, int type )
    {
        std::ostringstream strm;
        strm << "(change_player_type " << unum << " " << type << ")";
        M_stack.push( strm.str() );
                
    }
    
    virtual
    void
    doBuildSubstitution( int unum )
    {
        std::ostringstream strm;
        strm << "(change_player_type " << unum << ")";
        M_stack.push( strm.str() );
                
    }

    virtual
    void
    doBuildCantReconnect()
    {
    }

    virtual
    void
    doBuildInitError()
    {
    }

    virtual
    void
    doBuildNoMoreTeamOrPlayerOrGoalieError()
    {
    }
    
    virtual
    void
    doBuildNoSuchTeamOrAlreadyHaveCoachError()
    {
    }
 
    virtual
    void
    doBuildTooManyMovesError()
    {
    }

    virtual
    void
    doBuildUnknownCommandError()
    {
    }

    virtual
    void
    doBuildIllegalCommandError()
    {
        M_stack.push( "(error illegal_command_form)" );
    }

    virtual
    void
    doBuildSayMessageTooLongError()
    {
    }
 
    virtual
    void
    doBuildIllegalModeError()
    {
    }
 
    virtual
    void
    doBuildIllegalObjectFormError()
    {
    }
 
    virtual
    void
    doBuildSaidTooManyFreeformMessagesError()
    {
    }

    virtual
    void
    doBuildCannotSayFreeformWhilePlayonError()
    {
    }

    virtual
    void
    doBuildSaidTooManyMetaMessagesError()
    {
    }

    virtual
    void
    doBuildSaidTooManyAdviceMessagesError()
    {
    }

    virtual
    void
    doBuildSaidTooManyInfoMessagesError()
    {
    }

    virtual
    void
    doBuildSaidTooManyDefineMessagesError()
    {
    }

    virtual
    void
    doBuildCouldNotParseSayError()
    {
    }

    virtual
    void
    doBuildSaidTooManyMessagesError()
    {
    }

    virtual
    void
    doBuildUnknownError( const std::string& error )
    {
    }

    virtual
    void
    doBuildNoTeamFoundWarning()
    {
    }

    virtual
    void
    doBuildNoSuchPlayerWarning()
    {
    }

    virtual
    void
    doBuildCannotSubWhilePlayOnWarning()
    {
    }

    virtual
    void
    doBuildNoSubsLeftWarning()
    {
    }

    virtual
    void
    doBuildMaxOfThatPlayerTypeOnFieldWarning()
    {
    }

    virtual
    void
    doBuildCannotChangeGoalieWarning()
    {
    }

    virtual
    void
    doBuildUnknownWarning( const std::string& warning )
    {
    }
    
    
    virtual
    void
    doBuildClangVerOK( int min, int max )
    {
    }
          
    virtual
    void
    doBuildEyeOK( bool on )
    {
        std::ostringstream strm;
        strm << "(ok eye " << ( on ? "on" : "off" ) << ")";
        M_stack.push( strm.str() );
                
    }
        
    virtual
    void
    doBuildChangePlayerTypeOK( int unum, int type )
    {
        std::ostringstream strm;
        strm << "(ok change_player_type " 
             << unum << " " << type 
             << ")";
        M_stack.push( strm.str() );
               
    }
  
    virtual
    void
    doBuildUnknownOK( const std::string& msg )
    {
    }
};

std::pair< int, int >
diff( const std::string& a, const std::string& b )
{
    std::pair< int, int > rval( -1, -1 );

    if( a == b )
        return rval;

    if( !a.empty() && !b.empty() 
        && a[ 0 ] == b [ 0 ] )
    {
        rval = diff( a.substr( 1 ), b.substr( 1 ) );
        if( rval.first != -1 )
            ++( rval.first );
        if( rval.second != -1 )
            ++( rval.second );
        return rval;
    }
    else if( a.size() > 1 && !b.empty()
             && a[ 0 ] == '-' && a[ 1 ] == '0' && b[ 0 ] == '0' )
    {
        rval = diff( a.substr( 1 ), b );
        if( rval.first != -1 )
            ++( rval.first );
        return rval;       
    }
    else if( b.size() > 1 && !a.empty()
             && b[ 0 ] == '-' && b[ 1 ] == '0' && a[ 0 ] == '0' )
    {
        rval = diff( a, b.substr( 1 ) );
        if( rval.second != -1 )
            ++( rval.second );
        return rval;       
    }
    else if( !a.empty() && a[ 0 ] == ' ' )
    {
        rval = diff( a.substr( 1 ), b );
        if( rval.first != -1 )
            ++( rval.first );
        return rval;       
    }
    else if( !b.empty() && b[ 0 ] == ' ' )
    {
        rval = diff( a, b.substr( 1 ) );
        if( rval.second != -1 )
            ++( rval.second );
        return rval;       
    }
    else if( a.size() > 2 && a[ 0 ] == '.' && a[ 1 ] == '0' 
             && ( a[ 2 ] == ' ' || a[ 2 ] == ')' ) )
    {
        rval = diff( a.substr( 2 ), b );
        if( rval.first != -1 )
            rval.first += 2;
        return rval;       
    }
    else if( b.size() > 2 && b[ 0 ] == '.' && b[ 1 ] == '0' 
             && ( b[ 2 ] == ' ' || b[ 2 ] == ')' ) )
    {
        rval = diff( a, b.substr( 2 ) );
        if( rval.second != -1 )
            rval.second += 2;
        return rval;       
    }
    else
    {
        rval.first = rval.second = 0;
        return rval;
    }
}


int main()
{
    int rval = 0;
#ifndef HAVE_LIBRCSSCLANGPARSER
    CLangParser clang_parser;
#else
    rcss::clang::MsgBuilder builder;
    rcss::clang::Parser clang_parser( builder );
#endif
    TestParser parser( clang_parser );
    
    while( std::cin.good() )
    {
        std::string line;
		std::getline( std::cin, line );
        if( line.empty() )
            continue;

        std::istringstream strm( line );
        if( parser.parse( strm ) )
        {
            std::string parsed_data;
#ifdef HAVE_LIBRCSSCLANGPARSER
            if( !parser.getString().empty() || builder.getMsg() == NULL )
#endif
                parsed_data = parser.getString();
#ifdef HAVE_LIBRCSSCLANGPARSER
            else
            {
                std::ostringstream clang_strm;
                rcss::clang::Msg& msg = *( builder.getMsg() );
                clang_strm << "(hear " << msg.getTimeSend();
                if( msg.getSide() == 1 )
                    clang_strm << " online_coach_left ";
                else
                    clang_strm << " online_coach_right ";
                clang_strm << msg.getTimeRecv() << " " << msg << ")";
                parsed_data = clang_strm.str();
            }
#endif
            std::pair< int, int > diff_pos = diff( line, parsed_data );
            if( diff_pos.first != -1 && diff_pos.second != -1 )
            {
                std::cerr << "Input and parsed output differ\n";
                std::cerr << "Input >" << line << "<\n";
                std::cerr << "Output>" << parsed_data << "<\n";
                std::cerr << "Input differs at >" 
                     << line.substr( diff_pos.first ) << "<\n";
                std::cerr << "From output at >"
                     << parsed_data.substr( diff_pos.second ) << "<\n";
                rval = -1;
            }
            parser.clear();
        }
        else
        {
            std::cerr << "Error: Could not parse >" << line << "<\n";
            rval = -1;
        }
    }

	return rval;
}
