// -*-c++-*-

/***************************************************************************
                               rccparser.h  
                         Parser Abstract Base Class
                             -------------------
    begin                : 24-MAY-2002
    copyright            : (C) 2002 by Tom Howard
    email                : tomhoward@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU GPL as published by the Free Software   *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/

/*! \file
 *  This is the header for rcc::Parser class. It is the main header file
 *  for the RCCParser library.  It defines the
 *  rcc::Parser abstract base class, which your client needs to subclass
 *  to prevode custom behaviour during parsing.
 */

/*! \mainpage RoboCup Client Parser (RCCParser)
 *
 *  \section intro Introduction
 *
 *  RCCParser is a parser library for RoboCup clients for the
 *  <a href="http://sserver.sf.net">RoboCup Soccer Simulator</a>.
 *  It can be used for field players and coaches using protocol versions
 *  7 to 9.
 *
 *  The parser is Flex and Bison based so it should be fast compared to a
 *  hand written parser.
 *
 *  This program is free software; you can redistribute it and/or modify  
 *  it under the terms of the GNU GPL as published by the Free Software   
 *  Foundation; either version 2 of the License, or (at your option) any  
 *  later version.                                                       
 *  
 *  This product includes software (Flex) developed by the University of
 *  California, Berkeley and its contributors.
 *
 *  \section install Installation
 *
 *  \subsection basic Basic Installation

   The `configure' shell script attempts to guess correct values for
various system-dependent variables used during compilation.  It uses
those values to create a `Makefile' in each directory of the package.
It may also create one or more `.h' files containing system-dependent
definitions.  Finally, it creates a shell script `config.status' that
you can run in the future to recreate the current configuration, and a
file `config.log' containing compiler output (useful mainly for
debugging `configure').

   It can also use an optional file (typically called `config.cache'
and enabled with `--cache-file=config.cache' or simply `-C') that saves
the results of its tests to speed up reconfiguring (Caching is
disabled by default to prevent problems with accidental use of stale
cache files).  If you are using the cache, and at
some point `config.cache' contains results you don't want to keep, you
may remove or edit it.

   If you need to do unusual things to compile the package, please try
to figure out how `configure' could check whether to do them, and mail
diffs or instructions to rccparser-bugs@lists.sf.net so they can
be considered for the next release.

   The file `configure.in' is used to create
`configure' by a program called `autoconf'.  You only need
`configure.in' if you want to change it or regenerate `configure' using
a newer version of `autoconf'.

The simplest way to compile this package is:
<ol>
   <li>'cd' to the directory containing the package's source code and type
     `./configure' to configure the package for your system.  If you're
     using `csh' on an old version of System V, you might need to type
     `sh ./configure' instead to prevent `csh' from trying to execute
     `configure' itself.

     Running `configure' takes awhile.  While running, it prints some
     messages telling which features it is checking for.
   </li>
   <li>
     Type `make' to compile the package.
   </li>
   <li>
     Optionally, type `make check' to run any self-tests that come with
     the package.
   </li>
   <li>
     Type `make install' to install the programs and any data files and
     documentation.
   </li>
   <li>
     You can remove the program binaries and object files from the
     source code directory by typing `make clean'.  To also remove the
     files that `configure' created (so you can compile the package for
     a different kind of computer), type `make distclean'.  There is
     also a `make maintainer-clean' target, but that is intended mainly
     for the package's developers.  If you use it, you may have to get
     all sorts of other programs in order to regenerate files that came
     with the distribution.
   </li>
</ol>

 * \subsection compilers Compilers and Options

   Some systems require unusual options for compilation or linking that
the `configure' script does not know about.  Run `./configure --help'
for details on some of the pertinent environment variables.

   You can give `configure' initial values for configuration parameters
by setting variables in the command line or in the environment.  Here
is an example:

     ./configure CC=c89 CFLAGS=-O2 LIBS=-lposix

   See \ref variables "Defining Variables" for more information.

* \subsection multiplearch Compiling For Multiple Architectures

   You can compile the package for more than one kind of computer at the
same time, by placing the object files for each architecture in their
own directory.  To do this, you must use a version of `make' that
supports the `VPATH' variable, such as GNU `make'.  `cd' to the
directory where you want the object files and executables to go and run
the `configure' script.  `configure' automatically checks for the
source code in the directory that `configure' is in and in `..'.

   If you have to use a `make' that does not support the `VPATH'
variable, you have to compile the package for one architecture at a
time in the source code directory.  After you have installed the
package for one architecture, use `make distclean' before reconfiguring
for another architecture.

* \subsection names Installation Names

   By default, `make install' will install the package's files in
`/usr/local/bin', `/usr/local/man', etc.  You can specify an
installation prefix other than `/usr/local' by giving `configure' the
option `--prefix=PATH'.

   You can specify separate installation prefixes for
architecture-specific files and architecture-independent files.  If you
give `configure' the option `--exec-prefix=PATH', the package will use
PATH as the prefix for installing programs and libraries.
Documentation and other data files will still use the regular prefix.

   In addition, if you use an unusual directory layout you can give
options like `--bindir=PATH' to specify different values for particular
kinds of files.  Run `configure --help' for a list of the directories
you can set and what kinds of files go in them.

* \subsection options Optional Features

   `configure' also allows you to select and deselect optional features when
building RCCParser, the most relevant of these being `--enable-rccptest=ARG',
`--enable-shared=ARG' and `--enable-static=ARG'.

   `--enable-rccptest=ARG' can be used to enable of disable the building of 
the rccptest executable, which is used to test the parser by reading data to
be parsed from stdin.  By default, building rccptest is enabled.
  

   `--enable-shared=ARG' and `--enable-static=ARG' respectively enable and
disable the building of shared and static versions of the library.  Both are
enabled by default (that is, when the system supports shared libraries,
otherwise only static libraries are built).  There are advantages to using
shared libraries (that I wont go into here), but they can be a pain when
trying to debug a program. For that reason, it is recommended that you
disable shared libraries when debugging.

 * \subsection system Specifying the System Type

   There may be some features `configure' cannot figure out
automatically, but needs to determine by the type of machine the package
will run on.  Usually, assuming the package is built to be run on the
_same_ architectures, `configure' can figure that out, but if it prints
a message saying it cannot guess the machine type, give it the
`--build=TYPE' option.  TYPE can either be a short name for the system
type, such as `sun4', or a canonical name which has the form:

     CPU-COMPANY-SYSTEM

where SYSTEM can have one of these forms:

     OS KERNEL-OS

   See the file `config.sub' for the possible values of each field.  If
`config.sub' isn't included in this package, then this package doesn't
need to know the machine type.

   If you want to _use_ a cross compiler, that generates code for a
platform different from the build platform, you should specify the
"host" platform (i.e., that on which the generated programs will
eventually be run) with `--host=TYPE'.

 * \subsection variables Defining Variables

   Variables not defined in a site shell script can be set in the
environment passed to `configure'.  However, it is best to set
them in the `configure' command line, using `VAR=value'.  For example:

     ./configure CC=/usr/local2/bin/gcc

will cause the specified gcc to be used as the C compiler.

 * \subsection invocation `configure' Invocation

   `configure' recognizes the following options to control how it
operates.

`--help'
`-h'
     Print a summary of the options to `configure', and exit.

`--version'
`-V'
     Print the version of Autoconf used to generate the `configure'
     script, and exit.

`--cache-file=FILE'
     Enable the cache: use and save the results of the tests in FILE,
     traditionally `config.cache'.  FILE defaults to `/dev/null' to
     disable caching.

`--config-cache'
`-C'
     Alias for `--cache-file=config.cache'.

`--quiet'
`--silent'
`-q'
     Do not print messages saying which checks are being made.  To
     suppress all normal output, redirect it to `/dev/null' (any error
     messages will still be shown).

`--srcdir=DIR'
     Look for the package's source code in directory DIR.  Usually
     `configure' can determine that directory automatically.

`configure' also accepts some other, not widely useful, options.  Run
`configure --help' for more details.

* \section uninstall Uninstalling 

RCCParser can be easily removed by entering the src
directory and running `make uninstall'.  This will remove all the
files that where installed, but not any directories that were created
during the installation process.

* \section using Using The Parser

RCCParser takes care of parsing data from the simulator as a std::ostream,
what it doesn't take care of is what to do with the parsed data.  That is
the job of your client.  The default behaviour of the parser is to do nothing
with the parsed data.

To overide the default behaviour, you need to subclass rcc::Parser and overide
the virtual functions that the parser calls.

As a RoboCup message is parsed, the virtual functions will
be called to from the bottom up.  For
instance for the message

(init l 2 before_kick_off)

The follwing functions will be called:
 -# rcc::Parser::doBuildLeftSide()
 -# rcc::Parser::doBuildBeforeKickOffPlayMode()
 -# rcc::Parser::doBuildInit()

You will also need to pass a rcss::Parser object (note the different
namespace) to rcc::Parser when it is
initialized.  This so the rcssclangparser library that comes with rcsserver
can be used to parse CLang messages (thus making sure the CLang parsing is
always upto date).

Assuming the data to be parsed is in a char array, the following code would
cause parsing to occur.

\code
rcss::clang::MsgBuilder clang_builder;
rcss::clang::Parser clang_parser( clang_builder );
YourParser parser( clang_parser );

[...snip...]


istrstream strm( buffer ); // buffer is the char array with the data
if( !parser.parse( strm ) )
{ 
   // handle error 
}
\endcode

Instances of rcc::Parser derived classes can be used in a
multi-threaded environment, but you will either need prevent
simultaneous access by using a mutex before the call the parse, or by
using a different instance in each thread.  Normally clients only need one
instance of the parser, which is only accessed from a single thread.

* \section contact Contacts

The RCCParser <a href="http://rccparser.sf.net">home page</a> should be your
main point of contact for all matters concerning RCCParser, like getting
the latest versions, lodging support requests, bug reports and contacting
other users of RCCParser library.

At the home page you will also find access the the RCCParser mailing lists, 
rccparser-main@lists.sf.net and rccpaser-bugs@lists.sf.net.

* \subsection rccpaserbugs rccpaser-bugs@lists.sf.net

This secondary list is for reporting bugs or potential bugs in the
parser.  This could be anything like parse errors, building and
installation problems, or even defects in the documentation. I've
tried to make building, installing and using the parser as simple as
possible, so if you run into any problems whatsoever, you should post
it to this list.

* \subsection rccpasermain rccpaser-main@lists.sf.net

The rccparser-main@lists.sf.net list is the main list for the
RCCParser library, where you will mostly find announcements for the
latest releases.  If you have any questions or suggestions feel free
to post them here.

I hope your team development goes well, that this tool makes it
that little bit easier and I hope to see you at the next RoboCup World
Cup.

 */

#ifndef RCCPARSER_H
#define RCCPARSER_H

#include <rcssbase/parser.h>
//#include "rcclexer.h"

/** The Bison parsing function.
 * 
 * This is the parsing function produced by Bison.
 * You should \b not call this function directly.  Instead you should
 * use rcc::Parser::parse(), which makes sure that the correct arguments
 * are passed to RCC_parse().
 */

int RCC_parse( void * );


class RCCLexer;

/** The vendor namespace used by RCCParser */
namespace rcc
{
    class Lexer;

    class Holder
    {
    private:
        enum Type { S_CSTR, S_INT, S_DOUB, S_BOOL, S_STR };
        
        union 
        {
            const char* m_cstr;
            int m_int;
            double m_double;
            bool m_bool;
        };
        std::string m_str;
        Type m_type;

    public:
        Holder&
        operator=( const char* x )
        {
            m_cstr = x;
            m_type = S_CSTR;
            return *this;
        }

        Holder&
        operator=( int x )
        {
            m_int = x;
            m_type = S_INT;
            return *this;
        }
        
        Holder&
        operator=( double x )
        {
            m_double = x;
            m_type = S_DOUB;
            return *this;
        }
        
        Holder&
        operator=( bool x )
        {
            m_bool = x;
            m_type = S_BOOL;
            return *this;
        }
        
        Holder&
        operator=( const std::string& x )
        {
            m_str = x;
            m_type = S_STR;
            return *this;
        }
        
        const char*
        getCStr() const
        { return m_cstr; }

        int
        getInt() const
        { return m_int; }

        double
        getDouble() const
        { return m_double; }

        bool
        getBool() const
        { return m_bool; }

        std::string&
        getStr()
        { return m_str; }

        const std::string&
        getStr() const
        { return m_str; }
    };
    
    /** The parsing base class.
     *   
     *  This is the main class provided by the RCCParser library.  Subclass
     *  it and overide the virtual functions to provide custom behaviour with
     *  the parsed data during parsing.  The default behaviour is to do nothing with
     *  the parsed data.
     */
    class Parser
        : public rcss::Parser
    {
    public:
        typedef rcc::Lexer Lexer;
        
    private:
        class Param
        {
        private:
            Lexer* M_lexer;
            Parser& M_parser;

        public:
            Param( Parser& parser );

            ~Param();

            Param( const Param& );

            Param&
            operator=( const Param& );
            
            Parser&
            getParser()
            { return M_parser; }
            
            Lexer&
            getLexer()
            { return *M_lexer; }
            
            bool
            parseCLang( const char* message )
            { return M_parser.parseCLang( message ); }

            void
            msgParsed()
            { M_parser.doMsgParsed(); }

            void
            buildFullState( int time )
            { M_parser.doBuildFullState( time ); }

            void
            buildCounts( int kick,
                         int dash,
                         int turn,
                         int katch,
                         int move,
                         int turn_neck,
                         int change_view,
                         int say )
            {
                M_parser.doBuildCounts( kick, dash, turn, katch, move,
                                        turn_neck, change_view, say );
            }
        
            void
            buildCatchBallLeftPlayMode()
            { M_parser.doBuildCatchBallLeftPlayMode(); }

            void
            buildCatchBallRightPlayMode()
            { M_parser.doBuildCatchBallRightPlayMode(); }

            void
            buildBeforeKickOffPlayMode()
            { M_parser.doBuildBeforeKickOffPlayMode(); }

            void
            buildTimeOverPlayMode()
            { M_parser.doBuildTimeOverPlayMode(); }

            void
            buildPlayOnPlayMode()
            { M_parser.doBuildPlayOnPlayMode(); }

            void
            buildKickOffLeftPlayMode()
            { M_parser.doBuildKickOffLeftPlayMode(); }

            void
            buildKickOffRightPlayMode()
            { M_parser.doBuildKickOffRightPlayMode(); }

            void
            buildKickInLeftPlayMode()
            { M_parser.doBuildKickInLeftPlayMode(); }

            void
            buildKickInRightPlayMode()
            { M_parser.doBuildKickInRightPlayMode(); }

            void
            buildFreeKickLeftPlayMode()
            { M_parser.doBuildFreeKickLeftPlayMode(); }

            void
            buildFreeKickRightPlayMode()
            { M_parser.doBuildFreeKickRightPlayMode(); }

            void
            buildCornerKickLeftPlayMode()
            { M_parser.doBuildCornerKickLeftPlayMode(); }

            void
            buildCornerKickRightPlayMode()
            { M_parser.doBuildCornerKickRightPlayMode(); }

            void
            buildGoalKickLeftPlayMode()
            { M_parser.doBuildGoalKickLeftPlayMode(); }

            void
            buildGoalKickRightPlayMode()
            { M_parser.doBuildGoalKickRightPlayMode(); }

            void
            buildAfterGoalLeftPlayMode()
            { M_parser.doBuildAfterGoalLeftPlayMode(); }

            void
            buildAfterGoalRightPlayMode()
            { M_parser.doBuildAfterGoalRightPlayMode(); }

            void
            buildDropBallPlayMode()
            { M_parser.doBuildDropBallPlayMode(); }

            void
            buildOffSideLeftPlayMode()
            { M_parser.doBuildOffSideLeftPlayMode(); }

            void
            buildOffsideRightPlayMode()
            { M_parser.doBuildOffsideRightPlayMode(); }

            void
            buildPenaltyKickLeftPlayMode()
            { M_parser.doBuildPenaltyKickLeftPlayMode(); }

            void
            buildPenaltyKickRightPlayMode()
            { M_parser.doBuildPenaltyKickRightPlayMode(); }

            void
            buildFirstHalfOverPlayMode()
            { M_parser.doBuildFirstHalfOverPlayMode(); }

            void
            buildPausePlayMode()
            { M_parser.doBuildPausePlayMode(); }

            void
            buildHumanPlayMode()
            { M_parser.doBuildHumanPlayMode(); }

            void
            buildFoulLeftPlayMode()
            { M_parser.doBuildFoulLeftPlayMode(); }

            void
            buildFoulRightPlayMode()
            { M_parser.doBuildFoulRightPlayMode(); }

            void
            buildFoulChargeLeftPlayMode()
            { M_parser.doBuildFoulChargeLeftPlayMode(); }

            void
            buildFoulChargeRightPlayMode()
            { M_parser.doBuildFoulChargeRightPlayMode(); }

            void
            buildFoulPushLeftPlayMode()
            { M_parser.doBuildFoulPushLeftPlayMode(); }

            void
            buildFoulPushRightPlayMode()
            { M_parser.doBuildFoulPushRightPlayMode(); }

            void
            buildFoulMultipleAttackerLeftPlayMode()
            { M_parser.doBuildFoulMultipleAttackerLeftPlayMode(); }

            void
            buildFoulMultipleAttackerRightPlayMode()
            { M_parser.doBuildFoulMultipleAttackerRightPlayMode(); }

            void
            buildFoulBallOutLeftPlayMode()
            { M_parser.doBuildFoulBallOutLeftPlayMode(); }

            void
            buildFoulBallOutRightPlayMode()
            { M_parser.doBuildFoulBallOutRightPlayMode(); }

            void
            buildBackPassLeftPlayMode()
            { M_parser.doBuildBackPassLeftPlayMode(); }

            void
            buildBackPassRightPlayMode()
            { M_parser.doBuildBackPassRightPlayMode(); }

            void
            buildFreeKickFaultLeftPlayMode()
            { M_parser.doBuildFreeKickFaultLeftPlayMode(); }

            void
            buildFreeKickFaultRightPlayMode() 
            { M_parser.doBuildFreeKickFaultRightPlayMode(); }

            void
            buildCatchFaultLeftPlayMode()
            { M_parser.doBuildCatchFaultLeftPlayMode(); }

            void
            buildCatchFaultRightPlayMode()
            { M_parser.doBuildCatchFaultRightPlayMode(); }

            void
            buildIndirectFreeKickLeftPlayMode()
            { M_parser.doBuildIndirectFreeKickLeftPlayMode(); }

            void
            buildIndirectFreeKickRightPlayMode()
            { M_parser.doBuildIndirectFreeKickRightPlayMode(); }

            void
            buildPenaltySetupLeftPlaymode() 
            { M_parser.doBuildPenaltySetupLeftPlaymode(); }
 
            void
            buildPenaltySetupRightPlaymode() 
            { M_parser.doBuildPenaltySetupRightPlaymode(); }
 
            void
            buildPenaltyReadyLeftPlaymode() 
            { M_parser.doBuildPenaltyReadyLeftPlaymode(); }
 
            void
            buildPenaltyReadyRightPlaymode() 
            { M_parser.doBuildPenaltyReadyRightPlaymode(); }
 
            void
            buildPenaltyTakenLeftPlaymode() 
            { M_parser.doBuildPenaltyTakenLeftPlaymode(); }
 
            void
            buildPenaltyTakenRightPlaymode() 
            { M_parser.doBuildPenaltyTakenRightPlaymode(); }
 
            void
            buildPenaltyMissLeftPlaymode() 
            { M_parser.doBuildPenaltyMissLeftPlaymode(); }
 
            void
            buildPenaltyMissRightPlaymode() 
            { M_parser.doBuildPenaltyMissRightPlaymode(); }
 
            void
            buildPenaltyScoreLeftPlaymode() 
            { M_parser.doBuildPenaltyScoreLeftPlaymode(); }
 
            void
            buildPenaltyScoreRightPlaymode() 
            { M_parser.doBuildPenaltyScoreRightPlaymode(); }
 
            void
            buildPenaltyOnFieldLeftPlaymode() 
            { M_parser.doBuildPenaltyOnFieldLeftPlaymode(); }
 
            void
            buildPenaltyOnFieldRightPlaymode() 
            { M_parser.doBuildPenaltyOnFieldRightPlaymode(); }
 
            void
            buildPenaltyFoulLeftPlaymode() 
            { M_parser.doBuildPenaltyFoulLeftPlaymode(); }
 
            void
            buildPenaltyFoulRightPlaymode() 
            { M_parser.doBuildPenaltyFoulRightPlaymode(); }
 
            void
            buildPenaltyWinnerLeftPlaymode() 
            { M_parser.doBuildPenaltyWinnerLeftPlaymode(); }
 
            void
            buildPenaltyWinnerRightPlaymode() 
            { M_parser.doBuildPenaltyWinnerRightPlaymode(); }
 
            void
            buildPenaltyDrawPlaymode() 
            { M_parser.doBuildPenaltyDrawPlaymode(); }
 

            void
            buildScore( int our, int opp )
            { M_parser.doBuildScore( our, opp ); }

            void
            buildBall( double x, double y,
                       double delta_x, double delta_y )
            { M_parser.doBuildBall( x, y, delta_x, delta_y ); }

            void
            buildPlayer( int unum,
                         double x, double y, 
                         double delta_x, double delta_y,
                         double orientation, double head_orientation,
                         int stamina, double effort, double recovery )
            { 
                M_parser.doBuildPlayer( unum,
                                        x, y, 
                                        delta_x, delta_y,
                                        orientation, head_orientation,
                                        stamina, effort, recovery ); 
            }

            void
            buildPlayer( int unum, int type,
                         double pos_x, double pos_y,
                         double vel_x, double vel_y,
                         double body_dir, double head_dir,
                         double stamina, double effort,
                         double recovery )
            {
                M_parser.doBuildPlayer( unum, type, pos_x, pos_y, vel_x, vel_y,
                                        body_dir, head_dir, stamina, effort, recovery );
            }

            void
            buildPlayer( int unum, int type,
                         double pos_x, double pos_y,
                         double vel_x, double vel_y,
                         double body_dir, double head_dir,
                         double arm_mag, double arm_dir,
                         double stamina, double effort,
                         double recovery )
            {
                M_parser.doBuildPlayer( unum, type, pos_x, pos_y, vel_x, vel_y,
                                        body_dir, head_dir, arm_mag, arm_dir,
                                        stamina, effort, recovery );
            }

            void
            buildGoalie( int unum,
                         double pos_x, double pos_y,
                         double vel_x, double vel_y,
                         double body_dir, double head_dir,
                         double stamina, double effort,
                         double recovery )
            {
                M_parser.doBuildGoalie( unum, pos_x, pos_y, vel_x, vel_y,
                                        body_dir, head_dir, stamina, effort, recovery );
            }        

            void
            buildGoalie( int unum,
                         double pos_x, double pos_y,
                         double vel_x, double vel_y,
                         double body_dir, double head_dir,
                         double arm_mag, double arm_dir,
                         double stamina, double effort,
                         double recovery )
            {
                M_parser.doBuildGoalie( unum, pos_x, pos_y, vel_x, vel_y,
                                        body_dir, head_dir, arm_mag, arm_dir,
                                        stamina, effort, recovery );
            }        

            void
            buildLeftSide()
            { M_parser.doBuildLeftSide(); }

            void
            buildRightSide()
            { M_parser.doBuildRightSide(); }

            void
            buildOurSide()
            { M_parser.doBuildOurSide(); }

            void
            buildOppSide()
            { M_parser.doBuildOppSide(); }


            void
            buildNeutralSide()
            { M_parser.doBuildNeutralSide(); }


            void
            buildSenseBody( int time,
                            double stamina,
                            double effort,
                            double speed_mag,
                            double speed_head,
                            double head_angle,
                            int kick_count,
                            int dash_count,
                            int turn_count,
                            int say_count,
                            int turn_neck_count,
                            int catch_count,
                            int move_count,
                            int chg_view_count )
            {
                M_parser.doBuildSenseBody( time, stamina, effort, speed_mag, speed_head,
                                           head_angle, kick_count, dash_count, turn_count,
                                           say_count, turn_neck_count, catch_count, move_count, 
                                           chg_view_count );
            }

            void
            buildArmState( int movable_in,
                           int expires_in,
                           double target_x,
                           double target_y,
                           int count )
            { M_parser.doBuildArmState( movable_in, expires_in, target_x, target_y, count ); }

            void
            buildFocusState( int count )
            { M_parser.doBuildFocusState( count ); }

            void
            buildFocusState( int unum, int count )
            { M_parser.doBuildFocusState( unum, count ); }


            void
            buildTackleState( int expires_in, int count )
            { M_parser.doBuildTackleState( expires_in, count ); }

            void
            buildViewMode()
            { M_parser.doBuildViewMode(); }

            void
            buildViewQualityHigh()
            { M_parser.doBuildViewQualityHigh(); }

            void
            buildViewQualityLow()
            { M_parser.doBuildViewQualityLow(); }

            void
            buildViewWidthNarrow()
            { M_parser.doBuildViewWidthNarrow(); }

            void
            buildViewWidthNormal()
            { M_parser.doBuildViewWidthNormal(); }

            void
            buildViewWidthWide()
            { M_parser.doBuildViewWidthWide(); }

            void
            buildServerParam ( double gwidth, 
                               double inertia_moment, 
                               double psize, 
                               double pdecay, 
                               double prand,
                               double pweight, 
                               double pspeed_max, 
                               double paccel_max, 
                               double stamina_max, 
                               double stamina_inc,
                               double recover_init, 
                               double recover_dthr, 
                               double recover_min, 
                               double recover_dec, 
                               double effort_init,
                               double effort_dthr, 
                               double effort_min, 
                               double effort_dec, 
                               double effort_ithr, 
                               double effort_inc,
                               double kick_rand, 
                               int team_actuator_noise, 
                               double prand_factor_l, 
                               double prand_factor_r, 
                               double kick_rand_factor_l,
                               double kick_rand_factor_r, 
                               double bsize, 
                               double bdecay, 
                               double brand, 
                               double bweight,
                               double bspeed_max, 
                               double baccel_max, 
                               double dprate, 
                               double kprate, 
                               double kmargin,
                               double ctlradius, 
                               double ctlradius_width, 
                               double maxp, 
                               double minp, 
                               double maxm,
                               double minm, 
                               double maxnm, 
                               double minnm, 
                               double maxn, 
                               double minn,
                               double visangle, 
                               double visdist, 
                               double windir, 
                               double winforce, 
                               double winang,
                               double winrand, 
                               double kickable_area, 
                               double catch_area_l, 
                               double catch_area_w, 
                               double catch_prob,
                               int goalie_max_moves, 
                               double ckmargin, 
                               double offside_area, 
                               int win_no, 
                               int win_random,
                               int say_cnt_max, 
                               int SayCoachMsgSize, 
                               int clang_win_size, 
                               int clang_define_win, 
                               int clang_meta_win,
                               int clang_advice_win, 
                               int clang_info_win, 
                               int clang_mess_delay, 
                               int clang_mess_per_cycle, 
                               int half_time,
                               int sim_st, 
                               int send_st, 
                               int recv_st, 
                               int sb_step, 
                               int lcm_st,
                               int SayMsgSize, 
                               int hear_max, 
                               int hear_inc, 
                               int hear_decay, 
                               int cban_cycle,
                               int slow_down_factor, 
                               int useoffside, 
                               int kickoffoffside, 
                               double offside_kick_margin, 
                               double audio_dist,
                               double dist_qstep, 
                               double land_qstep, 
                               double dir_qstep, 
                               double dist_qstep_l, 
                               double dist_qstep_r,
                               double land_qstep_l, 
                               double land_qstep_r, 
                               double dir_qstep_l, 
                               double dir_qstep_r, 
                               int CoachMode,
                               int CwRMode, 
                               int old_hear, 
                               int sv_st, 
                               int start_goal_l, 
                               int start_goal_r,
                               int fullstate_l, 
                               int fullstate_r, 
                               int drop_time )
            {
                M_parser.doBuildServerParam( gwidth, 
                                             inertia_moment, 
                                             psize, 
                                             pdecay, 
                                             prand,
                                             pweight, 
                                             pspeed_max, 
                                             paccel_max, 
                                             stamina_max, 
                                             stamina_inc,
                                             recover_init, 
                                             recover_dthr, 
                                             recover_min, 
                                             recover_dec, 
                                             effort_init,
                                             effort_dthr, 
                                             effort_min, 
                                             effort_dec, 
                                             effort_ithr, 
                                             effort_inc,
                                             kick_rand, 
                                             team_actuator_noise, 
                                             prand_factor_l, 
                                             prand_factor_r, 
                                             kick_rand_factor_l,
                                             kick_rand_factor_r, 
                                             bsize, 
                                             bdecay, 
                                             brand, 
                                             bweight,
                                             bspeed_max, 
                                             baccel_max, 
                                             dprate, 
                                             kprate, 
                                             kmargin,
                                             ctlradius, 
                                             ctlradius_width, 
                                             maxp, 
                                             minp, 
                                             maxm,
                                             minm, 
                                             maxnm, 
                                             minnm, 
                                             maxn, 
                                             minn,
                                             visangle, 
                                             visdist, 
                                             windir, 
                                             winforce, 
                                             winang,
                                             winrand, 
                                             kickable_area, 
                                             catch_area_l, 
                                             catch_area_w, 
                                             catch_prob,
                                             goalie_max_moves, 
                                             ckmargin, 
                                             offside_area, 
                                             win_no, 
                                             win_random,
                                             say_cnt_max, 
                                             SayCoachMsgSize, 
                                             clang_win_size, 
                                             clang_define_win, 
                                             clang_meta_win,
                                             clang_advice_win, 
                                             clang_info_win, 
                                             clang_mess_delay, 
                                             clang_mess_per_cycle, 
                                             half_time,
                                             sim_st, 
                                             send_st, 
                                             recv_st, 
                                             sb_step, 
                                             lcm_st,
                                             SayMsgSize, 
                                             hear_max, 
                                             hear_inc, 
                                             hear_decay, 
                                             cban_cycle,
                                             slow_down_factor, 
                                             useoffside, 
                                             kickoffoffside, 
                                             offside_kick_margin, 
                                             audio_dist,
                                             dist_qstep, 
                                             land_qstep, 
                                             dir_qstep, 
                                             dist_qstep_l, 
                                             dist_qstep_r,
                                             land_qstep_l, 
                                             land_qstep_r, 
                                             dir_qstep_l, 
                                             dir_qstep_r, 
                                             CoachMode,
                                             CwRMode, 
                                             old_hear, 
                                             sv_st, 
                                             start_goal_l, 
                                             start_goal_r,
                                             fullstate_l, 
                                             fullstate_r, 
                                             drop_time );
            }

            void
            buildServerParam()
            { M_parser.doBuildServerParam(); }

            void
            buildParam( const std::string& name,
                        int value )
            { M_parser.doBuildParam( name, value ); }

            void
            buildParam( const std::string& name,
                        double value )
            { M_parser.doBuildParam( name, value ); }

            void
            buildParam( const std::string& name,
                        const std::string& value )
            { M_parser.doBuildParam( name, value ); }

            void
            buildPlayerParam( int player_types,
                              int subs_max,
                              int pt_max,
                              double player_speed_max_delta_min,
                              double player_speed_max_delta_max,
                              double stamina_inc_max_delta_factor,
                              double player_decay_delta_min,
                              double player_decay_delta_max,
                              double inertia_moment_delta_factor,
                              double dash_power_rate_delta_min,
                              double dash_power_rate_delta_max,
                              double player_size_delta_factor,
                              double kickable_margin_delta_min,
                              double kickable_margin_delta_max,
                              double kick_rand_delta_factor,
                              double extra_stamina_delta_min,
                              double extra_stamina_delta_max,
                              double effort_max_delta_factor,
                              double effort_min_delta_factor )
            {
                M_parser.doBuildPlayerParam( player_types,
                                             subs_max,
                                             pt_max,
                                             player_speed_max_delta_min,
                                             player_speed_max_delta_max,
                                             stamina_inc_max_delta_factor,
                                             player_decay_delta_min,
                                             player_decay_delta_max,
                                             inertia_moment_delta_factor,
                                             dash_power_rate_delta_min,
                                             dash_power_rate_delta_max,
                                             player_size_delta_factor,
                                             kickable_margin_delta_min,
                                             kickable_margin_delta_max,
                                             kick_rand_delta_factor,
                                             extra_stamina_delta_min,
                                             extra_stamina_delta_max,
                                             effort_max_delta_factor,
                                             effort_min_delta_factor );
            }

            void
            buildPlayerParam()
            { M_parser.doBuildPlayerParam(); }

            void
            buildPlayerType ( int id,
                              double player_speed_max,
                              double stamina_inc_max,
                              double player_decay,
                              double inertia_moment,
                              double dash_power_rate,
                              double player_size,
                              double kickable_margin,
                              double kick_rand,
                              double extra_stamina,
                              double effort_max,
                              double effort_min )
            {
                M_parser.doBuildPlayerType ( id,
                                             player_speed_max,
                                             stamina_inc_max,
                                             player_decay,
                                             inertia_moment,
                                             dash_power_rate,
                                             player_size,
                                             kickable_margin,
                                             kick_rand,
                                             extra_stamina,
                                             effort_max,
                                             effort_min );
            }

            void
            buildPlayerType()
            { M_parser.doBuildPlayerType(); }
 
            void
            buildVisual( int time )
            { M_parser.doBuildVisual( time ); }

            void
            buildGlobalVisual( int time )
            { M_parser.doBuildGlobalVisual( time ); }

            void
            buildLine( double dist, double dir )
            { M_parser.doBuildLine( dist, dir ); }
 
            void
            buildLine( double dir )
            { M_parser.doBuildLine( dir ); }
 
            void
            buildTopLocation()
            { M_parser.doBuildTopLocation(); }
 
            void
            buildBottomLocation() 
            { M_parser.doBuildBottomLocation(); }
 
            void
            buildLeftLocation() 
            { M_parser.doBuildLeftLocation(); }
 
            void
            buildRightLocation()
            { M_parser.doBuildRightLocation(); }
 
            void
            buildCenterLocation()
            { M_parser.doBuildCenterLocation(); }
 
            void
            buildPenaltyLocation()
            { M_parser.doBuildPenaltyLocation(); }
 
            void
            buildGoalLocation()
            { M_parser.doBuildGoalLocation(); }

            void
            buildFlagOffset( int offset )
            { M_parser.doBuildFlagOffset( offset ); }

            void
            buildFlag( double dist, double dir,
                       double dist_chg, double dir_chg )
            { M_parser.doBuildFlag( dist, dir, dist_chg, dir_chg ); }

            void
            buildFlag( bool close, double dist, double dir )
            { M_parser.doBuildFlag( close, dist, dir ); }

            void
            buildFlag( bool close, double dir ) 
            { M_parser.doBuildFlag( close, dir ); }

            void
            buildGoal( double dist, double dir,
                       double dist_chg, double dir_chg )
            { M_parser.doBuildGoal( dist, dir, dist_chg, dir_chg ); }

            void
            buildGoal( bool close, double dist, double dir )
            { M_parser.doBuildGoal( close, dist, dir ); }

            void
            buildGoal( bool close, double dir ) 
            { M_parser.doBuildGoal( close, dir ); }

            void
            buildGlobalGoal( double x, double y )
            { M_parser.doBuildGlobalGoal( x, y ); }

            void
            buildPlayerVisBall( double dist, double dir,
                       double dist_chg, double dir_chg )
            { M_parser.doBuildPlayerVisBall( dist, dir, dist_chg, dir_chg ); }

            void
            buildBall( bool close, double dist, double dir )
            { M_parser.doBuildBall( close, dist, dir ); }

            void
            buildBall( bool close, double dir )
            { M_parser.doBuildBall( close, dir ); }

            void
            buildGlobalBall( double x, double y,
                             double delta_x, double delta_y )
            { M_parser.doBuildGlobalBall( x, y, delta_x, delta_y ); }

            void
            buildPlayer( double dist, double dir,
                         double dist_chg, double dir_chg,
                         double orientation, double head_orientation,
                         double point_dir, bool tackling )
            { 
                M_parser.doBuildPlayer( dist, dir, dist_chg, dir_chg,
                                        orientation, head_orientation, point_dir, tackling );
            }

            void
            buildPlayer( double dist, double dir,
                         double dist_chg, double dir_chg,
                         double point_dir, bool tackling )
            {
                M_parser.doBuildPlayer( dist, dir, dist_chg, dir_chg,
                                        point_dir, tackling );
            }

            void
            buildPlayer( double dist, double dir ,
                         double point_dir, bool tackling )
            { M_parser.doBuildPlayer( dist, dir, point_dir, tackling ); }

            void
            buildPlayer( double dist, double dir,
                         double dist_chg, double dir_chg,
                         double orientation, double head_orientation,
                         bool tackling )
            { 
                M_parser.doBuildPlayer( dist, dir, dist_chg, dir_chg,
                                        orientation, head_orientation, tackling );
            }

            void
            buildPlayer( double dist, double dir,
                         double dist_chg, double dir_chg,
                         bool tackling )
            { M_parser.doBuildPlayer( dist, dir, dist_chg, dir_chg, tackling ); }

            void
            buildPlayer( bool close, double dist, double dir ,
                         bool tackling )
            { M_parser.doBuildPlayer( close, dist, dir, tackling ); }

            void
            buildPlayer( bool close, double dir )
            { M_parser.doBuildPlayer( close, dir ); }

            void
            buildGlobalPlayer( double x, double y,
                               double delta_x, double delta_y,
                               double orientation, double head_orientation, 
                               bool tackling )
            { 
                M_parser.doBuildGlobalPlayer( x, y, delta_x, delta_y,
                                              orientation, head_orientation,
                                              tackling );
            }

            void
            buildGlobalPlayer( double x, double y,
                               double delta_x, double delta_y,
                               double orientation, double head_orientation, 
                               double point_dir, bool tackling )
            { 
                M_parser.doBuildGlobalPlayer( x, y, delta_x, delta_y,
                                              orientation, head_orientation,
                                              point_dir, tackling );
            }

            void
            buildTeamName( const std::string& name ) 
            { M_parser.doBuildTeamName( name ); }

            void
            buildUNum( int unum )
            { M_parser.doBuildUNum( unum ); }

            void
            buildGoalie()
            { M_parser.doBuildGoalie(); }

            void
            buildScore( int time, int our, int opp )
            { M_parser.doBuildScore( time, our, opp ); }

            void
            buildInit( int unum )
            { M_parser.doBuildInit( unum ); }

            void
            buildInit()
            { M_parser.doBuildInit(); }
    
            void
            buildCoachInit()
            { M_parser.doBuildCoachInit(); }
    
            void
            buildCoachAudio( int time, const std::string& msg )
            { M_parser.doBuildCoachAudio( time, msg ); }

            void
            buildRefAudio( int time )
            { M_parser.doBuildRefAudio( time ); }
    
            void
            buildGoalRefAudio( int time, int score )
            { M_parser.doBuildGoalRefAudio( time, score ); }

            void
            buildUnknownRefAudio( int time, const std::string& message )
            { M_parser.doBuildUnknownRefAudio( time, message ); }

            void
            buildPlayerAudio( int time, double dir, const std::string& msg )
            { M_parser.doBuildPlayerAudio( time, dir, msg ); }

            void
            buildPlayerAudio( int time, const std::string& msg )
            { M_parser.doBuildPlayerAudio( time, msg ); }

            void
            buildPlayerAudio( int time, double dir, int unum,
                              const std::string& msg )
            { M_parser.doBuildPlayerAudio( time, dir, unum, msg ); }

            void
            buildPlayerAudio( int time )
            { M_parser.doBuildPlayerAudio( time ); }

            void
            buildPlayerAudio( int time, int unum )
            { M_parser.doBuildPlayerAudio( time, unum ); }

            void
            buildSelfAudio( int time, const std::string& msg )
            { M_parser.doBuildSelfAudio( time, msg ); }

            void
            buildSubstitution( int unum, int type )
            { M_parser.doBuildSubstitution( unum, type ); }

            void
            buildSubstitution( int unum )
            { M_parser.doBuildSubstitution( unum ); }

            void
	        buildClangPlayerVersionMsg( )
            { M_parser.doBuildClangPlayerVersionMsg( ); }

            void
	        buildClangPlayerVersion( int min, int max )
            { M_parser.doBuildClangPlayerVersion( min, max ); }

            void
            buildCantReconnect()
            { M_parser.doBuildCantReconnect(); }

            void
            buildInitError()
            { M_parser.doBuildInitError(); }

            void
            buildNoMoreTeamOrPlayerOrGoalieError()
            { M_parser.doBuildNoMoreTeamOrPlayerOrGoalieError(); }
    
            void
            buildNoSuchTeamOrAlreadyHaveCoachError()
            { M_parser.doBuildNoSuchTeamOrAlreadyHaveCoachError(); }

            void
            buildTooManyMovesError()
            { M_parser.doBuildTooManyMovesError(); }

            void
            buildUnknownCommandError()
            { M_parser.doBuildUnknownCommandError(); }

            void
            buildIllegalCommandError()
            { M_parser.doBuildIllegalCommandError(); }

            void
            buildSayMessageTooLongError()
            { M_parser.doBuildSayMessageTooLongError(); }

            void
            buildIllegalModeError()
            { M_parser.doBuildIllegalModeError(); }

            void
            buildIllegalObjectFormError()
            { M_parser.doBuildIllegalObjectFormError(); }

            void
            buildSaidTooManyFreeformMessagesError()
            { M_parser.doBuildSaidTooManyFreeformMessagesError(); }

            void
            buildCannotSayFreeformWhilePlayonError()
            { M_parser.doBuildCannotSayFreeformWhilePlayonError(); }

            void
            buildSaidTooManyMetaMessagesError()
            { M_parser.doBuildSaidTooManyMetaMessagesError(); }

            void
            buildSaidTooManyAdviceMessagesError()
            { M_parser.doBuildSaidTooManyAdviceMessagesError(); }

            void
            buildSaidTooManyInfoMessagesError()
            { M_parser.doBuildSaidTooManyInfoMessagesError(); }

            void
            buildSaidTooManyDefineMessagesError()
            { M_parser.doBuildSaidTooManyDefineMessagesError(); }

            void
            buildCouldNotParseSayError()
            { M_parser.doBuildCouldNotParseSayError(); }

            void
            buildSaidTooManyMessagesError()
            { M_parser.doBuildSaidTooManyMessagesError(); }

            void
            buildUnknownError( const std::string& error )
            { M_parser.doBuildUnknownError( error ); }

            void
            buildNoTeamFoundWarning()
            { M_parser.doBuildNoTeamFoundWarning(); }

            void
            buildNoSuchPlayerWarning()
            { M_parser.doBuildNoSuchPlayerWarning(); }

            void
            buildCannotSubWhilePlayOnWarning()
            { M_parser.doBuildCannotSubWhilePlayOnWarning(); }

            void
            buildNoSubsLeftWarning()
            { M_parser.doBuildNoSubsLeftWarning(); }

            void
            buildMaxOfThatPlayerTypeOnFieldWarning()
            { M_parser.doBuildMaxOfThatPlayerTypeOnFieldWarning(); }

            void
            buildCannotChangeGoalieWarning()
            { M_parser.doBuildCannotChangeGoalieWarning(); }

            void
            buildCompressionWarning() 
            { M_parser.doBuildCompressionWarning(); }

            void
            buildUnknownWarning( const std::string& warning )
            { M_parser.doBuildUnknownWarning( warning ); }

            void
            buildClangVerOK( int min, int max )
            { M_parser.doBuildClangVerOK( min, max ); }

            void
            buildEyeOK( bool on )
            { M_parser.doBuildEyeOK( on ); }

            void
            buildSayOK( )
            { M_parser.doBuildSayOK( ); }

            void
            buildChangePlayerTypeOK( int unum, int type )
            { M_parser.doBuildChangePlayerTypeOK( unum, type ); }
 
            void
            buildCompressionOK( int level ) 
            { M_parser.doBuildCompressionOK( level ); }


            void
            buildUnknownOK( const std::string& msg )
            { M_parser.doBuildUnknownOK( msg ); }
        };

    private:

        Param M_param;
        rcss::Parser& M_clang_parser;

        friend int RCC_lex( rcc::Holder*,
                            rcc::Parser::Param& param );
        friend rcc::Parser::Param& getParam(void*);
        friend class Param;
        friend class RCCLexer;

      
        virtual
        bool
        doParse( std::istream& strm )
        {
            return ( setStream( strm )
                     && ( ::RCC_parse( (void*)&M_param ) == 0 ) );
        }

    
        bool
        setStream( std::istream& strm );

        bool
        parseCLang( const char* message );
    public:

        /** Parser constructor.
         *
         *  You will need to call this constuctor from your subclass's contructor.
         *  The clang_parser parameter is a reference to another parser which will
         *  be used for parsing CLang messages.  This is compatible with the CLang      
         *  parser that is part of the simulator.
         *
         */
        Parser( rcss::Parser& clang_parser )
            : M_param( *this ),
              M_clang_parser( clang_parser )
        {}

        virtual
        ~Parser()
        {}

#ifdef DOXYGEN_ONLY
/* The following functions are actually inherited from rcss::Parser,
   but I cannot find a way to document them as such. */

        //! \name Parsing Functions
        //@{
        /** Main parsing function.
         *
         *  This is the main parsing function of the rccparser library.  Pass the
         *  stream you wish to parse to commence parsing. 
         *  
         *  \param strm The stream on which parsing will occur.
         *
         *  \return True if there where no errors, false otherwise.
         */
        bool
        parse( std::istream& strm );

        /** Secondary parsing function.
         *
         *  Sometimes it may be usefull to parse data that has been
         *  stored in a file (usually to detect errors).  This function
         *  opens the file specfied and passes the std::ifstream to
         *  parse( std::istream& strm ).
         *  
         *  \param file The name of the file on which parsing will occur.
         *
         *  \return True if there where no errors, false otherwise.
         */
        bool
        parse( const std::string& file );
        //@}
#endif
    protected:
        /** This function is called every time a message has been parsed.
         *
         *  Override this function in your subclass to execute code \b after a
         *  message has been parsed.
         *
         *  \pre A message has been parsed.
         */
        virtual
        void
        doMsgParsed() {} 

        //! \name CLang Parsing Functions
        //@{
        /** This function is called every time a CLang message has been parsed,
         *  if doParsedCLang( const char* msg ) has not been overridden.
         *
         *  Override this function in your subclass to
         *  execute code \b after a CLang message has been parsed.
         *
         *  \param msg The CLang message that has just been parsed
         *
         *  \warning This function is deprecated. Override doParsedCLang( const char* msg )
         *           instead.
         *
         *  \pre doParsedCLang( const char* msg ) has not been overridden.
         *
         *  \pre A CLang message has been parsed.
         *
         *  \sa 
         *      - doParsedCLang( const char* msg ).
         */
        virtual
        void
        doParsedCLang( const std::string& msg ) 
        {}


        /** This function is called every time a CLang message has been parsed.
         *
         *  Override this function in your subclass to
         *  execute code \b after a CLang message has been parsed.
         *
         *  If this function is not overridden then the default behaviour is to call
         *  the deprecated function doParsedCLang( const std::string& msg ).
         *
         *  \param msg Contains the CLang message that has just been parsed.
         *
         *  \pre A CLang message has been parsed.
         *
         *  \sa 
         *      - doParsedCLang( const std::string& msg ).
         */
        virtual
        void
        doParsedCLang( const char* msg ) 
        { doParsedCLang( std::string( msg ) ); }
        //@}


        //! \name Full State Parsing Functions
        //@{
        /** This function is called after parsing a full state message.
         *
         *  Override this function in your subclass to handle full
         *  state messages.  You will also need to overwrite the functions for parsing the
         *  subsections of the full state message if you want to be able to
         *  use the data from those subsections.
         *
         *  This message is only sent to field players and only if sending of
         *  fullstate messages is specifically turned on in the simulator.
         *
         *  \param time The time (in server cycles) that the full state
         *              message was sent by the simulator.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for
         *       the team this client is in.
         *
         *  \pre An entire full state message has been parsed.
         *  
         *  The following is an example (with line breaks added for clarity)
         *  of a version 8 or 9 full state message:
         *  \verbatim
            (fullstate 28 (pmode kick_off_l) 
                          (vmode high normal)
                          (count 0 0 0 0 0 0 0 0)
                          (arm (movable 0) (expires 0) (target 0 0) (count 0))
                          (score 0 0)
                          ((b) 0 0 0 0)
                          ((p l 1 g) -3 -37 0 0 0 0 (stamina 4000 1 1))
                          ((p l 2 0) -10 -38 0 0 0 0 (stamina 4000 1 1)))\endverbatim
         *
         *  The following is is example (with line breaks added for clarity) in version 7:
         *  \verbatim
            (fullstate 28 (pmode kick_off_l) 
                          (vmode high normal)
                          (score 0 0)
                          (ball 0 0 0 0)
                          (l_1 -3 -37 0 0 0 0 4000 1 1)
                          (l_2 -10 -38 0 0 0 0 4000 1 1))\endverbatim
         *
         *  In these examples the \e time parameter would be 28.
         *
         *  \sa 
         *      - doBuildCounts( int kick, int dash, int turn, int katch,
         *                     int move, int turn_neck, int change_view, int say )
         *      -doBuildArmState( int movable_in,
         *                        int expires_in,
         *                        double target_x,
         *                        double target_y,
         *                        int count )        
         *      - doBuildScore( int our, int opp )
         *      - doBuildBall( double x, double y, double delta_x, double delta_y )
         *      - doBuildPlayer( int unum,
         *                     double x, double y, 
         *                     double delta_x, double delta_y,
         *                     double orientation, double head_orientation,
         *                     int stamina, double effort, double recovery )
         *      - doBuildPlayer( int unum, int type,
         *                     double pos_x, double pos_y,
         *                     double vel_x, double vel_y,
         *                     double body_dir, double head_dir,
         *                     double stamina, double effort,
         *                     double recovery )
         *      - doBuildPlayer( int unum, int type,
         *                     double pos_x, double pos_y,
         *                     double vel_x, double vel_y,
         *                     double body_dir, double head_dir,
         *                     double arm_mag, double arm_dir,
         *                     double stamina, double effort,
         *                     double recovery )
         *      - doBuildGoalie( int unum,
         *                     double pos_x, double pos_y,
         *                     double vel_x, double vel_y,
         *                     double body_dir, double head_dir,
         *                     double stamina, double effort,
         *                     double recovery )
         *      - doBuildGoalie( int unum,
         *                     double pos_x, double pos_y,
         *                     double vel_x, double vel_y,
         *                     double body_dir, double head_dir,
         *                     double arm_mag, double arm_dir, 
         *                     double stamina, double effort,
         *                     double recovery )
         *      - doBuildCatchBallLeftPlayMode()
         *      - doBuildCatchBallRightPlayMode()
         *      - doBuildBeforeKickOffPlayMode()
         *      - doBuildTimeOverPlayMode()
         *      - doBuildPlayOnPlayMode()
         *      - doBuildKickOffLeftPlayMode()
         *      - doBuildKickOffRightPlayMode()
         *      - doBuildKickInLeftPlayMode()
         *      - doBuildKickInRightPlayMode()
         *      - doBuildFreeKickLeftPlayMode()
         *      - doBuildFreeKickRightPlayMode()
         *      - doBuildCornerKickLeftPlayMode()
         *      - doBuildCornerKickRightPlayMode()
         *      - doBuildGoalKickLeftPlayMode()
         *      - doBuildGoalKickRightPlayMode()
         *      - doBuildAfterGoalLeftPlayMode()
         *      - doBuildAfterGoalRightPlayMode()
         *      - doBuildDropBallPlayMode()
         *      - doBuildOffSideLeftPlayMode()
         *      - doBuildOffsideRightPlayMode()
         *      - doBuildPenaltyKickLeftPlayMode()
         *      - doBuildPenaltyKickRightPlayMode()
         *      - doBuildFirstHalfOverPlayMode()
         *      - doBuildPausePlayMode()
         *      - doBuildHumanPlayMode()
         *      - doBuildFoulLeftPlayMode()
         *      - doBuildFoulRightPlayMode()
         *      - doBuildFoulChargeLeftPlayMode()
         *      - doBuildFoulChargeRightPlayMode()
         *      - doBuildFoulPushLeftPlayMode()
         *      - doBuildFoulPushRightPlayMode()
         *      - doBuildFoulMultipleAttackerLeftPlayMode()
         *      - doBuildFoulMultipleAttackerRightPlayMode()
         *      - doBuildFoulBallOutLeftPlayMode()
         *      - doBuildFoulBallOutRightPlayMode()
         *      - doBuildBackPassLeftPlayMode()
         *      - doBuildBackPassRightPlayMode()
         *      - doBuildFreeKickFaultLeftPlayMode()
         *      - doBuildFreeKickFaultRightPlayMode()
         */
        virtual
        void
        doBuildFullState( int time ) {}
 

        /** This function is called after parsing
         *  the count section of a version 8 or 9 full state message.
         *
         *  Override this function in your subclass to handle the count section
         *  of a version 8 or 9 full state message.
         *
         *  \param kick The number of kick commands executed by the player.
         *  \param dash The number of dash commands executed by the player.
         *  \param turn The number of turn commands executed by the player.
         *  \param katch The number of catch commands executed by the player.
         *  \param move The number of move commands executed by the player.
         *  \param turn_neck The number of turn_neck commands executed by the player.
         *  \param change_view The number of change_view commands executed by the player.
         *  \param say The number of say commands executed by the player.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre The count section of the full state message has been parsed.
         *  
         *  The following is an example of the count section of a full state message:
         *  \verbatim
            (count 5 2 3 9 4 6 7 1)\endverbatim
         *
         *  In this example the:
         *      - \e kick        parameter would be 5.
         *      - \e dash        parameter would be 2.
         *      - \e turn        parameter would be 3.
         *      - \e katch       parameter would be 9.
         *      - \e move        parameter would be 4.
         *      - \e turn_neck   parameter would be 6.
         *      - \e change_view parameter would be 7.
         *      - \e say         parameter would be 1.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         */
        virtual
        void
        doBuildCounts( int kick, int dash, int turn, int katch,
                       int move, int turn_neck, int change_view, int say ) {}


        /** This function is called after parsing
         *  the score section of a full state message.
         *
         *  Override this function in your subclass to handle the score section
         *  of a full state message.
         *
         *  \param our If protocol version 7 is being used, then this is the
         *             number of goals the left team has scored.  Otherwise this
         *             is the number of goals the client's team has scored.
         *  \param our If protocol version 7 is being used, then this is the
         *             number of goals the right team has scored.  Otherwise this
         *             is the number of goals the opponent team has scored.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre The score section of the full state message has been parsed.
         *  
         *  The following is an example of the score section of a full state message:
         *  \verbatim
            (score 2 1)\endverbatim
         *
         *  In this example the:
         *      - \e our parameter would be 2.
         *      - \e opp parameter would be 1.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         */
        virtual
        void
        doBuildScore( int our, int opp ) {}


        /** This function is called after parsing
         *  the ball section of a full state message.
         *
         *  Override this function in your subclass to handle the ball section
         *  of a full state message.
         *
         *  \param x       The x co-ordinate of the ball.
         *  \param y       The y co-ordinate of the ball.
         *  \param delta_x The x velocity of the ball.
         *  \param delta_y The y velocity of the ball.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre The ball section of the full state message has been parsed.
         *  
         *  The following is a protocol version 8 or 9 example of the score section of a full state message:
         *  \verbatim
            ((b) -20.5 10.7 0.56 -0.3)\endverbatim
         *  The following is the same exmaple in protocol version 7:
         *  \verbatim
            (ball -20.5 10.7 0.56 -0.3)\endverbatim

         *  In these examples the:
         *      - \e x       parameter would be -20.5.
         *      - \e y       parameter would be 10.7.
         *      - \e delta_x parameter would be 0.56.
         *      - \e delta_y parameter would be -0.3.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         */
        virtual
        void
        doBuildBall( double x, double y, 
                     double delta_x, double delta_y ) {}


        /** This function is called after parsing
         *  a player section of a version 7 full state message.
         *
         *  Override this function in your subclass to handle player sections
         *  of version 7 full state messages.
         *
         *  \param unum             The uniform number of the player.
         *  \param x                The x co-ordinate of the player.
         *  \param y                The y co-ordinate of the player.
         *  \param delta_x          The x velocity of the player.
         *  \param delta_y          The y velocity of the player.
         *  \param orientation      The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param stamina          The player's stamina.
         *  \param effort           The player's effort. 
         *  \param recovery         The player's recovery.
         *
         *  \pre The client is using protocol version 7.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre A player section of a full state message has been parsed.
         *  
         *  The following is an example of the player section of a version 7 full state message:
         *  \verbatim
            (l_5 -20.5 10.7 0.56 -0.3 13.5 4.8 2964.5 0.93 0.54)\endverbatim
         *
         *  In this examples the:
         *      - \e unum             parameter would be 5.
         *      - \e x                parameter would be -20.5.
         *      - \e y                parameter would be 10.7.
         *      - \e delta_x          parameter would be 0.56.
         *      - \e delta_y          parameter would be -0.3.
         *      - \e orientation      parameter would be 13.5.
         *      - \e head_orientation parameter would be 4.8.
         *      - \e stamina          parameter would be 2965.
         *      - \e effort           parameter would be 0.93.
         *      - \e recovery         parameter would be 0.54.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildLeftSide()
         *      - doBuildRightSide()
        */
        virtual
        void
        doBuildPlayer( int unum,
                       double x, double y, 
                       double delta_x, double delta_y,
                       double orientation, double head_orientation,
                       int stamina, double effort, double recovery ) {}


        /** This function is called after parsing a player
         *  section of a version 8 or 9 full state message, where the player is
         *  \b not pointing and the player is \b not a goalie.
         *
         *  Override this function in your subclass to handle non-pointing
         *  non-goalie player sections of version 8 or 9 full state messages.
         *
         *  \param unum             The uniform number of the player.
         *  \param type             The player's type.
         *  \param x                The x co-ordinate of the player.
         *  \param y                The y co-ordinate of the player.
         *  \param delta_x          The x velocity of the player.
         *  \param delta_y          The y velocity of the player.
         *  \param orientation      The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param stamina          The player's stamina.
         *  \param effort           The player's effort. 
         *  \param recovery         The player's recovery.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre A non-pointing, non-goalie player section of a version 8 or 9 full state message has been parsed.
         *  
         *  The following is an example of a non-pointing non-goalie player
         *  section of a version 8 or 9 full state message:
         *  \verbatim
            ((p l 5 4) -20.5 10.7 0.56 -0.3 13.5 4.8 (stamina 2964.5 0.93 0.54))\endverbatim
         *
         *  In this examples the:
         *      - \e unum             parameter would be 5.
         *      - \e type             parameter would be 4.
         *      - \e x                parameter would be -20.5.
         *      - \e y                parameter would be 10.7.
         *      - \e delta_x          parameter would be 0.56.
         *      - \e delta_y          parameter would be -0.3.
         *      - \e orientation      parameter would be 13.5.
         *      - \e head_orientation parameter would be 4.8.
         *      - \e stamina          parameter would be 2964.5.
         *      - \e effort           parameter would be 0.93.
         *      - \e recovery         parameter would be 0.54.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildLeftSide()
         *      - doBuildRightSide()
         */
        virtual
        void
        doBuildPlayer( int unum, int type,
                       double pos_x, double pos_y,
                       double vel_x, double vel_y,
                       double body_dir, double head_dir,
                       double stamina, double effort,
                       double recovery ) {}
       
       
        /** This function is called after parsing a player
         *  section of a version 8 or 9 full state message, where the player
         *  \b is pointing and the player is \b not a goalie.
         *
         *  Override this function in your subclass to handle pointing
         *  non-goalie player sections of version 8 or 9 full state messages.
         *
         *  \param unum             The uniform number of the player.
         *  \param type             The player's type.
         *  \param x                The x co-ordinate of the player.
         *  \param y                The y co-ordinate of the player.
         *  \param delta_x          The x velocity of the player.
         *  \param delta_y          The y velocity of the player.
         *  \param orientation      The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param arm_mag          The distance from the center of the field to the point
         *                          the player is pointing to.
         *  \param arm_dir          The direction the player is pointing in.
         *  \param stamina          The player's stamina.
         *  \param effort           The player's effort. 
         *  \param recovery         The player's recovery.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre A pointing non-goalie player section of a version 8 or 9 full state message has been parsed.
         *  
         *  The following is an example of a pointing non-goalie player
         *  section of a version 8 or 9 full state message:
         *  \verbatim
            ((p l 5 4) -20.5 10.7 0.56 -0.3 13.5 4.8 13.6 -10.2 (stamina 2964.5 0.93 0.54))\endverbatim
         *
         *  In this examples the:
         *      - \e unum             parameter would be 5.
         *      - \e type             parameter would be 4.
         *      - \e x                parameter would be -20.5.
         *      - \e y                parameter would be 10.7.
         *      - \e delta_x          parameter would be 0.56.
         *      - \e delta_y          parameter would be -0.3.
         *      - \e orientation      parameter would be 13.5.
         *      - \e head_orientation parameter would be 4.8.
         *      - \e arm_mag          parameter would be 13.6.
         *      - \e arm_dir          parameter would be -10.2.
         *      - \e stamina          parameter would be 2964.5.
         *      - \e effort           parameter would be 0.93.
         *      - \e recovery         parameter would be 0.54.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildLeftSide()
         *      - doBuildRightSide()
         */
        virtual
        void
        doBuildPlayer( int unum, int type,
                       double pos_x, double pos_y,
                       double vel_x, double vel_y,
                       double body_dir, double head_dir,
                       double arm_mag, double arm_dir,
                       double stamina, double effort,
                       double recovery ) {}
             

        /** This function is called after parsing a player
         *  section of a version 8 or 9 full state message, where the player is
         *  \b not pointing and the player \b is a goalie.
         *
         *  Override this function in your subclass to handle non-pointing
         *  goalie player sections of version 8 or 9 full state messages.
         *
         *  \param unum             The uniform number of the player.
         *  \param x                The x co-ordinate of the player.
         *  \param y                The y co-ordinate of the player.
         *  \param delta_x          The x velocity of the player.
         *  \param delta_y          The y velocity of the player.
         *  \param orientation      The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param stamina          The player's stamina.
         *  \param effort           The player's effort. 
         *  \param recovery         The player's recovery.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre A non-pointing, goalie player section of a version 8 or 9 full state message has been parsed.
         *  
         *  The following is an example of a non-pointing goalie player
         *  section of a version 8 or 9 full state message:
         *  \verbatim
            ((p l 5 g) -20.5 10.7 0.56 -0.3 13.5 4.8 (stamina 2964.5 0.93 0.54))\endverbatim
         *
         *  In this examples the:
         *      - \e unum             parameter would be 5.
         *      - \e x                parameter would be -20.5.
         *      - \e y                parameter would be 10.7.
         *      - \e delta_x          parameter would be 0.56.
         *      - \e delta_y          parameter would be -0.3.
         *      - \e orientation      parameter would be 13.5.
         *      - \e head_orientation parameter would be 4.8.
         *      - \e stamina          parameter would be 2964.5.
         *      - \e effort           parameter would be 0.93.
         *      - \e recovery         parameter would be 0.54.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildLeftSide()
         *      - doBuildRightSide()
         */
        virtual
        void
        doBuildGoalie( int unum,
                       double pos_x, double pos_y,
                       double vel_x, double vel_y,
                       double body_dir, double head_dir,
                       double stamina, double effort,
                       double recovery ) {}


        /** This function is called after parsing a player
         *  section of a version 8 or 9 full state message, where the player
         *  \b is pointing and the player \b is a goalie.
         *
         *  Override this function in your subclass to handle pointing
         *  goalie player sections of version 8 or 9 full state messages.
         *
         *  \param unum             The uniform number of the player.
         *  \param x                The x co-ordinate of the player.
         *  \param y                The y co-ordinate of the player.
         *  \param delta_x          The x velocity of the player.
         *  \param delta_y          The y velocity of the player.
         *  \param orientation      The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param arm_mag          The distance from the center of the field to the point
         *                          the player is pointing to.
         *  \param arm_dir          The direction the player is pointing in.
         *  \param stamina          The player's stamina.
         *  \param effort           The player's effort. 
         *  \param recovery         The player's recovery.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre Full state message sending is turned on in the simulator for the team this client is in.
         *
         *  \pre A pointing goalie player section of a version 8 or 9 full state message has been parsed.
         *  
         *  The following is an example of a pointing goalie player
         *  section of a version 8 or 9 full state message:
         *  \verbatim
            ((p l 5 g) -20.5 10.7 0.56 -0.3 13.5 4.8 13.6 -10.2 (stamina 2964.5 0.93 0.54))\endverbatim
         *
         *  In this examples the:
         *      - \e unum             parameter would be 5.
         *      - \e x                parameter would be -20.5.
         *      - \e y                parameter would be 10.7.
         *      - \e delta_x          parameter would be 0.56.
         *      - \e delta_y          parameter would be -0.3.
         *      - \e orientation      parameter would be 13.5.
         *      - \e head_orientation parameter would be 4.8.
         *      - \e arm_mag          parameter would be 13.6.
         *      - \e arm_dir          parameter would be -10.2.
         *      - \e stamina          parameter would be 2964.5.
         *      - \e effort           parameter would be 0.93.
         *      - \e recovery         parameter would be 0.54.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildLeftSide()
         *      - doBuildRightSide()
         */
        virtual
        void
        doBuildGoalie( int unum,
                       double pos_x, double pos_y,
                       double vel_x, double vel_y,
                       double body_dir, double head_dir,
                       double arm_mag, double arm_dir, 
                       double stamina, double effort,
                       double recovery ) {}
        //@}


        //! \name PlayMode Parsing Functions
        //@{
        /** This function is called after parsing
         *  a \c goalie_catch_ball_l play mode.
         *
         *  Override this function in your subclass to handle the \c goalie_catch_ball_l play mode.
         *
         *  \note: This is and \c goalie_catch_ball_r are only ever sent as part
         *         full state message.
         *
         *  \pre A full state message is being parsed.
         *
         *  \pre A \c goalie_catch_ball_l play mode has just been parsed.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildCatchBallRightPlayMode()
         */
        virtual
        void
        doBuildCatchBallLeftPlayMode() {}


        /** This function is called after parsing
         *  a \c goalie_catch_ball_r play mode.
         *
         *  Override this function in your subclass to handle the \c goalie_catch_ball_r play mode.
         *
         *  \note: This is and \c goalie_catch_ball_l are only ever sent as part
         *         full state message.
         *
         *  \pre A full state message is being parsed.
         *
         *  \pre A \c goalie_catch_ball_r play mode has just been parsed.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildCatchBallLeftPlayMode()
         */
        virtual
        void
        doBuildCatchBallRightPlayMode() {}


        /** This function is called after parsing
         *  a \c before_kick_off play mode.
         *
         *  Override this function in your subclass to handle the \c before_kick_off play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c before_kick_off play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildBeforeKickOffPlayMode() {}


        /** This function is called after parsing of
         *  a \c time_over play mode.
         *
         *  Override this function in your subclass to handle the \c time_over play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c time_over play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildTimeOverPlayMode() {}


        /** This function is called after parsing of
         *  a \c play_on play mode.
         *
         *  Override this function in your subclass to handle the \c play_on play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c play_on play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPlayOnPlayMode() {}


        /** This function is called after parsing of
         *  a \c kick_off_l play mode.
         *
         *  Override this function in your subclass to handle the \c kick_off_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c kick_off_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildKickOffLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c kick_off_r play mode.
         *
         *  Override this function in your subclass to handle the \c kick_off_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c kick_off_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildKickOffRightPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c kick_in_l play mode.
         *
         *  Override this function in your subclass to handle the \c kick_in_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c kick_in_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildKickInLeftPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c kick_in_r play mode.
         *
         *  Override this function in your subclass to handle the \c kick_in_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c kick_in_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildKickInRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c free_kick_l play mode.
         *
         *  Override this function in your subclass to handle the \c free_kick_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c free_kick_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFreeKickLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c free_kick_r play mode.
         *
         *  Override this function in your subclass to handle the \c free_kick_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c free_kick_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFreeKickRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c corner_kick_l play mode.
         *
         *  Override this function in your subclass to handle the \c corner_kick_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c corner_kick_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildCornerKickLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c corner_kick_r play mode.
         *
         *  Override this function in your subclass to handle the \c corner_kick_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c corner_kick_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildCornerKickRightPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c goal_kick_l play mode.
         *
         *  Override this function in your subclass to handle the \c goal_kick_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c goal_kick_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildGoalKickLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c goal_kick_r play mode.
         *
         *  Override this function in your subclass to handle the \c goal_kick_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c goal_kick_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildGoalKickRightPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c goal_l play mode.
         *
         *  Override this function in your subclass to handle the \c goal_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c goal_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildAfterGoalLeftPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c goal_r play mode.
         *
         *  Override this function in your subclass to handle the \c goal_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c goal_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildAfterGoalRightPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c drop_ball play mode.
         *
         *  Override this function in your subclass to handle the \c drop_ball play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c drop_ball play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildDropBallPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c offside_l play mode.
         *
         *  Override this function in your subclass to handle the \c offside_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c offside_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildOffSideLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c offside_r play mode.
         *
         *  Override this function in your subclass to handle the \c offside_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c offside_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildOffsideRightPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c penalty_kick_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_kick_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_kick_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyKickLeftPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c penalty_kick_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_kick_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_kick_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyKickRightPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c first_half_over play mode.
         *
         *  Override this function in your subclass to handle the \c first_half_over play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c first_half_over play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFirstHalfOverPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c pause play mode.
         *
         *  Override this function in your subclass to handle the \c pause play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c pause play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPausePlayMode() {}
 

        /** This function is called after parsing of
         *  a \c human_judge play mode.
         *
         *  Override this function in your subclass to handle the \c human_judge play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c human_judge play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildHumanPlayMode() {}
 

        /** This function is called after parsing of
         *  a \c foul_l play mode.
         *
         *  Override this function in your subclass to handle the \c foul_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_r play mode.
         *
         *  Override this function in your subclass to handle the \c foul_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_charge_l play mode.
         *
         *  Override this function in your subclass to handle the \c foul_charge_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_charge_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulChargeLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_charge_r play mode.
         *
         *  Override this function in your subclass to handle the \c foul_charge_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_charge_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulChargeRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_push_l play mode.
         *
         *  Override this function in your subclass to handle the \c foul_push_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_push_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulPushLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_push_r play mode.
         *
         *  Override this function in your subclass to handle the \c foul_push_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_push_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulPushRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_multiple_attack_l play mode.
         *
         *  Override this function in your subclass to handle the \c foul_multiple_attack_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_multiple_attack_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulMultipleAttackerLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_multiple_attack_r play mode.
         *
         *  Override this function in your subclass to handle the \c foul_multiple_attack_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_multiple_attack_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulMultipleAttackerRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_ballout_l  play mode.
         *
         *  Override this function in your subclass to handle the \c foul_ballout_l  play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_ballout_l  play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulBallOutLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c foul_ballout_r  play mode.
         *
         *  Override this function in your subclass to handle the \c foul_ballout_r  play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c foul_ballout_r  play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFoulBallOutRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c back_pass_l play mode.
         *
         *  Override this function in your subclass to handle the \c back_pass_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c back_pass_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildBackPassLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c back_pass_r play mode.
         *
         *  Override this function in your subclass to handle the \c back_pass_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c back_pass_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildBackPassRightPlayMode() {}


        /** This function is called after parsing of
         *  a \c free_kick_fault_l play mode.
         *
         *  Override this function in your subclass to handle the \c free_kick_fault_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c free_kick_fault_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFreeKickFaultLeftPlayMode() {}


        /** This function is called after parsing of
         *  a \c free_kick_fault_r play mode.
         *
         *  Override this function in your subclass to handle the \c free_kick_fault_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c free_kick_fault_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildFreeKickFaultRightPlayMode() {}

        /** This function is called after parsing of
         *  a \c catch_fault_l play mode.
         *
         *  Override this function in your subclass to handle the \c catch_fault_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c catch_fault_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildCatchFaultLeftPlayMode() {}

        /** This function is called after parsing of
         *  a \c catch_fault_r play mode.
         *
         *  Override this function in your subclass to handle the \c catch_fault_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c catch_fault_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildCatchFaultRightPlayMode() {}
        
       /** This function is called after parsing of
         *  a \c indirect_free_kick_l play mode.
         *
         *  Override this function in your subclass to handle the \c indirect_free_kick_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c indirect_free_kick_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildIndirectFreeKickLeftPlayMode() {}
        
       /** This function is called after parsing of
         *  a \c indirect_free_kick_r play mode.
         *
         *  Override this function in your subclass to handle the \c indirect_free_kick_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c indirect_free_kick_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildIndirectFreeKickRightPlayMode() {}

       /** This function is called after parsing of
         *  a \c penalty_setup_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_setup_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_setup_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltySetupLeftPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_setup_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_setup_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_setup_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltySetupRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_ready_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_ready_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_ready_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyReadyLeftPlaymode() {}
 
       /** This function is called after parsing of
         *  a \c penalty_ready_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_ready_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_ready_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyReadyRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_taken_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_taken_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_taken_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyTakenLeftPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_taken_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_taken_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_taken_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyTakenRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_miss_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_miss_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_miss_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyMissLeftPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_miss_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_miss_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_miss_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyMissRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_score_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_score_l play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_score_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyScoreLeftPlaymode() {}
 
       /** This function is called after parsing of
         *  a \c penalty_score_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_score_r play mode.
         *
         *  \pre Either:
         *            - a full state message is being parsed,
         *            - a init or reconnect message is being parsed or
         *            - a referee audio message is beign parsed.
         *
         *  \pre A \c penalty_score_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildFullState( int time )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyScoreRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_onfield_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_onfield_l play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_onfield_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyOnFieldLeftPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_onfield_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_onfield_r play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_onfield_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyOnFieldRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_foul_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_foul_l play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_foul_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyFoulLeftPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_foul_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_foul_r play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_foul_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyFoulRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_winner_l play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_winner_l play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_winner_l play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyWinnerLeftPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_winner_r play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_winner_r play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_winner_r play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyWinnerRightPlaymode() {}
        
       /** This function is called after parsing of
         *  a \c penalty_draw play mode.
         *
         *  Override this function in your subclass to handle the \c penalty_draw play mode.
         *
         *  \pre A referee audio message is beign parsed.
         *
         *  \pre A \c penalty_draw play mode has just been parsed.
         *
         *  \sa
         *      - doBuildRefAudio( int time )
         */
        virtual
        void
        doBuildPenaltyDrawPlaymode() {}

        //@}
      
        //! \name Side Parsing Functions
        //@{
        /** This function is called after parsing a token representing the left team side.
         *
         *  Override this function in your subclass to handle left team sides.
         *
         *  \pre Either:
         *            - a player section of a full state message is being parsed,
         *            - an init message or reconnect message is being parsed,
         *            - a focus section of a sense body message is being parsed,
         *              where the client is focused on a player,
         *            - a referee goal audio message is being parsed or
         *            - a non CLang coach audio message is being parsed.
         *
         *  \pre A token representing the left team has been parsed.
         *
         *  \sa
         *      - doBuildPlayer( int unum,
         *                       double x, double y, 
         *                       double delta_x, double delta_y,
         *                       double orientation, double head_orientation,
         *                       int stamina, double effort, double recovery )
         *      - doBuildPlayer( int unum, int type,
         *                       double pos_x, double pos_y,
         *                       double vel_x, double vel_y,
         *                       double body_dir, double head_dir,
         *                       double arm_mag, double arm_dir,
         *                       double stamina, double effort,
         *                       double recovery )
         *      - doBuildGoalie( int unum,
         *                       double pos_x, double pos_y,
         *                       double vel_x, double vel_y,
         *                       double body_dir, double head_dir,
         *                       double stamina, double effort,
         *                       double recovery )
         *      - doBuildGoalie( int unum,
         *                       double pos_x, double pos_y,
         *                       double vel_x, double vel_y,
         *                       double body_dir, double head_dir,
         *                       double arm_mag, double arm_dir, 
         *                       double stamina, double effort,
         *                       double recovery )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildCoachInit()
         *      - doBuildFocusState( int unum, int count )
         *      - doBuildGoalRefAudio( int time, int score )
         *      - doBuildCoachAudio( int time, const std::string& msg )
         *
         */
        virtual
        void
        doBuildLeftSide() {}

        /** This function is called after parsing a token representing the right team side.
         *
         *  Override this function in your subclass to handle right team sides.
         *
         *  \pre Either:
         *            - a player section of a full state message is being parsed,
         *            - an init message or reconnect message is being parsed,
         *            - a focus section of a sense body message is being parsed,
         *            - a referee goal audio message is being parsed or
         *            - a non-CLang coach audio message is being parsed.
         *
         *  \pre A token representing the right team has been parsed.
         *
         *  \sa
         *      - doBuildPlayer( int unum,
         *                       double x, double y, 
         *                       double delta_x, double delta_y,
         *                       double orientation, double head_orientation,
         *                       int stamina, double effort, double recovery )
         *      - doBuildPlayer( int unum, int type,
         *                       double pos_x, double pos_y,
         *                       double vel_x, double vel_y,
         *                       double body_dir, double head_dir,
         *                       double arm_mag, double arm_dir,
         *                       double stamina, double effort,
         *                       double recovery )
         *      - doBuildGoalie( int unum,
         *                       double pos_x, double pos_y,
         *                       double vel_x, double vel_y,
         *                       double body_dir, double head_dir,
         *                       double stamina, double effort,
         *                       double recovery )
         *      - doBuildGoalie( int unum,
         *                       double pos_x, double pos_y,
         *                       double vel_x, double vel_y,
         *                       double body_dir, double head_dir,
         *                       double arm_mag, double arm_dir, 
         *                       double stamina, double effort,
         *                       double recovery )
         *      - doBuildInit( int unum )
         *      - doBuildInit()
         *      - doBuildCoachInit()
         *      - doBuildFocusState( int unum, int count )
         *      - doBuildGoalRefAudio( int time, int score )
         *      - doBuildCoachAudio( int time, const std::string& msg )
         *
         */
        virtual
        void
        doBuildRightSide() {}

        /** This function is called after parsing a token representing the client's team's side.
         *
         *  Override this function in your subclass to handle client team sides.
         *
         *  \pre A version 8 or 9 player audio message is being parsed.
         *
         *  \pre A token representing the client's team has been parsed.
         *
         *  \sa 
         *      - doBuildPlayerAudio( int time, double dir, const std::string& msg )
         *      - doBuildPlayerAudio( int time, double dir, int unum,
         *                            const std::string& msg )
         *      - doBuildPlayerAudio( int time, int unum )
         *      - doBuildPlayerAudio( int time )
         *
         */
        virtual
        void
        doBuildOppSide() {}

        /** This function is called after parsing a token representing opponent's team's side.
         *
         *  Override this function in your subclass to handle opponent team sides.
         *
         *  \pre A version 8 or 9 player audio message is being parsed.
         *
         *  \pre A token representing the opponent team has been parsed.
         *
         *  \sa 
         *      - doBuildPlayerAudio( int time, double dir, const std::string& msg )
         *      - doBuildPlayerAudio( int time, double dir, int unum,
         *                            const std::string& msg )
         *      - doBuildPlayerAudio( int time, int unum )
         *      - doBuildPlayerAudio( int time )
         *
         */
        virtual
        void
        doBuildOurSide() {}

        /** This function is called after parsing a token representing the neither side.
         *
         *  Override this function in your subclass to handle neutral sides.
         *
         *  \pre A offline coach audio message is being parsed.
         *
         *  \pre A token representing the offline coach been parsed.
         *
         *  \sa
         *      - doBuildCoachAudio( int time, const std::string& msg )
         *
         */
        virtual
        void
        doBuildNeutralSide() {}
        //@}

        //! \name Sense Body Parsing Functions
        //@{
        /** This function is called after parsing a sense body message.
         *
         *  Override this function in your subclass to handle sense body
         *  messages.  You will also need to overwrite the functions for parsing the
         *  subsections of the sense body message if you want to be able to use the data 
         *  from those subsections.
         *
         *  \note This function deprecates doBuildSenseBody( int time, 
         *                                                   int stamina,
         *                                                   int effort,
         *                                                   double speed_mag,
         *                                                   double speed_head,
         *                                                   double head_angle,
         *                                                   int kick_count,
         *                                                   int dash_count,
         *                                                   int turn_count,
         *                                                   int say_count,
         *                                                   int turn_neck_count,
         *                                                   int catch_count,
         *                                                   int move_count,
         *                                                   int chg_view_count ).
         *        If this function is not overridden, it will call the deprecated version.
         *
         *  \param time            The time (in server cycles) that the
         *                         message was sent by the simulator.
         *  \param stamina         The client's stamina.
         *  \param effort          The client's effort.
         *  \param speed_mag       The magnitude of the client's velocity.
         *  \param speed_head      The heading of the client's velocity, relative to it's body direction.
         *  \param head_angle      The client's head angle relative to it's body direction.
         *  \param kick_count      The number of kick commands the client has executed.
         *  \param dash_count      The number of dash commands the client has executed.
         *  \param turn_count      The number of turn commands the client has executed.
         *  \param say_count       The number of say commands the client has executed.
         *  \param turn_neck_count The number of turn neck commands the client has executed.
         *  \param catch_count     The number of catch commands the client has executed.
         *  \param move_count      The number of move commands the client has executed.
         *  \param chg_view_count  The number of change view commands the client has executed.
         *
         *  \pre The client is a field player.
         *
         *  The following is an example (with line breaks added for clarity) of a version 8 or 9 sense body message:
         *  \verbatim
            (sense_body 2127 (view_mode high normal)
                             (stamina 2694.5 0.93)
                             (speed 0.2 -3.56)
                             (head_angle 12.7)
                             (kick 128)
                             (dash 897)
                             (turn 1256)
                             (say 54)
                             (turn_neck 563)
                             (catch 0)
                             (move 2)
                             (change_view 89)
                             (arm (movable 0) (expires 0) (target 0 0) (count 0))
                             (focus (target none) (count 0))
                             (tackle (expires 0) (count 0)))\endverbatim
         *
         *  The following is the same example (with line breaks added for clarity) in the version 7 protocol:
         *  \verbatim
            (sense_body 2127 (view_mode high normal)
                             (stamina 2694.5 0.93)
                             (speed 0.2 -3.56)
                             (head_angle 12.7)
                             (kick 128)
                             (dash 897)
                             (turn 1256)
                             (say 54)
                             (turn_neck 563)
                             (catch 0)
                             (move 2)
                             (change_view 89)\endverbatim
         *
         *  In these examples the:
         *      - \e time             parameter would be 2127
         *      - \e stamina          parameter would be 2694.5.
         *      - \e effort           parameter would be 0.93.
         *      - \e speed_mag        parameter would be 0.2.
         *      - \e speed_head       parameter would be -3.56.
         *      - \e head_angle       parameter would be 12.7.
         *      - \e kick_count       parameter would be 128.
         *      - \e dash_count       parameter would be 897.
         *      - \e turn_count       parameter would be 1256.
         *      - \e say_count        parameter would be 54.
         *      - \e turn_neck_count  parameter would be 563.
         *      - \e catch_count      parameter would be 0.
         *      - \e move_count       parameter would be 2.
         *      - \e chg_view_count   parameter would be 89.
         *
         *  \sa
         *      - doBuildArmState( int movable_in,
         *                         int expires_in,
         *                         double target_x,
         *                         double target_y,
         *                         int count )
         *      - doBuildFocusState( int count )
         *      - doBuildFocusState( int unum, int count )
         *      - doBuildTackleState( int expires_in, int count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count ).
         */
        virtual
        void
        doBuildSenseBody( int time, 
                          double stamina,
                          double effort,
                          double speed_mag,
                          double speed_head,
                          double head_angle,
                          int kick_count,
                          int dash_count,
                          int turn_count,
                          int say_count,
                          int turn_neck_count,
                          int catch_count,
                          int move_count,
                          int chg_view_count ) 
        {
            doBuildSenseBody( time, 
                              (int)stamina, (int)effort,
                              speed_mag, speed_head, head_angle,
                              kick_count, dash_count, turn_count, say_count,
                              turn_neck_count, catch_count,
                              move_count, chg_view_count );
        }


        /** This function is called after parsing a sense body message, if 
         *  doBuildSenseBody( int time, 
         *                    double stamina,
         *                    double effort,
         *                    double speed_mag,
         *                    double speed_head,
         *                    double head_angle,
         *                    int kick_count,
         *                    int dash_count,
         *                    int turn_count,
         *                    int say_count,
         *                    int turn_neck_count,
         *                    int catch_count,
         *                    int move_count,
         *                    int chg_view_count ) has not been overridden.
         *
         *  Override this function in your subclass to handle sense body
         *  messages.  You will also need to overwrite the functions for parsing the
         *  subsections of the sense body message if you want to be able to use the data 
         *  from those subsections.
         *
         *  \Warning This function is deprecated.  Override doBuildSenseBody( int time, 
         *                                                                   double stamina,
         *                                                                   double effort,
         *                                                                   double speed_mag,
         *                                                                   double speed_head,
         *                                                                   double head_angle,
         *                                                                   int kick_count,
         *                                                                   int dash_count,
         *                                                                   int turn_count,
         *                                                                   int say_count,
         *                                                                   int turn_neck_count,
         *                                                                   int catch_count,
         *                                                                   int move_count,
         *                                                                   int chg_view_count )
         *           instead.
         *
         *  \param time            The time (in server cycles) that the
         *                         message was sent by the simulator.
         *  \param stamina         The client's stamina.
         *  \param effort          The client's effort.
         *  \param speed_mag       The magnitude of the client's velocity.
         *  \param speed_head      The heading of the client's velocity, relative to it's body direction.
         *  \param head_angle      The client's head angle relative to it's body direction.
         *  \param kick_count      The number of kick commands the client has executed.
         *  \param dash_count      The number of dash commands the client has executed.
         *  \param turn_count      The number of turn commands the client has executed.
         *  \param say_count       The number of say commands the client has executed.
         *  \param turn_neck_count The number of turn neck commands the client has executed.
         *  \param catch_count     The number of catch commands the client has executed.
         *  \param move_count      The number of move commands the client has executed.
         *  \param chg_view_count  The number of change view commands the client has executed.
         *
         *  \pre The client is a field player.
         *
         *  The following is an example (with line breaks added for clarity) of a version 8 or 9 sense body message:
         *  \verbatim
            (sense_body 2127 (view_mode high normal)
                             (stamina 2694.5 0.93)
                             (speed 0.2 -3.56)
                             (head_angle 12.7)
                             (kick 128)
                             (dash 897)
                             (turn 1256)
                             (say 54)
                             (turn_neck 563)
                             (catch 0)
                             (move 2)
                             (change_view 89)
                             (arm (movable 0) (expires 0) (target 0 0) (count 0))
                             (focus (target none) (count 0))
                             (tackle (expires 0) (count 0)))\endverbatim
         *
         *  The following is the same example (with line breaks added for clarity) in the version 7 protocol:
         *  \verbatim
            (sense_body 2127 (view_mode high normal)
                             (stamina 2694.5 0.93)
                             (speed 0.2 -3.56)
                             (head_angle 12.7)
                             (kick 128)
                             (dash 897)
                             (turn 1256)
                             (say 54)
                             (turn_neck 563)
                             (catch 0)
                             (move 2)
                             (change_view 89)\endverbatim
         *
         *  In these examples the:
         *      - \e time             parameter would be 2127.
         *      - \e stamina          parameter would be 2695.
         *      - \e effort           parameter would be 1.
         *      - \e speed_mag        parameter would be 0.2.
         *      - \e speed_head       parameter would be -3.56.
         *      - \e head_angle       parameter would be 12.7.
         *      - \e kick_count       parameter would be 128.
         *      - \e dash_count       parameter would be 897.
         *      - \e turn_count       parameter would be 1256.
         *      - \e say_count        parameter would be 54.
         *      - \e turn_neck_count  parameter would be 563.
         *      - \e catch_count      parameter would be 0.
         *      - \e move_count       parameter would be 2.
         *      - \e chg_view_count   parameter would be 89.
         *
         *  \sa
         *      - doBuildArmState( int movable_in,
         *                         int expires_in,
         *                         double target_x,
         *                         double target_y,
         *                         int count )
         *      - doBuildFocusState( int count )
         *      - doBuildFocusState( int unum, int count )
         *      - doBuildTackleState( int expires_in, int count )
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count ).
         */
        virtual
        void
        doBuildSenseBody( int time, 
                          int stamina,
                          int effort,
                          double speed_mag,
                          double speed_head,
                          double head_angle,
                          int kick_count,
                          int dash_count,
                          int turn_count,
                          int say_count,
                          int turn_neck_count,
                          int catch_count,
                          int move_count,
                          int chg_view_count ) 
        {}


        /** This function is called after parsing an arm
         *  section of a version 8 or 9 sense body message or
         *  a version 8 or 9 full state message.
         *
         *  Override this function in your subclass to handle
         *  the arm section of version 8 or 9 sense body messages
         *  and version 8 or 9 full state messages.
         *
         *  \param movable_in The number of cycles until the client or player can change where it's pointing to.
         *  \param expires_in The number of cycles until the client or player will automatically stop pointing.
         *  \param target_x   The x co-ordinate of the point the client or player is pointing to.
         *  \param target_y   The y co-ordinate of the point the client or player is pointing to.
         *  \param count      The number of times the client or player has excuted a point to command.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either:
                    - The arm section of a sense body message has been parsed or
                    - Full state message sending is turned on in the simulator for
                      the team this client is in and the arm section of a full state
                      message has just be parsed.
         *  
         *  The following is an example of the arm section of either a version 8 or 9 sense
         *  body message or a version 8 or 9 full state message:
         *  \verbatim
            (arm (movable 3) (expires 13) (target -10 36) (count 15))\endverbatim
         *
         *  In this examples the:
         *      - \e movable_in parameter would be 3.
         *      - \e expires_in parameter would be 13.
         *      - \e target_x   parameter would be -10.
         *      - \e target_y   parameter would be 36.
         *      - \e count      parameter would be 15.
         *
         *  \sa 
         *      - doBuildFullState( int time )
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *
         */
        virtual
        void
        doBuildArmState( int movable_in,
                         int expires_in,
                         double target_x,
                         double target_y,
                         int count ) {}
      
        /** This function is called after parsing an focus state
         *  section of a version 8 or 9 sense body, where the client is
         *  \b not focused on any player.
         *
         *  Override this function in your subclass to handle the non-focused
         *  focus state section of version 8 or 9 sense body messages.
         *
         *  \param count      The number of times the client has excuted an attention to command.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre A version 8 or 9 sense body message is being parsed.
         *
         *  \pre A non-focused focus state section of a version 8 or 9 sense body message has just been parsed.
         *
         *  The following is an example of the non-focused focus section of a version
         *  8 or 9 sense body message:
         *  \verbatim
            (focus (target none) (count 18))\endverbatim
         *
         *  In this examples the:
         *      - \e count      parameter would be 18.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFocusState( int unum, int count )
         */
        virtual
        void
        doBuildFocusState( int count ) {}

        /** This function is called after parsing a focus state
         *  section of a version 8 or 9 sense body, where the client \b is
         *  focused on a player.
         *
         *  Override this function in your subclass to handle the focused
         *  focus state section of version 8 or 9 sense body messages.
         *
         *  \param unum       The uniform number of the player the client is focued on.
         *  \param count      The number of times the client has excuted an attention to command.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre A version 8 or 9 sense body message is being parsed.
         *
         *  \pre A focused focus state section of a version 8 or 9 sense body message has just been parsed.
         *
         *  The following is an example of the focused focus section of a version
         *  8 or 9 sense body message:
         *  \verbatim
            (focus (target l 9) (count 18))\endverbatim
         *
         *  In this examples the:
         *      - \e unum       parameter would be 9.
         *      - \e count      parameter would be 18.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFocusState( int count )
         *
         */
        virtual
        void
        doBuildFocusState( int unum, int count ) {}


        /** This function is called after parsing a tackle state
         *  section of a version 8 or 9 sense body.
         *
         *  Override this function in your subclass to handle the 
         *  tackle state section of version 8 or 9 sense body messages.
         *
         *  \param expires_in The number of cycles till client finishes tackling.
         *  \param count      The number of times the client has excuted an tackle command.
         *
         *  \pre The client is using protocol version 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre A version 8 or 9 sense body message is being parsed.
         *
         *  \pre A tackle state section of a version 8 or 9 sense body message has just been parsed.
         *
         *  The following is an example of the focused focus section of a version
         *  8 or 9 sense body message:
         *  \verbatim
            (tackle (expires 3) (count 18))\endverbatim
         *
         *  In this examples the:
         *      - \e expires_in parameter would be 3.
         *      - \e count      parameter would be 18.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         */
        virtual
        void
        doBuildTackleState( int expires_in, int count ) {}
        //@}



        //! \name View Mode Parsing Functions
        //@{
        /** This function is called after parsing a view mode
         *  section of a sense body of full state message.
         *
         *  Override this function in your subclass to handle the 
         *  view mode section of sense body or full state messages.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either: 
         *          - a sense body message is being parsed.
         *          - a full state message is being parsed.
         *
         *  \pre The view mode section of a sense body or full state message has just been parsed.
         *
         *  The following is an example of the view_mode section of a
         *  sense body message:
         *  \verbatim
            (view_mode high normal)\endverbatim
         *
         *  The following is the same example as part of a 
         *  full state message:
         *  \verbatim
            (vmode high normal)\endverbatim
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFullState( int time )
         *      - doBuildViewQualityHigh()
         *      - doBuildViewQualityLow()
         *      - doBuildViewWidthNarrow()
         *      - doBuildViewWidthNormal()
         *      - doBuildViewWidthWide()
         */
        virtual
        void
        doBuildViewMode() {}


        /** This function is called after parsing a \c high
         *  view quality in the view mode section of a 
         *  sense body or full state message.
         *
         *  Override this function in your subclass to handle \c high
         *  view qualities in the view mode section of a
         *  sense body or full state message.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either: 
         *          - a sense body message is being parsed.
         *          - a full state message is being parsed.
         *
         *  \pre The view mode section of a sense body or full state is being parsed.
         *
         *  \pre The \c high view quality has just been parsed.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFullState( int time )
         *      - doBuildViewMode()
         *      - doBuildViewQualityLow()
         */
        virtual
        void
        doBuildViewQualityHigh() {}


        /** This function is called after parsing a \c low
         *  view quality in the view mode section of a 
         *  sense body or full state message.
         *
         *  Override this function in your subclass to handle \c low
         *  view qualities in the view mode section of a
         *  sense body or full state message.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either: 
         *          - a sense body message is being parsed.
         *          - a full state message is being parsed.
         *
         *  \pre The view mode section of a sense body or full state is being parsed.
         *
         *  \pre The \c low view quality has just been parsed.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFullState( int time )
         *      - doBuildViewMode()
         *      - doBuildViewQualityHigh()
         */
        virtual
        void
        doBuildViewQualityLow() {}


        /** This function is called after parsing a \c narrow
         *  view width in the view mode section of a 
         *  sense body or full state message.
         *
         *  Override this function in your subclass to handle \c narrow
         *  view qualities in the view mode section of a
         *  sense body or full state message.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either: 
         *          - a sense body message is being parsed.
         *          - a full state message is being parsed.
         *
         *  \pre The view mode section of a sense body or full state is being parsed.
         *
         *  \pre The \c narrow view width has just been parsed.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFullState( int time )
         *      - doBuildViewMode()
         *      - doBuildViewWidthNormal()
         *      - doBuildViewWidthWide()
         */
        virtual
        void
        doBuildViewWidthNarrow() {}


        /** This function is called after parsing a \c normal
         *  view width in the view mode section of a 
         *  sense body or full state message.
         *
         *  Override this function in your subclass to handle \c normal
         *  view qualities in the view mode section of a
         *  sense body or full state message.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either: 
         *          - a sense body message is being parsed.
         *          - a full state message is being parsed.
         *
         *  \pre The view mode section of a sense body or full state is being parsed.
         *
         *  \pre The \c normal view width has just been parsed.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFullState( int time )
         *      - doBuildViewMode()
         *      - doBuildViewWidthNarrow()
         *      - doBuildViewWidthWide()
         */
        virtual
        void
        doBuildViewWidthNormal() {}


        /** This function is called after parsing a \c wide
         *  view width in the view mode section of a 
         *  sense body or full state message.
         *
         *  Override this function in your subclass to handle \c wide
         *  view qualities in the view mode section of a
         *  sense body or full state message.
         *
         *  \pre The client is a field player.
         *
         *  \pre Either: 
         *          - a sense body message is being parsed.
         *          - a full state message is being parsed.
         *
         *  \pre The view mode section of a sense body or full state is being parsed.
         *
         *  \pre The \c wide view width has just been parsed.
         *
         *  \sa 
         *      - doBuildSenseBody( int time, 
         *                          double stamina,
         *                          double effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildSenseBody( int time, 
         *                          int stamina,
         *                          int effort,
         *                          double speed_mag,
         *                          double speed_head,
         *                          double head_angle,
         *                          int kick_count,
         *                          int dash_count,
         *                          int turn_count,
         *                          int say_count,
         *                          int turn_neck_count,
         *                          int catch_count,
         *                          int move_count,
         *                          int chg_view_count )
         *      - doBuildFullState( int time )
         *      - doBuildViewMode()
         *      - doBuildViewWidthNarrow()
         *      - doBuildViewWidthNormal()
         */
        virtual
        void
        doBuildViewWidthWide() {}
        //@}


        //! \name Parameter Parsing Functions
        //@{
        /** This ungainly function is called after parsing a version 7
         *  server parameter message.
         *
         *  Override this function in your subclass to handle version 7
         *  server parameter messages.
         *
         *  \param gwidth              The width of the goals.
         *  \param inertia_moment      The players inertia during a turn.
         *  \param psize               The size of the default player.
         *  \param pdecay              The velocity decay of the default player.
         *  \param prand               The maxmuim percent of random motion in player for each cycle.
         *  \param pweight             The weight of the player.
         *  \param pspeed_max          The maximum velocity of the default player.
         *  \param paccel_max          The maximum acceleration of the players.
         *  \param stamina_max         The maximum stamina of the players.
         *  \param stamina_inc         The maximum stamina increment of the default player.
         *  \param recover_init        The initial recovery value for players.
         *  \param recover_dthr        The recovery decrement threshold for players.
         *  \param recover_min         The minimum recovery a player can have.
         *  \param recover_dec         The player recovery decrement.
         *  \param effort_init         The intial and maximum effort for the default player.
         *  \param effort_dthr         The effort decrement threshold for players.
         *  \param effort_min          The minimum effort a default player can have.
         *  \param effort_dec          The player effort decrement.
         *  \param effort_ithr         The effort increment threshold for players.
         *  \param effort_inc          The player effort increment.
         *  \param kick_rand           The percent of random noise added to kicks.
         *  \param team_actuator_noise The flag for whether team specific actuator noise is used.
         *  \param prand_factor_l      The prand multiplication factor for the left team.
         *  \param prand_factor_r      The prand multiplication factor for the right team.
         *  \param kick_rand_factor_l  The kick_rand multiplication factor for the left team.
         *  \param kick_rand_factor_r  The kick_rand multiplication factor for the right team.
         *  \param bsize               The size of the ball.
         *  \param bdecay              The velocity decay of the ball.
         *  \param brand               The maxmuim percent of random motion in the ball for each cycle.
         *  \param bweight             The weight of the ball.
         *  \param bspeed_max          The maximum speed of the ball.
         *  \param baccel_max          The maximum velocity of the ball.
         *  \param dprate              The dash power rate for the default player.
         *  \param kprate              The kick power rate for players.
         *  \param kmargin             The kickable margin for the default player.
         *  \param ctlradius           Obsolete and no longer used.
         *  \param ctlradius_width     Obsolete and no longer used.
         *  \param maxp                The maximum power that can be used by a player in one command.
         *  \param minp                The minimum power that can be used by a player in one command.
         *  \param maxm                The maximum moment that can be used by a player in one command.
         *  \param minm                The minimum moment that can be used by a player in one command.
         *  \param maxnm               The maximum moment that can be used by a player in one turn neck command.
         *  \param minnm               The minimum moment that can be used by a player in one command.
         *  \param maxn                The maximum a player can turn it's neck.
         *  \param minn                The minimum a player can turn it's neck.
         *  \param visangle            The view width of a player in \c normal view mode.
         *  \param visdist             The distance under which objects can be "sensed" by players. 
         *  \param windir              The direction of the wind.
         *  \param winforce            The force of the wind.
         *  \param winang              Obsolete and no longer used.
         *  \param winrand             The randomisation factor in the wind.
         *  \param kickable_area       kmargin + bsize + psize.
         *  \param catch_area_l        The goalie catchable area length.
         *  \param catch_area_w        The goalie catchable area width.
         *  \param catch_prob          The goalie cacth probability.
         *  \param goalie_max_moves    The maximum number of move a goalie can make after a catch and before releasing the ball.
         *  \param ckmargin            The distance from the side and goal line that the ball is placed when a corner kick is awarded.
         *  \param offside_area        The distance to the ball that under which an offside player is considered to be in active offside.
         *  \param win_no              The flag for no wind.
         *  \param win_random          The flag for random wind.
         *  \param say_cnt_max         The maximum number of say messages an online coach can execute in one game. 
         *  \param SayCoachMsgSize     The maximum size of a online coach non-CLang or freeform say message.
         *  \param clang_win_size      The size of each CLang send window.
         *  \param clang_define_win    The number of CLang define messages allowed per CLang window.
         *  \param clang_meta_win      The number of CLang meta messages allowed per Clang window.
         *  \param clang_advice_win    The number of CLang advice messages allowed per Clang window.
         *  \param clang_info_win      The number of CLang info messages allowed per Clang window.
         *  \param clang_mess_delay    The number of cycles the server will delay CLang messages before sending them to the players.
         *  \param clang_mess_per_cycle The maximum number of CLang messages that a coach can be send per cycle.
         *  \param half_time           The length of a game half (in seconds).
         *  \param sim_st              The length of a simulator step (in ms).
         *  \param send_st             The length between player visuals when the player is in the \c normal \c high view mode(in ms).
         *  \param recv_st             The length between the simulator checking for client messages.
         *  \param sb_step             The length between player sense body messages.
         *  \param lcm_st              The lowest common multiple of the above steps.
         *  \param SayMsgSize          The maximum size of a player say message.
         *  \param hear_max            The maximum hearing capacity of a player.
         *  \param hear_inc            The player hear capacity increament per cycle.
         *  \param hear_decay          The amount the player hear capacity is reduced by every time they hear a player audio message.
         *  \param cban_cycle          The number of cycles for which a goalie cannot catch after executing a catch.
         *  \param slow_down_factor    The amount the sim_st, send_st, recv_st and sb_step have been mulitplied by.
         *  \param useoffside          The flag for whether offsides are used.
         *  \param kickoffoffside      The flag for whether players are moved to a onside position before kickoffs.
         *  \param offside_kick_margin The radius of the area that opponent players are moved from the ball when an offside has been awarded.
         *  \param audio_dist          The maximum distance from which players can hear each other.
         *  \param dist_qstep          The quatization factor for movable objects (players and the ball).
         *  \param land_qstep          The quatization factor for fixed objects (flags).
         *  \param dir_qstep           The quatization factor for direction change in a objects velocity.
         *  \param dist_qstep_l        The dist_qstep for the left team. 
         *  \param dist_qstep_r        The dist_qstep for the right team.  
         *  \param land_qstep_l        The land_qstep for the left team.
         *  \param land_qstep_r        The land_qstep for the right team.
         *  \param dir_qstep_l         The dir_qstep for the left team.
         *  \param dir_qstep_r         The dir_qstep for the right team.
         *  \param CoachMode           The flag for if the offline coach is allowed with the referee disabled. 
         *  \param CwRMode             The flag for if the offline coach is allowed with the referee enabled.
         *  \param old_hear            Obsolete and no longer used.
         *  \param sv_st               The time between sending coach visuals.
         *  \param start_goal_l        The number of goals that the left team has started on.
         *  \param start_goal_r        The number of goals that the right teams has started on.
         *  \param fullstate_l         The flag for whether full state messages are sent to the left players.
         *  \param fullstate_r         The flag for whether full state messages are sent to the right players.
         *  \param drop_time           The amount of time after a freekick in which the play mode is
         *                             changed to drop ball if the kick has not been taken.                 
         *
         *  \pre The client is using protocol version 7
         *
         *  \pre A server param message has just been parsed.
         *
         *  The following is an example (with line breaks added for what every clarity we can get) of a version 7 server param message:
         *  \verbatim
            (server_param 14.02 5 0.3 0.4 0.1 60 1 1 4000 45 0 0.3 0.5 0.002
                          0 0.3 0.6 0.005 0.6 0.01 0 0 1 1 1 1 0.085 0.94 0.05
                          0.2 2.7 2.7 0.006 0.027 0.7 2 1.7 100 -100 180 -180
                          180 -180 90 -90 90 3 0 0 0 0 1.085 2 1 1 2 1 5 0 0
                          128 128 300 1 1 1 1 50 1 3000 100 150 10 100 300 512
                          2 1 2 5 1 1 1 9.15 50 0.1 0.01 -1 -1 -1 -1 -1 -1 -1
                          0 0 0 100 0 0 0 0 200)\endverbatim
         *
         *  \sa
         *      - doBuildServerParam()
         */
        virtual
        void
        doBuildServerParam ( double gwidth, 
                             double inertia_moment, 
                             double psize, 
                             double pdecay, 
                             double prand,
                             double pweight, 
                             double pspeed_max, 
                             double paccel_max, 
                             double stamina_max, 
                             double stamina_inc,
                             double recover_init, 
                             double recover_dthr, 
                             double recover_min, 
                             double recover_dec, 
                             double effort_init,
                             double effort_dthr, 
                             double effort_min, 
                             double effort_dec, 
                             double effort_ithr, 
                             double effort_inc,
                             double kick_rand, 
                             int team_actuator_noise, 
                             double prand_factor_l, 
                             double prand_factor_r, 
                             double kick_rand_factor_l,
                             double kick_rand_factor_r, 
                             double bsize, 
                             double bdecay, 
                             double brand, 
                             double bweight,
                             double bspeed_max, 
                             double baccel_max, 
                             double dprate, 
                             double kprate, 
                             double kmargin,
                             double ctlradius, 
                             double ctlradius_width, 
                             double maxp, 
                             double minp, 
                             double maxm,
                             double minm, 
                             double maxnm, 
                             double minnm, 
                             double maxn, 
                             double minn,
                             double visangle, 
                             double visdist, 
                             double windir, 
                             double winforce, 
                             double winang,
                             double winrand, 
                             double kickable_area, 
                             double catch_area_l, 
                             double catch_area_w, 
                             double catch_prob,
                             int goalie_max_moves, 
                             double ckmargin, 
                             double offside_area, 
                             int win_no, 
                             int win_random,
                             int say_cnt_max, 
                             int SayCoachMsgSize, 
                             int clang_win_size, 
                             int clang_define_win, 
                             int clang_meta_win,
                             int clang_advice_win, 
                             int clang_info_win, 
                             int clang_mess_delay, 
                             int clang_mess_per_cycle, 
                             int half_time,
                             int sim_st, 
                             int send_st, 
                             int recv_st, 
                             int sb_step, 
                             int lcm_st,
                             int SayMsgSize, 
                             int hear_max, 
                             int hear_inc, 
                             int hear_decay, 
                             int cban_cycle,
                             int slow_down_factor, 
                             int useoffside, 
                             int kickoffoffside, 
                             double offside_kick_margin, 
                             double audio_dist,
                             double dist_qstep, 
                             double land_qstep, 
                             double dir_qstep, 
                             double dist_qstep_l, 
                             double dist_qstep_r,
                             double land_qstep_l, 
                             double land_qstep_r, 
                             double dir_qstep_l, 
                             double dir_qstep_r, 
                             int CoachMode,
                             int CwRMode, 
                             int old_hear, 
                             int sv_st, 
                             int start_goal_l, 
                             int start_goal_r,
                             int fullstate_l, 
                             int fullstate_r, 
                             int drop_time ) {}

        /** This function is called after parsing a version 8 or 9
         *  server parameter message.
         *
         *  Override this function in your subclass to handle version 8 or 9
         *  server parameter messages.
         *
         *  \pre The client is using protocol version 8 or 9
         *
         *  \pre A version 8 or 9 server param message has just been parsed.
         *
         *  The following is an example (with line breaks added for clarity) of a version 8 or 9 server param message:
         *  \verbatim
            (server_param (catch_ban_cycle 5)(clang_advice_win 1)(clang_define_win 1)
                          (clang_del_win 1)(clang_info_win 1)(clang_mess_delay 50)
                          (clang_mess_per_cycle 1)(clang_meta_win 1)(clang_rule_win 1)
                          (clang_win_size 300)(coach_port 6001)(drop_ball_time 200)
                          (freeform_send_period 20)(freeform_wait_period 600)(game_log_compression 0)
                          (game_log_version 3)(goalie_max_moves 2)(half_time 3000)(hear_decay 1)
                          (hear_inc 1)(hear_max 1)(max_goal_kicks 3)(olcoach_port 6002)(point_to_ban 5)
                          (point_to_duration 20)(port 6000)(recv_step 10)(say_coach_cnt_max 128)
                          (say_coach_msg_size 128)(say_msg_size 10)(send_step 150)(send_vi_step 100)
                          (sense_body_step 100)(simulator_step 100)(slow_down_factor 1)(start_goal_l 0)
                          (start_goal_r 0)(synch_micro_sleep 1)(synch_offset 60)(tackle_cycles 10)
                          (text_log_compression 0)(game_log_dir "/home/thoward/data")
                          (game_log_fixed_name "rcssserver")(landmark_file "~/.rcssserver-landmark.xml")
                          (log_date_format "%Y%m%d%H%M-")(text_log_dir "/home/thoward/data")
                          (text_log_fixed_name "rcssserver")(coach 0)(coach_w_referee 1)
                          (old_coach_hear 0)(wind_none 0)(wind_random 0)(back_passes 1)
                          (forbid_kick_off_offside 1)(free_kick_faults 1)(fullstate_l 1)(fullstate_r 1)
                          (game_log_dated 1)(game_log_fixed 1)(game_logging 1)(log_times 0)(profile 1)
                          (proper_goal_kicks 0)(record_messages 0)(send_comms 0)(synch_mode 0)
                          (team_actuator_noise 0)(text_log_dated 1)(text_log_fixed 1)(text_logging 1)
                          (use_offside 1)(verbose 0)(audio_cut_dist 50)(ball_accel_max 2.7)
                          (ball_decay 0.94)(ball_rand 0.05)(ball_size 0.085)(ball_speed_max 2.7)
                          (ball_weight 0.2)(catch_probability 1)(catchable_area_l 2)(catchable_area_w 1)
                          (ckick_margin 1)(control_radius 2)(dash_power_rate 0.006)(effort_dec 0.005)
                          (effort_dec_thr 0.3)(effort_inc 0.01)(effort_inc_thr 0.6)(effort_init 0)
                          (effort_min 0.6)(goal_width 14.02)(inertia_moment 5)(kick_power_rate 0.027)
                          (kick_rand 0)(kick_rand_factor_l 1)(kick_rand_factor_r 1)(kickable_margin 0.7)
                          (maxmoment 180)(maxneckang 90)(maxneckmoment 180)(maxpower 100)
                          (minmoment -180)(minneckang -90)(minneckmoment -180)(minpower -100)
                          (offside_active_area_size 2.5)(offside_kick_margin 9.15)(player_accel_max 1)
                          (player_decay 0.4)(player_rand 0.1)(player_size 0.3)(player_speed_max 1)
                          (player_weight 60)(prand_factor_l 1)(prand_factor_r 1)(quantize_step 0.1)
                          (quantize_step_l 0.01)(recover_dec 0.002)(recover_dec_thr 0.3)
                          (recover_min 0.5)(slowness_on_top_for_left_team 1)
                          (slowness_on_top_for_right_team 1)(stamina_inc_max 45)(stamina_max 4000)
                          (stopped_ball_vel 0.01)(tackle_back_dist 0.5)(tackle_dist 2.5)
                          (tackle_exponent 6)(tackle_power_rate 0.027)(tackle_width 1.25)
                          (visible_angle 90)(visible_distance 3)(wind_ang 0)(wind_dir 0)(wind_force 0)
                          (wind_rand 0))\endverbatim
         *
         *  \sa
         *      - doBuildServerParam( double gwidth, 
         *                   double inertia_moment, 
         *                   double psize, 
         *                   double pdecay, 
         *                   double prand,
         *                   double pweight, 
         *                   double pspeed_max, 
         *                   double paccel_max, 
         *                   double stamina_max, 
         *                   double stamina_inc,
         *                   double recover_init, 
         *                   double recover_dthr, 
         *                   double recover_min, 
         *                   double recover_dec, 
         *                   double effort_init,
         *                   double effort_dthr, 
         *                   double effort_min, 
         *                   double effort_dec, 
         *                   double effort_ithr, 
         *                   double effort_inc,
         *                   double kick_rand, 
         *                   int team_actuator_noise, 
         *                   double prand_factor_l, 
         *                   double prand_factor_r, 
         *                   double kick_rand_factor_l,
         *                   double kick_rand_factor_r, 
         *                   double bsize, 
         *                   double bdecay, 
         *                   double brand, 
         *                   double bweight,
         *                   double bspeed_max, 
         *                   double baccel_max, 
         *                   double dprate, 
         *                   double kprate, 
         *                   double kmargin,
         *                   double ctlradius, 
         *                   double ctlradius_width, 
         *                   double maxp, 
         *                   double minp, 
         *                   double maxm,
         *                   double minm, 
         *                   double maxnm, 
         *                   double minnm, 
         *                   double maxn, 
         *                   double minn,
         *                   double visangle, 
         *                   double visdist, 
         *                   double windir, 
         *                   double winforce, 
         *                   double winang,
         *                   double winrand,
         *                   double kickable_area, 
         *                   double catch_area_l, 
         *                   double catch_area_w, 
         *                   double catch_prob,
         *                   int goalie_max_moves, 
         *                   double ckmargin, 
         *                   double offside_area, 
         *                   int win_no, 
         *                   int win_random,
         *                   int say_cnt_max, 
         *                   int SayCoachMsgSize, 
         *                   int clang_win_size, 
         *                   int clang_define_win, 
         *                   int clang_meta_win,
         *                   int clang_advice_win, 
         *                   int clang_info_win, 
         *                   int clang_mess_delay, 
         *                   int clang_mess_per_cycle, 
         *                   int half_time,
         *                   int sim_st, 
         *                   int send_st, 
         *                   int recv_st, 
         *                   int sb_step, 
         *                   int lcm_st,
         *                   int SayMsgSize, 
         *                   int hear_max, 
         *                   int hear_inc, 
         *                   int hear_decay, 
         *                   int cban_cycle,
         *                   int slow_down_factor, 
         *                   int useoffside, 
         *                   int kickoffoffside, 
         *                   double offside_kick_margin, 
         *                   double audio_dist,
         *                   double dist_qstep, 
         *                   double land_qstep, 
         *                   double dir_qstep, 
         *                   double dist_qstep_l, 
         *                   double dist_qstep_r,
         *                   double land_qstep_l, 
         *                   double land_qstep_r, 
         *                   double dir_qstep_l, 
         *                   double dir_qstep_r, 
         *                   int CoachMode,
         *                   int CwRMode, 
         *                   int old_hear, 
         *                   int sv_st, 
         *                   int start_goal_l, 
         *                   int start_goal_r,
         *                   int fullstate_l, 
         *                   int fullstate_r, 
         *                   int drop_time )
         *      - doBuildParam( const std::string& name,
         *                      int value )
         *      - doBuildParam( const std::string& name,
         *                      double value )
         *      - doBuildParam( const std::string& name,
         *                      const std::string& value )
         *
         */
        virtual
        void
        doBuildServerParam() {}


        /** This function is called after parsing a param section of a
         *  version 8 or 9 server parameter, player parameter or player type
         *  message, where the type of the
         *  parameter is an integer.
         *
         *  Override this function in your subclass to handle integer
         *  parameters of version 8 or 9 a server parameter, player parameter
         *  or player type message.
         *
         *  \param name The name of the parameter.
         *  \param value The vaule of the parameter.
         *
         *  \pre The client is using protocol version 8 or 9
         *
         *  \pre Either:
         *           - A version 8 or 9 server parameter message is being parsed.
         *           - A version 8 or 9 player parameter message is being parsed.
         *           - A version 8 or 9 player type message is being parsed.
         *  
         *  \pre A integer parameter section of a version 8 or 9 server parameter, player parameter
         *       or player type message has just been parsed.
         *
         *  The following is an example of a integer parameter section of a
         *  version 8 or 9 server param message:
         *  \verbatim
            (catch_ban_cycle 5)\endverbatim
         *
         *  \sa
         *      - doBuildServerParam()
         *      - doBuildPlayerParam()
         *      - doBuildPlayerType()
         *      - doBuildParam( const std::string& name,
         *                      double value )
         *      - doBuildParam( const std::string& name,
         *                      const std::string& value )
         */
        virtual
        void
        doBuildParam( const std::string& name,
                      int value ) {}


        /** This function is called after parsing a param section of a
         *  version 8 or 9 server parameter, player parameter or player type
         *  message, where the type of the
         *  parameter is an real.
         *
         *  Override this function in your subclass to handle real
         *  parameters of version 8 or 9 a server parameter, player parameter
         *  or player type message.
         *
         *  \param name The name of the parameter.
         *  \param value The vaule of the parameter.
         *
         *  \pre The client is using protocol version 8 or 9
         *
         *  \pre Either:
         *           - A version 8 or 9 server parameter message is being parsed.
         *           - A version 8 or 9 player parameter message is being parsed.
         *           - A version 8 or 9 player type message is being parsed.
         *  
         *  \pre A real parameter section of a version 8 or 9 server parameter, player parameter
         *       or player type message has just been parsed.
         *
         *  The following is an example of a real parameter section of a
         *  version 8 or 9 server param message:
         *  \verbatim
            (stopped_ball_vel 0.01)\endverbatim
         *
         *  \sa
         *      - doBuildServerParam()
         *      - doBuildPlayerParam()
         *      - doBuildPlayerType()
         *      - doBuildParam( const std::string& name,
         *                      int value )
         *      - doBuildParam( const std::string& name,
         *                      const std::string& value )
         */
        virtual
        void
        doBuildParam( const std::string& name,
                      double value ) {}
      

        /** This function is called after parsing a param section of a
         *  version 8 or 9 server parameter, player parameter or player type
         *  message, where the type of the
         *  parameter is an string.
         *
         *  Override this function in your subclass to handle string
         *  parameters of version 8 or 9 a server parameter, player parameter
         *  or player type message.
         *
         *  \param name The name of the parameter.
         *  \param value The vaule of the parameter.
         *
         *  \pre The client is using protocol version 8 or 9
         *
         *  \pre Either:
         *           - A version 8 or 9 server parameter message is being parsed.
         *           - A version 8 or 9 player parameter message is being parsed.
         *           - A version 8 or 9 player type message is being parsed.
         *  
         *  \pre A string parameter section of a version 8 or 9 server parameter, player parameter
         *       or player type message has just been parsed.
         *
         *  The following is an example of a string parameter section of a
         *  version 8 or 9 server param message:
         *  \verbatim
            (landmark_file "~/.rcssserver-landmark.xml")\endverbatim
         *
         *  \sa
         *      - doBuildServerParam()
         *      - doBuildPlayerParam()
         *      - doBuildPlayerType()
         *      - doBuildParam( const std::string& name,
         *                      int value )
         *      - doBuildParam( const std::string& name,
         *                      const std::string& value )
         */
        virtual
        void
        doBuildParam( const std::string& name,
                      const std::string& value ) {}


        /** This function is called after parsing a version 7
         *  player parameter message.
         *
         *  Override this function in your subclass to handle version 7
         *  player parameter messages.
         *
         *  \param player_types                 The total number of player types.
         *  \param subs_max                     The total number of substitutions allowed per game.
         *  \param pt_max                       The maximum number of non-default player types a
         *                                      team is allowed to have ant any one time.
         *  \param player_speed_max_delta_min   The minimum delta factor for the max player speed.
         *  \param player_speed_max_delta_max   The maximum delta factor for the max player speed.
         *  \param stamina_inc_max_delta_factor The stamina increment max delta factor.
         *  \param player_decay_delta_min       The minimum delta factor for the player decay.
         *  \param player_decay_delta_max       The maximum delta factor for the player decay.
         *  \param inertia_moment_delta_factor  The inertia moment delta factor.
         *  \param dash_power_rate_delta_min    The minimum delta factor for the dash power rate.
         *  \param dash_power_rate_delta_max    The maximum delta factor for the dash power rate.
         *  \param player_size_delta_factor     The player size delta factor.
         *  \param kickable_margin_delta_min    The minimum delta factor for the kickable margin.
         *  \param kickable_margin_delta_max    The maximum delta factor for the kickable margin.
         *  \param kick_rand_delta_factor       The kick rand delta factor.
         *  \param extra_stamina_delta_min      The minimum delta factor for the extra stamina.
         *  \param extra_stamina_delta_max      The maximum delta factor for the extra stamina.
         *  \param effort_max_delta_factor      The effort max delta factor.
         *  \param effort_min_delta_factor      The effort min delta factor.
         *
         *  \pre The client is using protocol version 7
         *
         *  \pre A version 7 player parameter meaage has just been parsed.
         *
         *  The following is an example (with line breaks
         *  added for clarity) of a version 7 player param message:
         *  \verbatim
            (player_param 7 3 3 0 0.2 -100 0 0.2 25 0 0.002
                          -100 0 0.2 0.5 0 100 -0.002 -0.002)\endverbatim
         *
         *  \sa
         *      - doBuildPlayerParam()
         *
         */
        virtual
        void
        doBuildPlayerParam( int player_types,
                            int subs_max,
                            int pt_max,
                            double player_speed_max_delta_min,
                            double player_speed_max_delta_max,
                            double stamina_inc_max_delta_factor,
                            double player_decay_delta_min,
                            double player_decay_delta_max,
                            double inertia_moment_delta_factor,
                            double dash_power_rate_delta_min,
                            double dash_power_rate_delta_max,
                            double player_size_delta_factor,
                            double kickable_margin_delta_min,
                            double kickable_margin_delta_max,
                            double kick_rand_delta_factor,
                            double extra_stamina_delta_min,
                            double extra_stamina_delta_max,
                            double effort_max_delta_factor,
                            double effort_min_delta_factor ) {}


        /** This function is called after parsing a version 8 or 9
         *  player parameter message.
         *
         *  Override this function in your subclass to handle version 8 or 9
         *  player parameter messages.
         *
         *  \pre The client is using protocol version 8 or 9
         *
         *  \pre A version 8 or 9 player param message has just been parsed.
         *
         *  The following is an example (with line breaks added for clarity) of a version 8 or 9 player param message:
         *  \verbatim
            (player_param (player_types 7)(pt_max 3)(random_seed -1)(subs_max 3)
                          (dash_power_rate_delta_max 0)(dash_power_rate_delta_min 0)
                          (effort_max_delta_factor -0.002)(effort_min_delta_factor -0.002)
                          (extra_stamina_delta_max 100)(extra_stamina_delta_min 0)
                          (inertia_moment_delta_factor 25)(kick_rand_delta_factor 0.5)
                          (kickable_margin_delta_max 0.2)(kickable_margin_delta_min 0)
                          (new_dash_power_rate_delta_max 0.002)
                          (new_dash_power_rate_delta_min 0)
                          (new_stamina_inc_max_delta_factor -10000)
                          (player_decay_delta_max 0.2)(player_decay_delta_min 0)
                          (player_size_delta_factor -100)(player_speed_max_delta_max 0.2)
                          (player_speed_max_delta_min 0)(stamina_inc_max_delta_factor 0))\endverbatim
         *
         *  \sa
         *      - doBuildPlayerParam( int player_types,
         *                  int subs_max,
         *                  int pt_max,
         *                  double player_speed_max_delta_min,
         *                  double player_speed_max_delta_max,
         *                  double stamina_inc_max_delta_factor,
         *                  double player_decay_delta_min,
         *                  double player_decay_delta_max,
         *                  double inertia_moment_delta_factor,
         *                  double dash_power_rate_delta_min,
         *                  double dash_power_rate_delta_max,
         *                  double player_size_delta_factor,
         *                  double kickable_margin_delta_min,
         *                  double kickable_margin_delta_max,
         *                  double kick_rand_delta_factor,
         *                  double extra_stamina_delta_min,
         *                  double extra_stamina_delta_max,
         *                  double effort_max_delta_factor,
         *                  double effort_min_delta_factor )
         *      - doBuildParam( const std::string& name,
         *                      int value )
         *      - doBuildParam( const std::string& name,
         *                      double value )
         *      - doBuildParam( const std::string& name,
         *                      const std::string& value )
         *
         */
        virtual
        void
        doBuildPlayerParam() {}


        /** This function is called after parsing a version 7
         *  player type message.
         *
         *  Override this function in your subclass to handle version 7
         *  player type messages.
         *
         *  \param id               The player type's ID
         *  \param player_speed_max The maximum speed of the player type.
         *  \param stamina_inc_max  The maximum stamina increment of the player type.
         *  \param player_decay     The velocity decay of the player type.
         *  \param inertia_moment   The player type's inertia in turns.
         *  \param dash_power_rate  The dash power rate of the player type.
         *  \param player_size      The size of the player type.
         *  \param kickable_margin  The kickable margin of the player type.
         *  \param kick_rand        The percent of random nosie in kicks by this player type.
         *  \param extra_stamina    The extra stamina for this player type.
         *  \param effort_max       The maximum effort for this player type.
         *  \param effort_min       The minimum effort for this player type.
         *
         *  \pre The client is using protocol version 7
         *
         *  \pre A version 7 player type meaage has just been parsed.
         *
         *  The following is an exampleof a version 7 player type message:
         *  \verbatim
            (player_type 0 1 45 0.4 5 0.006 0.3 0.7 0 0 1 0.6)\endverbatim
         *
         *  In this example the:
         *      - \e id               parameter would be 0.
         *      - \e player_speed_max parameter would be 1.
         *      - \e stamina_inc_max  parameter would be 45.
         *      - \e player_decay     parameter would be 0.4.
         *      - \e inertia_moment   parameter would be 5.
         *      - \e dash_power_rate  parameter would be 0.006.
         *      - \e player_size      parameter would be 0.3.
         *      - \e kickable_margin  parameter would be 0.7.
         *      - \e kick_rand        parameter would be 0.
         *      - \e extra_stamina    parameter would be 0.
         *      - \e effort_max       parameter would be 1.
         *      - \e effort_min       parameter would be 0.6.
         *       
         *  \sa
         *      - doBuildPlayerType()
         *
         */
        virtual
        void
        doBuildPlayerType ( int id,
                            double player_speed_max,
                            double stamina_inc_max,
                            double player_decay,
                            double inertia_moment,
                            double dash_power_rate,
                            double player_size,
                            double kickable_margin,
                            double kick_rand,
                            double extra_stamina,
                            double effort_max,
                            double effort_min ) {}


        /** This function is called after parsing a version 8 or 9
         *  player type message.
         *
         *  Override this function in your subclass to handle version 8 or 9
         *  player type messages.
         *
         *  \pre The client is using protocol version 8 or 9
         *
         *  \pre A version 8 or 9 player type message has just been parsed.
         *
         *  The following is an example (with line breaks added for clarity) of a version 8 or 9 player type message:
         *  \verbatim
            (player_type (id 0) (player_speed_max 1) (stamina_inc_max 45)
                         (player_decay 0.4) (inertia_moment 5) (dash_power_rate 0.006)
                         (player_size 0.3) (kickable_margin 0.7) (kick_rand 0)
                         (extra_stamina 0) (effort_max 1) (effort_min 0.6))\endverbatim
         *
         *  \sa
         *      - doBuildPlayerType( int id,
         *                           double player_speed_max,
         *                           double stamina_inc_max,
         *                           double player_decay,
         *                           double inertia_moment,
         *                           double dash_power_rate,
         *                           double player_size,
         *                           double kickable_margin,
         *                           double kick_rand,
         *                           double extra_stamina,
         *                           double effort_max,
         *                           double effort_min )
         *      - doBuildParam( const std::string& name,
         *                      int value )
         *      - doBuildParam( const std::string& name,
         *                      double value )
         *      - doBuildParam( const std::string& name,
         *                      const std::string& value )
         *
         */
        virtual
        void
        doBuildPlayerType() {}
        //@}


        //! \name Player Visual Parsing Functions
        //@{
        /** This function is called after parsing a player visual message.
         *
         *  Override this function in your subclass to handle player visual
         *  messages.
         *
         *  \param time The time (in server cycles) that the player visual
         *              message was sent by the simulator.
         *
         *  \pre The client is a field player.
         *
         *  \pre An entire player visual message has been parsed.
         *
         *  The following is an example (with line breaks added for clarity)
         *  of a player visual message:
         *  \verbatim
            (see 101 ((g r) 72.2 0)
                     ((f c) 20.1 0 -0 0)
                     ((f r t) 79.8 -25)
                     ((f r b) 79.8 25)
                     ((f g r b) 73 6)
                     ((f g r t) 73 -6)
                     ((f p r b) 59.7 20)
                     ((f p r c) 56.3 0)
                     ((f p r t) 59.7 -20)
                     ((f t r 20) 55.7 -44)
                     ((f t r 30) 63.4 -38)
                     ((f t r 40) 71.5 -33)
                     ((f t r 50) 79.8 -29)
                     ((f b r 20) 55.7 44)
                     ((f b r 30) 63.4 38)
                     ((f b r 40) 71.5 33)
                     ((f b r 50) 79.8 29)
                     ((f r 0) 77.5 0)
                     ((f r t 10) 78.3 -7)
                     ((f r t 20) 79.8 -14)
                     ((f r t 30) 83.1 -21)
                     ((f r b 10) 78.3 7)
                     ((f r b 20) 79.8 14)
                     ((f r b 30) 83.1 21)
                     ((b) 24.5 -13 0.98 0.6)
                     ((p "one" 2) 22.2 32 0 -0 -81 -82)
                     ((p "one" 3) 22.2 -36 0 -0.3 104 65)
                     ((p "one" 4) 22.2 -17 -0 -0.1 179 -180)
                     ((p "one") 24.5 0)
                     ((p "one" 6) 27.1 27 0 -0 -100 -100)
                     ((p "one" 7) 22.2 -14 0.444 0.2 -1 0)
                     ((p "one" 8) 20.1 -7 -0 0.1 -101 -68)
                     ((p "one" 9) 30 35 0 -0 -101 -101)
                     ((p "one" 10) 33.1 -42 0 -0 109 109)
                     ((p "two") 30 -28)
                     ((p "two") 30 -28)
                     ((p "two" 3) 24.5 20 -0 -0 -87 -87)
                     ((p "two" 5) 27.1 1 0 0 -170 -135)
                     ((p "two" 6) 27.1 4 0 0 -124 -159)
                     ((p "two" 7) 22.2 34 0 -0 -12 -80)
                     ((p "two" 8) 14.9 5 0 -0 -4 -33)
                     ((p "two" 10) 24.5 40 0 -0 -77 -78)
                     ((p) 66.7 0)
                     ((l r) 72.2 90))\endverbatim
         *
         *  In this example the
         *      - \e time parameter would be 101.
         *
         *  \sa
         *      - doBuildLine( double dist, double dir )
         *      - doBuildLine( double dir )
         *      - doBuildFlag( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( bool close,
         *                     double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( bool close, double dist, double dir )
         *      - doBuildFlag( bool close, double dir )
         *      - doBuildGoal( bool close, double dist, double dir )
         *      - doBuildGoal( bool close, double dir )
         *      - doBuildBall( bool close,
         *                     double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildBall( bool close, double dist, double dir )
         *      - doBuildBall( bool close, double dir )
         *      - doBuildPlayer( bool close,
         *                       double dist, double dir,
         *                       double dist_chg, double dir_chg,
         *                       double orientation, double head_orientation,
         *                       double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
         *                       double dist, double dir,
         *                       double dist_chg, double dir_chg,
         *                       double orientation, double head_orientation,
         *                       bool tackling )
         *      - doBuildPlayer( bool close,
         *                       double dist, double dir,
         *                       double dist_chg, double dir_chg,
         *                       double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
         *                       double dist, double dir,
         *                       double dist_chg, double dir_chg,
         *                       bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
         *                       double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
         *                       bool tackling )
         *      - doBuildPlayer( bool close,
         *                       double dist, double dir,
         *                       double dist_chg, double dir_chg,
         *                       double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
         *                       double dist, double dir,
         *                       double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         */
        virtual
        void
        doBuildVisual( int time ) {}
        
        /** This function is called after parsing a line section of player visual message, 
         *  when the client is in a high quality view mode.
         *
         *  Override this function in your subclass to handle high quality line sections of player visual
         *  messages.
         *
         *  \param dist The distance (in meters) from the client to the point the center of
         *              the client's field of view intersects the line.
         *  \param dir  The angle of the line to the center of the client's field of view
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire line section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a line section of a player visual message:
         *  \verbatim
            ((l r) 72.2 90)\endverbatim
         *
         *  In this example the:
         *      - \e dist parameter would be 72.2.
         *      - \e dir  paramter would be 90.
         *
         *  \sa
         *      - doBuildLine( double dir )
         *      - doBuildVisual( int time )
         *      - doBuildTopLocation()
         *      - doBuildBottomLocation()
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *
         */
        virtual
        void
        doBuildLine( double dist, double dir ) {}

        /** This function is called after parsing a line section of player visual message, 
         *  when the client is in a low quality view mode.
         *
         *  Override this function in your subclass to handle low quality line sections of player visual
         *  messages.
         *
         *  \param dist The distance (in meters) from the client to the point the center of
         *              the client's field of view intersects the line.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a low quality view mode.
         *
         *  \pre An entire line section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a line section of a player visual message:
         *  \verbatim
            ((l r) 90)\endverbatim
         *
         *  In this example the:
         *      - \e dir  paramter would be 90.
         *
         *  \sa
         *      - doBuildLine( double dist, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildTopLocation()
         *      - doBuildBottomLocation()
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *
         */
        virtual
        void
        doBuildLine( double dir ) {}


        /** This function is called after parsing a flag section of player visual message, 
         *  when the client is in a high quality view mode, the flag is not far away and
         *  doBuildFlag( double dist, double dir, double dist_chg, double dir_chg ) has not
         *  been overridden.
         *
         *  Override this function in your subclass to handle high quality near flag sections
         *  of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildFlag( double dist, double dir,
         *           double dist_chg, double dir_chg ) instead.
         *
         *  \param close    false if the flag is in the client's view cone, true otherwise.  Since
         *                  dist_chg and dir_chg are only ever sent if the flag is in the client's
         *                  view cone, close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the flag.
         *  \param dir      The angle from the client's center of view to the flag.
         *  \param dist_chg The distance change of the flag relative to the player. Since the flag
         *                  is stationary this can be used with \e dir_chg to augment the client's velocity
         *                  information found in the sense body messages.
         *  \param dir_chg  This is the direction change of the flag relative to the client.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire flag section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a near flag section of a player visual message:
         *  \verbatim
            ((f t 0) 26.8 33 0 0.7)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dist     parameter would be 26.8.
         *      - \e dir      parameter would be 33.
         *      - \e dist_chg parameter would be 0.
         *      - \e dir_chg  parameter would be 0.7.
         *
         *  \sa
         *      - doBuildFlag( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( bool close, double dist, double dir )
         *      - doBuildFlag( bool close, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildFlagOffset( int offset )
         *      - doBuildTopLocation()
         *      - doBuildBottomLocation()
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *      - doBuildCenterLocation()
         *      - doBuildPenaltyLocation()
         *      - doBuildGoalLocation()
         *
         */
        virtual
        void
        doBuildFlag( bool close,
                     double dist, double dir,
                     double dist_chg, double dir_chg ) {}

        /** This function is called after parsing a flag section of player visual message, 
         *  when the client is in a high quality view mode and the flag is not far away.
         *
         *  Override this function in your subclass to handle high quality near flag sections
         *  of player visual messages.
         *
         *  \note This function deprecates doBuildFlag( bool close, double dist, double dir,
         *                                              double dist_chg, double dir_chg ).
         *
         *  \param dist     The distance (in meters) from the client to the flag.
         *  \param dir      The angle from the client's center of view to the flag.
         *  \param dist_chg The distance change of the flag relative to the player. Since the flag
         *                  is stationary this can be used with \e dir_chg to augment the client's velocity
         *                  information found in the sense body messages.
         *  \param dir_chg  This is the direction change of the flag relative to the client.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire flag section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a near flag section of a player visual message:
         *  \verbatim
            ((f t 0) 26.8 33 0 0.7)\endverbatim
         *
         *  In this example the:
         *      - \e dist     parameter would be 26.8.
         *      - \e dir      parameter would be 33.
         *      - \e dist_chg parameter would be 0.
         *      - \e dir_chg  parameter would be 0.7.
         *
         *  \sa
         *      - doBuildFlag( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( bool close, double dist, double dir )
         *      - doBuildFlag( bool close, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildFlagOffset( int offset )
         *      - doBuildTopLocation()
         *      - doBuildBottomLocation()
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *      - doBuildCenterLocation()
         *      - doBuildPenaltyLocation()
         *      - doBuildGoalLocation()
         *
         */
        virtual
        void
        doBuildFlag( double dist, double dir,
                     double dist_chg, double dir_chg ) 
        {
            doBuildFlag( false, dist, dir, dist_chg, dir_chg );
        }

        /** This function is called after parsing a flag section of player visual message, 
         *  when the client is in a high quality view mode and the flag is close or far away.
         *
         *  Override this function in your subclass to handle high quality close or far flag sections
         *  of player visual messages.
         *
         *  \param close    false if the flag is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the flag.
         *  \param dir      The angle from the client's center of view to the flag.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire flag section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a far flag section of a player visual message:
         *  \verbatim
            ((f l t 20) 67.4 -31)\endverbatim
         *
         *  In this example the:
         *      - \e dist     parameter would be 67.4.
         *      - \e dir      parameter would be -31.
         *
         *  The following is an example of a close flag section of a player visual message:
         *  \verbatim
            ((F) 2.3 80)\endverbatim
         *
         *  In this example the:
         *      - \e dist     parameter would be 2.3.
         *      - \e dir      parameter would be 80.
         *
         *  \sa
         *      - doBuildFlag( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( bool close, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildFlagOffset( int offset )
         *      - doBuildTopLocation()
         *      - doBuildBottomLocation()
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *      - doBuildCenterLocation()
         *      - doBuildPenaltyLocation()
         *      - doBuildGoalLocation()
         *
         */
        virtual
        void
        doBuildFlag( bool close, double dist, double dir ) {}

        /** This function is called after parsing a flag section of player visual message, 
         *  when the client is in a low quality view mode and the flag is close or far away.
         *
         *  Override this function in your subclass to handle low quality close or far flag sections
         *  of player visual messages.
         *
         *  \param close    false if the flag is in the client's view cone, true otherwise.
         *  \param dir      The angle from the client's center of view to the flag.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a low quality view mode.
         *
         *  \pre An entire flag section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a far flag section of a player visual message:
         *  \verbatim
            ((f l t 20) -31)\endverbatim
         *
         *  In this example the:
         *      - \e dir      parameter would be -31.
         *
         *  The following is an example of a close flag section of a player visual message:
         *  \verbatim
            ((F) 80)\endverbatim
         *
         *  In this example the:
         *      - \e dir      parameter would be 80.
         *
         *  \sa
         *      - doBuildFlag( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildFlag( bool close, double dist, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildFlagOffset( int offset )
         *      - doBuildTopLocation()
         *      - doBuildBottomLocation()
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *      - doBuildCenterLocation()
         *      - doBuildPenaltyLocation()
         *      - doBuildGoalLocation()
         *
         */
        virtual
        void
        doBuildFlag( bool close, double dir ) {}

        /** This function is called after parsing a goal section of player visual message, 
         *  when the client is in a high quality view mode, the goal is not far away and
         *  doBuildGoal( double dist, double dir, double dist_chg, double dir_chg ) has not
         *  been overridden.
         *
         *  Override this function in your subclass to handle high quality near goal sections
         *  of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildGoal( double dist, double dir,
         *           double dist_chg, double dir_chg ) instead.
         *
         *  \param close    false if the goal is in the client's view cone, true otherwise.  Since
         *                  dist_chg and dir_chg are only ever sent if the goal is in the client's
         *                  view cone, close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the goal.
         *  \param dir      The angle from the client's center of view to the goal.
         *  \param dist_chg The distance change of the goal relative to the player. Since the flag
         *                  is stationary this can be used with \e dir_chg to augment the client's velocity
         *                  information found in the sense body messages.
         *  \param dir_chg  This is the direction change of the goal relative to the client.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire goal section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a near goal section of a player visual message:
         *  \verbatim
            ((g r) 34.8 -19 0 -0.1)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dist     parameter would be 34.8.
         *      - \e dir      parameter would be -19.
         *      - \e dist_chg parameter would be 0.
         *      - \e dir_chg  parameter would be -0.1.
         *
         *  \sa
         *      - doBuildGoal( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildGoal( bool close, double dist, double dir )
         *      - doBuildGoal( bool close, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *
         */
        virtual
        void
        doBuildGoal( bool close,
                     double dist, double dir,
                     double dist_chg, double dir_chg ) {}


        /** This function is called after parsing a goal section of player visual message, 
         *  when the client is in a high quality view mode and the goal is not far away.
         *
         *  Override this function in your subclass to handle high quality near goal sections
         *  of player visual messages.
         *
         *  \note This function deprecates doBuildGoal( bool close, double dist, double dir,
         *                                              double dist_chg, double dir_chg ).
         *
         *  \param dist     The distance (in meters) from the client to the goal.
         *  \param dir      The angle from the client's center of view to the goal.
         *  \param dist_chg The distance change of the goal relative to the player. Since the flag
         *                  is stationary this can be used with \e dir_chg to augment the client's velocity
         *                  information found in the sense body messages.
         *  \param dir_chg  This is the direction change of the goal relative to the client.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire goal section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a near goal section of a player visual message:
         *  \verbatim
            ((g r) 34.8 -19 0 -0.1)\endverbatim
         *
         *  In this example the:
         *      - \e dist     parameter would be 34.8.
         *      - \e dir      parameter would be -19.
         *      - \e dist_chg parameter would be 0.
         *      - \e dir_chg  parameter would be -0.1.
         *
         *  \sa
         *      - doBuildGoal( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildGoal( bool close, double dist, double dir )
         *      - doBuildGoal( bool close, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *
         */
        virtual
        void
        doBuildGoal( double dist, double dir,
                     double dist_chg, double dir_chg ) 
        {
            doBuildGoal( false, dist, dir, dist_chg, dir_chg );
        }

        /** This function is called after parsing a goal section of player visual message, 
         *  when the client is in a high quality view mode and the goal is close or far away.
         *
         *  Override this function in your subclass to handle high quality close or far goal sections
         *  of player visual messages.
         *
         *  \param close    false if the goal is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the goal.
         *  \param dir      The angle from the client's center of view to the goal.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire goal section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a far goal section of a player visual message:
         *  \verbatim
            ((g l) 82.3 -36)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dist     parameter would be 82.3.
         *      - \e dir      parameter would be -36.
         *
         *  The following is an example of a close goal section of a player visual message:
         *  \verbatim
            ((G) 2.9 166)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be true.
         *      - \e dist     parameter would be 2.9.
         *      - \e dir      parameter would be -166.
         *
         *  \sa
         *      - doBuildGoal( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildGoal( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildGoal( bool close, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *
         */
        virtual
        void
        doBuildGoal( bool close, double dist, double dir ) {}

        /** This function is called after parsing a goal section of player visual message, 
         *  when the client is in a low quality view mode and the goal is close or far away.
         *
         *  Override this function in your subclass to handle low quality close or far goal sections
         *  of player visual messages.
         *
         *  \param close    false if the goal is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the goal.
         *  \param dir      The angle from the client's center of view to the goal.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a low quality view mode.
         *
         *  \pre An entire goal section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a far goal section of a player visual message:
         *  \verbatim
            ((g l) -36)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dir      parameter would be -36.
         *
         *  The following is an example of a close goal section of a player visual message:
         *  \verbatim
            ((G) 166)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be true.
         *      - \e dir      parameter would be -166.
         *
         *  \sa
         *      - doBuildGoal( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildGoal( double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildGoal( bool close, double dist, double dir )
         *      - doBuildVisual( int time )
         *      - doBuildLeftLocation()
         *      - doBuildRightLocation()
         *
         */
        virtual
        void
        doBuildGoal( bool close, double dir ) {}

        /** This function is called after parsing a ball section of player visual message, 
         *  when the client is in a high quality view mode, the ball is not far away and
         *  doBuildPlayerVisBall( double dist, double dir, double dist_chg, double dir_chg ) has not
         *  been overridden.
         *
         *  Override this function in your subclass to handle high quality near ball sections
         *  of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildPlayerVisBall( double dist, double dir,
         *           double dist_chg, double dir_chg ) instead.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.  Since
         *                  dist_chg and dir_chg are only ever sent if the ball is in the client's
         *                  view cone, \e close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the ball.
         *  \param dir      The angle from the client's center of view to the ball.
         *  \param dist_chg The distance change of the ball relative to the player.
         *  \param dir_chg  This is the direction change of the ball relative to the client.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire ball section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a near ball section of a player visual message:
         *  \verbatim
            ((b) 24.5 -36 0.49 -0.8)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dist     parameter would be 24.5.
         *      - \e dir      parameter would be -36.
         *      - \e dist_chg parameter would be 0.49.
         *      - \e dir_chg  parameter would be -0.8.
         *
         *  \sa
         *      - doBuildPlayerVisBall( double dist, double dir,
         *                              double dist_chg, double dir_chg )
         *      - doBuildBall( bool close, double dist, double dir )
         *      - doBuildBall( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildBall( bool close,
                     double dist, double dir,
                     double dist_chg, double dir_chg ) {}



        /** This function is called after parsing a ball section of player visual message, 
         *  when the client is in a high quality view mode and the ball is not far away
         *
         *  Override this function in your subclass to handle high quality near ball sections
         *  of player visual messages.
         *
         *  \note This function deprecates doBuildBall( bool close, double dist, double dir,
         *           double dist_chg, double dir_chg ).
         *
         *  \param dist     The distance (in meters) from the client to the ball.
         *  \param dir      The angle from the client's center of view to the ball.
         *  \param dist_chg The distance change of the ball relative to the player.
         *  \param dir_chg  This is the direction change of the ball relative to the client.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire ball section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a near ball section of a player visual message:
         *  \verbatim
            ((b) 24.5 -36 0.49 -0.8)\endverbatim
         *
         *  In this example the:
         *      - \e dist     parameter would be 24.5.
         *      - \e dir      parameter would be -36.
         *      - \e dist_chg parameter would be 0.49.
         *      - \e dir_chg  parameter would be -0.8.
         *
         *  \sa
         *      - doBuildBall( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildBall( bool close, double dist, double dir )
         *      - doBuildBall( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayerVisBall( double dist, double dir,
                              double dist_chg, double dir_chg ) 
        {
            doBuildBall( false, dist, dir, dist_chg, dir_chg );
        }

        /** This function is called after parsing a ball section of player visual message, 
         *  when the client is in a high quality view mode and the ball is close or far away.
         *
         *  Override this function in your subclass to handle high quality close or far ball sections
         *  of player visual messages.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the ball.
         *  \param dir      The angle from the client's center of view to the ball.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire ball section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a far ball section of a player visual message:
         *  \verbatim
            ((b) 27.1 -37)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dist     parameter would be 27.1.
         *      - \e dir      parameter would be -37.
         *
         *  The following is an example of a close ball section of a player visual message:
         *  \verbatim
            ((B) 2.7 46)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be true.
         *      - \e dist     parameter would be 2.7.
         *      - \e dir      parameter would be 46.
         *
         *  \sa
         *      - doBuildBall( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildPlayerVisBall( double dist, double dir,
         *                              double dist_chg, double dir_chg )
         *      - doBuildBall( double dist, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildBall( bool close, double dist, double dir ) {}

        /** This function is called after parsing a ball section of player visual message, 
         *  when the client is in a low quality view mode and the ball is close or far away.
         *
         *  Override this function in your subclass to handle low quality close or far ball sections
         *  of player visual messages.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the ball.
         *  \param dir      The angle from the client's center of view to the ball.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a low quality view mode.
         *
         *  \pre An entire ball section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  The following is an example of a far ball section of a player visual message:
         *  \verbatim
            ((b) -37)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be false.
         *      - \e dir      parameter would be -37.
         *
         *  The following is an example of a close ball section of a player visual message:
         *  \verbatim
            ((B) 46)\endverbatim
         *
         *  In this example the:
         *      - \e close    parameter would be true.
         *      - \e dir      parameter would be 46.
         *
         *  \sa
         *      - doBuildBall( bool close, double dist, double dir,
         *                     double dist_chg, double dir_chg )
         *      - doBuildPlayerVisBall( double dist, double dir,
         *                              double dist_chg, double dir_chg )
         *      - doBuildBall( bool close, double dist, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildBall( bool close, double dir ) {}

        /** This function is called after parsing a version 8 or 9 player section of player visual message, 
         *  when the client is in a high quality view mode, the player is not far away, is pointing and
         *  doBuildPlayer( double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             double point_dir, bool tackling ) has not
         *  been overridden.
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  near pointing player sections of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildPlayer( double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             double point_dir, bool tackling ) instead.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.  Since
         *                  dist_chg, dir_chg, etc are only ever sent if the player is in the client's
         *                  view cone, \e close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param dist_chg The distance change of the player relative to the client.
         *  \param dir_chg  The the direction change of the player relative to the client.
         *  \param orientation  The the direction of the player's body relative to the client.
         *  \param head_orientation  The the direction of the player's head relative to the client.
         *  \param point_dir         The the direction of the player is pointing relative to the client.
         *  \param tackling          true if the player is tackling, false otherwise.
         *
         *  \pre The client is using protocol 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire near, pointing, player section of a player visual message has been parsed.
         *
         *  \pre A version 8 or 9 player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close,
                       double dist, double dir,
                       double dist_chg, double dir_chg,
                       double orientation, double head_orientation,
                       double point_dir, bool tackling )
        {
            doBuildPlayer( close, dist, dir, dist_chg, dir_chg,
                           orientation, head_orientation );
        }

        /** This function is called after parsing a version 8 or 9 player section of player visual message, 
         *  when the client is in a high quality view mode, the player is not far away and is pointing.
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  near pointing player sections of player visual messages.
         *
         *  \note This function is deprecates doBuildPlayer( bool close, double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             double point_dir, bool tackling ).
         *
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param dist_chg The distance change of the player relative to the client.
         *  \param dir_chg  The the direction change of the player relative to the client.
         *  \param orientation  The the direction of the player's body relative to the client.
         *  \param head_orientation  The the direction of the player's head relative to the client.
         *  \param point_dir         The the direction of the player is pointing relative to the client.
         *  \param tackling          true if the player is tackling, false otherwise.
         *
         *  \pre The client is using protocol 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire near, pointing, player section of a player visual message has been parsed.
         *
         *  \pre A version 8 or 9 player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( double dist, double dir,
                       double dist_chg, double dir_chg,
                       double orientation, double head_orientation,
                       double point_dir, bool tackling )
        {
            doBuildPlayer( false, dist, dir, dist_chg, dir_chg,
                           orientation, head_orientation, tackling );
        }

        /** This function is called after parsing a player section of player visual message, 
         *  when the client is in a high quality view mode, the player is not far away, is not
         *  pointing or the client is using protcol version 7 and
         *  doBuildPlayer( double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             bool tackling ) has not
         *  been overridden.
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  near non-pointing or version 7 high quality near player sections of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildPlayer( double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             bool tackling ) instead.
         *  \note    This function deprecates doBuildPlayer( bool close,
         *                                                   double dist, double dir,
         *                                                   double dist_chg, double dir_chg,
         *                                                   double orientation, double head_orientation ).
         *
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.  Since
         *                  dist_chg, dir_chg, etc are only ever sent if the player is in the client's
         *                  view cone, \e close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param dist_chg The distance change of the player relative to the client.
         *  \param dir_chg  The the direction change of the player relative to the client.
         *  \param orientation  The the direction of the player's body relative to the client.
         *  \param head_orientation  The the direction of the player's head relative to the client.
         *  \param tackling          false if the player is not tackling or protocol 7 is being used,
         *                           true otherwise.
         *
         *  \pre Either:
         *        - The client is using protocol 8 or 9 and the player is not pointing, or
         *        - The client is using protocol 7.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre Either an entire:
                     - near, non-pointing or
                     - version 7
                 player section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close,
                       double dist, double dir,
                       double dist_chg, double dir_chg,
                       double orientation, double head_orientation,
                       bool tackling )
        {
            doBuildPlayer( close, dist, dir, dist_chg, dir_chg,
                           orientation, head_orientation );
        }

        /** This function is called after parsing a player section of player visual message, 
         *  when the client is in a high quality view mode, the player is not far away and the
         *  player is not pointing or the client is using protocol 7. 
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  near non-pointing or version 7 high quality near player sections of player visual messages.
         *
         *  \note This function is deprecates doBuildPlayer( bool close, double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             bool tackling ).
         *
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param dist_chg The distance change of the player relative to the client.
         *  \param dir_chg  The the direction change of the player relative to the client.
         *  \param orientation  The the direction of the player's body relative to the client.
         *  \param head_orientation  The the direction of the player's head relative to the client.
         *  \param tackling          false if the player is not tackling or the client is using protocol 7,
         *                           true otherwise.
         *
         *  \pre Either:
         *        - The client is using protocol 8 or 9 and the player is not pointing, or
         *        - The client is using protocol 7.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire near, non-pointing or version 7, player section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *        
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( double dist, double dir,
                       double dist_chg, double dir_chg,
                       double orientation, double head_orientation,
                       bool tackling )
        {
            doBuildPlayer( false, dist, dir, dist_chg, dir_chg,
                           orientation, head_orientation, tackling );
        }

        /** This function never called.
         *
         *  \warning This function is never called because the server never sends dist_chg and dir_chg
         *           without body and head directions.
         */
        virtual
        void
        doBuildPlayer( bool close,
                       double dist, double dir,
                       double dist_chg, double dir_chg,
                       double point_dir, bool tackling )
        {
            doBuildPlayer( close, dist, dir, dist_chg, dir_chg );
        }        
       
        /** This function never called.
         *
         *  \warning This function is never called because the server never sends dist_chg and dir_chg
         *           without body and head directions.
         */
        virtual
        void
        doBuildPlayer( double dist, double dir,
                       double dist_chg, double dir_chg,
                       double point_dir, bool tackling )
        {
            doBuildPlayer( false, dist, dir, dist_chg, dir_chg, tackling );
        }

        /** This function never called.
         *
         *  \warning This function is never called because the server never sends dist_chg and dir_chg
         *           without body and head directions.
         */
        virtual
        void
        doBuildPlayer( bool close,
                       double dist, double dir,
                       double dist_chg, double dir_chg,
                       bool tackling )
        {
            doBuildPlayer( close, dist, dir, dist_chg, dir_chg );
        }

        /** This function never called.
         *
         *  \warning This function is never called because the server never sends dist_chg and dir_chg
         *           without body and head directions.
         */
        virtual
        void
        doBuildPlayer( double dist, double dir,
                       double dist_chg, double dir_chg,
                       bool tackling )
        {
            doBuildPlayer( false, dist, dir, dist_chg, dir_chg, tackling );
        }

        /** This function is called after parsing a version 8 or 9 player section of player visual message, 
         *  when the client is in a high quality view mode, the player is far away, is pointing and
         *  doBuildPlayer( double dist, double dir,
         *             double point_dir, bool tackling ) has not
         *  been overridden.
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  far pointing player sections of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildPlayer( double dist, double dir,
         *             double point_dir, bool tackling ) instead.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.  Since
         *                  point_dir and tackling are only ever sent if the player is in the client's
         *                  view cone, \e close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param point_dir         The the direction of the player is pointing relative to the client.
         *  \param tackling          true if the player is tackling, false otherwise.
         *
         *  \pre The client is using protocol 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire far, pointing, player section of a player visual message has been parsed.
         *
         *  \pre A version 8 or 9 player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close, double dist, double dir,
                       double point_dir, bool tackling )
        {
            doBuildPlayer( close, dist, dir );
        }

        /** This function is called after parsing a version 8 or 9 player section of player visual message, 
         *  when the client is in a high quality view mode, the player is far away and is pointing.
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  far pointing player sections of player visual messages.
         *
         *  \note This function is deprecates doBuildPlayer( bool close, double dist, double dir,
         *             double point_dir, bool tackling ).
         *
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param point_dir         The the direction of the player is pointing relative to the client.
         *  \param tackling          true if the player is tackling, false otherwise.
         *
         *  \pre The client is using protocol 8 or 9.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire far, pointing, player section of a player visual message has been parsed.
         *
         *  \pre A version 8 or 9 player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( double dist, double dir,
                       double point_dir, bool tackling )
        {
            doBuildPlayer( false, dist, dir, tackling );
        }


        /** This function is called after parsing a player section of player visual message, 
         *  when the client is in a high quality view mode, the player is far away and not pointing
         *  or the player is very far away and the client is using protocol 7,
         *  or the player is very close, but outside of the view cone.
         *
         *  Override this function in your subclass to handle high quality
         *  far non-pointing or version 7, or very close player sections of player visual messages.
         *
         *  \note This function is deprecates doBuildPlayer( bool close, double dist, double dir ).
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param tackling true if the player is tackling and is in the clients view cone, false otherwise.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire far non-pointing or version 7, or very close player section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close, double dist, double dir,
                       bool tackling )
        {
            doBuildPlayer( close, dist, dir );
        }


        /** This function is called after parsing a player section of player visual message, 
         *  when the client is in a high quality view mode, the player is not far away, is not
         *  pointing or the client is using protcol version 7 and
         *  doBuildPlayer( bool close, double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             bool tackling ), and 
         *  doBuildPlayer( double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             bool tackling ) have not
         *  been overridden.
         *
         *  Override this function in your subclass to handle version 8 or 9 high quality
         *  near non-pointing or version 7 high quality near player sections of player visual messages.
         *
         *  \warning This function is deprecated.  Overide doBuildPlayer( double dist, double dir,
         *             double dist_chg, double dir_chg,
         *             double orientation, double head_orientation,
         *             bool tackling ) instead.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.  Since
         *                  dist_chg, dir_chg, etc are only ever sent if the player is in the client's
         *                  view cone, \e close is always false, thus this parameter is redundant.
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *  \param dist_chg The distance change of the player relative to the client.
         *  \param dir_chg  The the direction change of the player relative to the client.
         *  \param orientation  The the direction of the player's body relative to the client.
         *  \param head_orientation  The the direction of the player's head relative to the client.
         *
         *  \pre Either:
         *        - The client is using protocol 8 or 9 and the player is not pointing, or
         *        - The client is using protocol 7.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre Either an entire:
                     - near, non-pointing or
                     - version 7
                 player section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close,
                       double dist, double dir,
                       double dist_chg, double dir_chg,
                       double orientation, double head_orientation ) {}
      

        /** This function never called.
         *
         *  \warning This function is never called because the server never sends dist_chg and dir_chg
         *           without body and head directions.
         */
        virtual
        void
        doBuildPlayer( bool close,
                       double dist, double dir,
                       double dist_chg, double dir_chg ) {}

        /** This function is called after parsing a player section of player visual message, 
         *  when the client is in a high quality view mode, the player is far away and not pointing
         *  or the player is very far away and the client is using protocol 7,
         *  or the player is very close, but outside of the view cone,
         *  and doBuildPlayer( bool close,
         *  double dist, double dir, bool tackling ) has not been overridden.
         *
         *  Override this function in your subclass to handle high quality
         *  far non-pointing or version 7, or very close player sections of player visual messages.
         *
         *  \warning This function is deprecated.  Override doBuildPlayer( bool close,
         *  double dist, double dir, bool tackling ) instead.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.
         *  \param dist     The distance (in meters) from the client to the player.
         *  \param dir      The angle from the client's center of view to the player.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a high quality view mode.
         *
         *  \pre An entire far non-pointing or version 7, or very close player section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close, double dist, double dir ) {}

        /** This function is called after parsing a player section of player visual message, 
         *  when the client is in a low quality view mode.
         *
         *  Override this function in your subclass to handle low quality
         *  player sections of player visual messages.
         *
         *  \param close    false if the ball is in the client's view cone, true otherwise.
         *  \param dir      The angle from the client's center of view to the player.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is in a low quality view mode.
         *
         *  \pre An entire player section of a player visual message has been parsed.
         *
         *  \pre A player visual message is being parsed.
         *
         *  \sa
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation,
                                 bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( double dist, double dir,
                                 double point_dir, bool tackling )
         *      - doBuildPlayer( bool close, double dist, double dir, bool tackling )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg,
                                 double orientation, double head_orientation )
         *      - doBuildPlayer( bool close,
                                 double dist, double dir,
                                 double dist_chg, double dir_chg )
         *      - doBuildPlayer( bool close, double dist, double dir )
         *      - doBuildVisual( int time )
         *
         */
        virtual
        void
        doBuildPlayer( bool close, double dir ) {}
        //@}
    


        //! \name Coach Visual Parsing Functions
        //@{
        /** This function is called after parsing a coach visual message.
         *
         *  Override this function in your subclass to handle coach visual
         *  messages.
         *
         *  \param time The time (in server cycles) that the coach visual
         *              message was sent by the simulator.
         *
         *  \pre The client is a coach player.
         *
         *  \pre An entire coach visual message has been parsed.
         *
         */
        virtual
        void
        doBuildGlobalVisual( int time ) {}

        /** This function is called after parsing a goal section 
         *  of a coach visual message.
         *
         *  Override this function in your subclass to handle goal
         *  sections of a coach visual message.
         *
         *  \param x The x co-ordinate of the goal.
         *  \param y The y co-ordinate of the goal.
         *
         *  \pre The client is a coach.
         *
         *  \pre An entire goal section of coach visual message has
         *       been parsed.
         *
         *  \pre A coach visual message is being parsed.
         *
         *  \sa
         *      - doBuildGlobalVisual( int time )
         */
        virtual
        void
        doBuildGlobalGoal( double x, double y ) {}

        /** This function is called after parsing a ball section 
         *  of a coach visual message.
         *
         *  Override this function in your subclass to handle ball
         *  sections of a coach visual message.
         *
         *  \param x       The x co-ordinate of the ball.
         *  \param y       The y co-ordinate of the ball.
         *  \param delta_x The magnitude of the ball's velociy along the
         *                 x co-ordinate.
         *  \param delta_y The magnitude of the ball's velociy along the
         *                 y co-ordinate.
         *
         *  \pre The client is a coach.
         *
         *  \pre An entire ball section of coach visual message has
         *       been parsed.
         *
         *  \pre A coach visual message is being parsed.
         *
         *  \sa
         *      - doBuildGlobalVisual( int time )
         */
        virtual
        void
        doBuildGlobalBall( double x, double y,
                           double delta_x, double delta_y ) {}

        /** This function is called after parsing a player section 
         *  of a coach visual message, where the player is pointing.
         *
         *  Override this function in your subclass to handle pointing player
         *  sections of a coach visual message.
         *
         *  \note This function deprecates doBuildGlobalPlayer( x, y,
                                 delta_x, delta_y, 
                                 orientation, head_orientation ).
         *                        
         *  \param x       The x co-ordinate of the player.
         *  \param y       The y co-ordinate of the player.
         *  \param delta_x The magnitude of the player's velociy along the
         *                 x co-ordinate.
         *  \param delta_y The magnitude of the player's velociy along the
         *                 y co-ordinate.
         *  \param orientation The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param point_dir The direction the player is pointing in.
         *  \param tackle True if the player is tackling.  False otherwise.
         *
         *  \pre The player is pointing.
         *
         *  \pre The client is a coach.
         *
         *  \pre An entire player section of coach visual message has
         *       been parsed.
         *
         *  \pre A coach visual message is being parsed.
         *
         *  \sa
         *      - doBuildGlobalVisual( int time )
         */
        virtual
        void
        doBuildGlobalPlayer( double x, double y,
                             double delta_x, double delta_y,
                             double orientation, double head_orientation,
                             double point_dir, bool tackle )
        {
            doBuildGlobalPlayer( x, y,
                                 delta_x, delta_y, 
                                 orientation, head_orientation );
        }
      
        /** This function is called after parsing a non-pointing player section 
         *  of a coach visual message.
         *
         *  Override this function in your subclass to handle non-pointing player
         *  sections of a coach visual message.
         *
         *  \note This function deprecates doBuildGlobalPlayer( x, y,
                                 delta_x, delta_y, 
                                 orientation, head_orientation ).
         *                        
         *  \param x       The x co-ordinate of the player.
         *  \param y       The y co-ordinate of the player.
         *  \param delta_x The magnitude of the player's velociy along the
         *                 x co-ordinate.
         *  \param delta_y The magnitude of the player's velociy along the
         *                 y co-ordinate.
         *  \param orientation The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *  \param tackle True if the player is tackling.  False otherwise.
         *
         *  \pre The player is not pointing.
         *
         *  \pre The client is a coach.
         *
         *  \pre An entire player section of coach visual message has
         *       been parsed.
         *
         *  \pre A coach visual message is being parsed.
         *
         *  \sa
         *      - doBuildGlobalVisual( int time )
         */
        virtual
        void
        doBuildGlobalPlayer( double x, double y,
                             double delta_x, double delta_y,
                             double orientation, double head_orientation, 
                             bool tackle )
        {
            doBuildGlobalPlayer( x, y,
                                 delta_x, delta_y, 
                                 orientation, head_orientation );
        }
      

        /** This function is called after parsing a player section 
         *  of a coach visual message.
         *
         *  Override this function in your subclass to handle player
         *  sections of a coach visual message.
         *
         *  \warning This function deprecated.  Use 
         doBuildGlobalPlayer( double x, double y,
                             double delta_x, double delta_y,
                             double orientation, double head_orientation,
                             double point_dir, bool tackle ) and
         doBuildGlobalPlayer( double x, double y,
                             double delta_x, double delta_y,
                             double orientation, double head_orientation, 
                             bool tackle ) instead.
         *                        
         *  \param x       The x co-ordinate of the player.
         *  \param y       The y co-ordinate of the player.
         *  \param delta_x The magnitude of the player's velociy along the
         *                 x co-ordinate.
         *  \param delta_y The magnitude of the player's velociy along the
         *                 y co-ordinate.
         *  \param orientation The direction the player's body is facing.
         *  \param head_orientation The direction the player's head is facing.
         *
         *  \pre The client is a coach.
         *
         *  \pre An entire player section of coach visual message has
         *       been parsed.
         *
         *  \pre A coach visual message is being parsed.
         *
         *  \sa
         *      - doBuildGlobalVisual( int time )
         */
        virtual
        void
        doBuildGlobalPlayer( double x, double y,
                             double delta_x, double delta_y,
                             double orientation, double head_orientation ) {}
        //@}



        //! \name Field Location Parsing Functions
        //@{
        /** This function is called after parsing a flag offset section of
         *  a flag section of a player visual message.
         *
         *  Override this function in your subclass to handle flag offsets
         *  of flags sections of player visual messages.
         *
         *  \param offset The Flags offset, which helps specify it's location.
         *
         *  \pre The client is a field player.
         *
         *  \pre A flag offset of a flag section of a player visual message has been parsed.
         *
         *  \pre A flag section of a player visual message is being parsed.
         *
         *  \pre A a player visual message is being parsed.
         *
         *  \sa
         *       - doBuildFlag( bool close,
         *                      double dist, double dir,
         *                      double dist_chg, double dir_chg )
         *       - doBuildFlag( bool close, double dist, double dir )
         *       - doBuildFlag( bool close, double dir ) {}
         *
         */
        virtual
        void
        doBuildFlagOffset( int offset ) {}


        /** This function is called after parsing a top location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle top
         *  field locations.
         *
         *  \pre A top loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildTopLocation() {}
 

        /** This function is called after parsing a bottom location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle bottom
         *  field locations.
         *
         *  \pre A bottom loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildBottomLocation() {}
 
        /** This function is called after parsing a left location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle left
         *  field locations.
         *
         *  \pre A left loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildLeftLocation() {}
 
        /** This function is called after parsing a right location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle right
         *  field locations.
         *
         *  \pre A right loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildRightLocation() {}
 
        /** This function is called after parsing a center location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle center
         *  field locations.
         *
         *  \pre A center loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildCenterLocation() {}
 
        /** This function is called after parsing a penalty location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle penalty
         *  field locations.
         *
         *  \pre A penalty loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildPenaltyLocation() {}
 
        /** This function is called after parsing a goal location of
         *  a field object or a visual
         *
         *  Override this function in your subclass to handle goal
         *  field locations.
         *
         *  \pre A goal loation of a field object has been parsed.
         *
         *  \pre A field object is being parsed.
         *
         *  \pre A a visual message is being parsed.
         *
         */
        virtual
        void
        doBuildGoalLocation() {}
        //@}

        /** This function is called after parsing a team name.
         *
         *  Override this function in your subclass to handle team names.
         *
         *  \pre A team name has been parsed.
         *
         */
        virtual
        void
        doBuildTeamName( const std::string& name ) {}

        /** This function is called after parsing a uniform number.
         *
         *  Override this function in your subclass to handle uniform numbers.
         *
         *  \pre A uniform number has been parsed.
         *
         */
        virtual
        void
        doBuildUNum( int unum ) {}

        /** This function is called after parsing a goalie indicator of a
         *  player name.
         *
         *  Override this function in your subclass to handle goalie
         *  indicators.
         *
         *  \pre A goalie inidicator has been parsed.
         *
         */
        virtual
        void
        doBuildGoalie() {}

        /** This function is called after parsing a score message
         *
         *  Override this function in your subclass to handle score messages
         *
         *  \pre An entire score message has been parsed.
         *
         */
        virtual
        void
        doBuildScore( int time, int our, int opp ) {}


        //! \name Initialisation Parsing Functions
        //@{
        /** This function is called after parsing a player init message
         *
         *  Override this function in your subclass to handle player
         *  init messages
         *
         *  \pre The client is a field player
         *
         *  \pre An entire player init message has been parsed.
         *
         */
        virtual
        void
        doBuildInit( int unum ) {}

        /** This function is called after parsing a player reconnect message
         *
         *  Override this function in your subclass to handle player reconnect messages
         *
         *  \pre The client is a player
         *
         *  \pre An entire player reconnect has been parsed.
         *
         */
        virtual
        void
        doBuildInit() {}
    
        /** This function is called after parsing an coach init message
         *
         *  Override this function in your subclass to handle player
         *  coach init messages
         *
         *  \pre The client is an coach
         *
         *  \pre An entire coach init message has been parsed.
         *
         */
        virtual
        void
        doBuildCoachInit() {}
        //@}


    
        //! \name Audio Parsing Functions
        //@{
        /** This function is called after parsing an offline coach or
         *  non-CLang online coach audio message
         *
         *  Override this function in your subclass to handle offline or
         *  non-CLang online coach audio messages.
         *
         *  \pre An entire offline coach or non-Clang online coach audio
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildCoachAudio( int time, const std::string& msg ) {}

        /** This function is called after parsing a non-goal referee message
         *
         *  Override this function in your subclass to handle non-goal referee
         *  messages.
         *
         *  \pre An entire non-goal referee message has been parsed.
         *
         */
        virtual
        void
        doBuildRefAudio( int time ) {}

        /** This function is called after parsing a goal referee message
         *
         *  Override this function in your subclass to handle goal referee
         *  messages.
         *
         *  \pre An entire goal referee message has been parsed.
         *
         */
        virtual
        void
        doBuildGoalRefAudio( int time, int score ) {}

        /** This function is called after parsing an unknown referee message.
         *  This will normally only be called if new referee messages have
         *  been added to the simulator, but not to this parser.  By using
         *  this function you can provide support in your team for the new
         *  messages without having to wait for the parser to be updated.
         *
         *  Override this function in your subclass to handle unknown referee
         *  messages.
         *
         *  \pre An entire unknown referee message has been parsed.
         *
         */
        virtual
        void
        doBuildUnknownRefAudio( int time, const std::string& message ) {}

        /** This function is called after parsing a version 7 player audio
         *  message or a version 8 opponent player audio message, where
         *  the client is a field player.
         *
         *  Override this function in your subclass to handle version 7 player
         *  audio or version 8 opponent player audio messages.
         *
         *  \pre The client is a field player
         *
         *  \pre Either:
         *         - The client is using protocol version 7, or
         *         - The audio message is from an opponent
         *
         *  \pre An entire version 7 player audio or version 8 opponent player
         *       audio message has been parsed.
         *
         */
        virtual
        void
        doBuildPlayerAudio( int time, double dir, const std::string& msg ) {}

        /** This function is called after parsing a player audio message,
         *  where the client is an online coach.
         *
         *  Override this function in your subclass to player audio
         *  messages to online coaches
         *
         *  \pre The client is an online coach
         *
         *  \pre An entire player audio message has been parsed.
         *
         */
        virtual
        void
        doBuildPlayerAudio( int time, const std::string& msg ) {}

        /** This function is called after parsing a version 8 team-mate
         *  player audio message, where the client is a field player.
         *
         *  Override this function in your subclass to handle version 8
         *  team-mate player audio messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client is using protocol version 8
         *
         *  \pre The message is from a team-mate
         *
         *  \pre An entire version 8 team-mate player audio message has
         *       been parsed.
         *
         */
        virtual
        void
        doBuildPlayerAudio( int time, double dir, int unum,
                            const std::string& msg ) {}

        /** This function is called after parsing a version 8 team-mate
         *  player audio message, where the client is a field player and
         *  the client has already received a team-mate player audio message
         *  this cycle
         *
         *  Override this function in your subclass to handle version 8
         *  incomplete team-mate player audio messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client is using protocol version 8
         *
         *  \pre The message is from a team-mate
         *
         *  \pre An entire version 8 incomplete team-mate player audio
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildPlayerAudio( int time, int unum ) {}

        /** This function is called after parsing a version 8 opponent
         *  player audio message, where the client is a field player and
         *  the client has already received an opponent player audio message
         *  this cycle
         *
         *  Override this function in your subclass to handle version 8
         *  incomplete opponent player audio messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client is using protocol version 8
         *
         *  \pre The message is from an opponent
         *
         *  \pre An entire version 8 incomplete opponent player audio
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildPlayerAudio( int time ) {}

        /** This function is called after parsing a player audio message,
         *  where the client is a field player and the message was sent by
         *  the client.
         *
         *  Override this function in your subclass to handle self player audio
         *  messages
         *
         *  \pre The client is a field player
         *
         *  \pre The message is from the client
         *
         *  \pre An entire self player audio
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildSelfAudio( int time, const std::string& msg ) {}
        //@}


    
        //! \name Substitution Parsing Functions
        //@{
        /** This function is called after parsing a substitutution message,
         *  where the player being sub'ed is a team-mate
         *
         *  Override this function in your subclass to handle team-mate 
         *  substitution messages.
         *
         *  \pre The client is a field player
         *
         *  \pre A team-mate has been substituted.
         *
         *  \pre An entire team-mate substitution
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildSubstitution( int unum, int type ) {}
    
        /** This function is called after parsing a substitutution message,
         *  where the player being sub'ed is an opponent
         *
         *  Override this function in your subclass to handle opponent
         *  substitution messages.
         *
         *  \pre The client is a field player
         *
         *  \pre An opponent has been substituted.
         *
         *  \pre An entire opponent substitution
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildSubstitution( int unum ) {}
        //@}

        //! \name Clang version parsing functions
        //@{
        /** This function is called after parsing a clang player version message
	 *  (where the player reports what version of clang it supports)
         *
         *  Override this function in your subclass to handle clang player version
	 *  messages
         *
         *  \pre The client is an online coach
         *
         *  \pre An entire clang player version message has been parsed
	 *
         *  \pre zero of more clang player version tokens have been parsed
         */
         virtual
         void
	 doBuildClangPlayerVersionMsg( ) {}

        /** This function is called after parsing a player clang version
	 *  token (where the player reports what version of clang it supports)
         *
         *  Override this function in your subclass to handle clang player
	 *  version messages
         *
         *  \pre The client is an online coach
         *
	 *  \pre A player specification has been parsed
         */
         virtual
	 void
	 doBuildClangPlayerVersion( int min, int max ) {}
        //@}

    
        //! \name Error Parsing Functions
        //@{
        /** This function is called after parsing a can't reconnect error
         *  message.
         *
         *  Override this function in your subclass to handle can't reconnect
         *  error messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client has sent a reconnect message.
         *
         *  \pre The server is in play-on playmode
         *
         *  \pre An entire can't reconnect error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildCantReconnect() {}

        /** This function is called after parsing a reconnection error
         *  message.
         *
         *  Override this function in your subclass to handle reconnection
         *  error messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client has sent a reconnect message.
         *
         *  \pre Either:
         *          - There was an error in the reconnect message,
         *          - There was an error connecting to the clients socket, or
         *          - The team-name and uniform number do not match any
         *            field player
         *
         *  \pre An entire reconnection error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildInitError() {}

        /** This function is called after parsing a player initialisation error
         *  message.
         *
         *  Override this function in your subclass to handle player initialisation 
         *  error messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client has sent an init message.
         *
         *  \pre Either:
         *          - There was an error in the init message,
         *          - There were illegal characters in the team-name,
         *          - The team-name was too long,
         *          - There are two teams already connected and the team-name
         *            matched neither,
         *          - There are already 11 players connected for the team
         *            specified,
         *          - The client tried to connect as a goalie and the team
         *            specified already has a goalie connected,
         *          - The client tried to connect with a non-existant protocol
         *            version, or
         *          - There was an error connecting to the clients socket.
         *
         *  \pre An entire player initialisation error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildNoMoreTeamOrPlayerOrGoalieError() {}
    
        /** This function is called after parsing an online coach 
         *  initialisation error message.
         *
         *  Override this function in your subclass to handle online coach
         *  initialisation error messages.
         *
         *  \pre The client is an online coach.
         *
         *  \pre The client has sent an init message.
         *
         *  \pre Either:
         *          - There was an error in the init message,
         *          - The version number specified was less than 5,
         *          - The team-name specified was too long,
         *          - The team-name specified does not match any team that is
         *            connected,
         *          - The specified team already has a coach connected,
         *          - The client tried to connect with a non-existant protocol
         *            version, or
         *          - There was an error connecting to the clients socket.
         *
         *  \pre An entire coach initialisation error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildNoSuchTeamOrAlreadyHaveCoachError() {}
 
        /** This function is called after parsing a too many moves error message.
         *
         *  Override this function in your subclass to handle too many moves 
         *  error messages.
         *
         *  \pre The client is a field player.
         *
         *  \pre The client is a goalie.
         *
         *  \pre The client has caught the ball.
         *
         *  \pre The client has already moved the prescribed maximum amount of
         *       times while the ball is caught and another move command was
         *       sent.
         *
         *  \pre An entire too many moves error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildTooManyMovesError() {}

        /** This function is called after parsing an unknown command error message.
         *
         *  Override this function in your subclass to handle unknown command 
         *  error messages.
         *
         *  \pre The client has send a command that the server does not
         *       recognise
         *
         *  \pre An entire unknown commad error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildUnknownCommandError() {}

        /** This function is called after parsing an illegal command form error message.
         *
         *  Override this function in your subclass to handle illegal command form 
         *  error messages.
         *
         *  \pre The client has sent a command whoes format does not match
         *       that expected by the server
         *
         *  \pre An entire illegal command form error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildIllegalCommandError() {}

        /** This function is called after parsing a say too long error message.
         *
         *  Override this function in your subclass to handle say too long
         *  error messages.
         *
         *  \pre The client has sent a say command and the message is longer than those prescribed by the server.
         *
         *  \pre An entire say too long error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildSayMessageTooLongError() {}
 
        /** This function is called after parsing an illegal mode error
         *  message.
         *
         *  Override this function in your subclass to handle illegal mode
         *  error messages.
         *
         *  \pre The client is an offline coach
         *
         *  \pre The client has sent a change_mode message with a playmode
         *       of PM_Null
         *
         *  \pre An entire illegal mode
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildIllegalModeError() {}
 
        /** This function is called after parsing an illegal object form error
         *  message.
         *
         *  Override this function in your subclass to handle illegal object
         *  form error messages.
         *
         *  \pre The client is an offline coach
         *
         *  \pre The client has sent a move message, where the the name of the
         *       object to be moved is not in the format expected by the
         *       server.
         *
         *  \pre An entire illegal object form
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildIllegalObjectFormError() {}
 
        /** This function is called after parsing a said too many freeform
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  said too many freeform error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a freeform CLang message and exceeded
         *       it's current quota
         *
         *  \pre An entire said too many freeform
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildSaidTooManyFreeformMessagesError() {}

        /** This function is called after parsing a cannot say freeform
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  cannot say freeform error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a freeform CLang message and the playmode
         *       is playon.
         *
         *  \pre An entire cannot say freeform
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildCannotSayFreeformWhilePlayonError() {}

        /** This function is called after parsing a said too many meta
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  said too many meta error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a meta CLang message and exceeded
         *       it's current quota
         *
         *  \pre An entire said too many meta
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildSaidTooManyMetaMessagesError() {}

        /** This function is called after parsing a said too many advice
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  said too many advice error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a advice CLang message and exceeded
         *       it's current quota
         *
         *  \pre An entire said too many advice
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildSaidTooManyAdviceMessagesError() {}

        /** This function is called after parsing a said too many info
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  said too many info error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a info CLang message and exceeded
         *       it's current quota
         *
         *  \pre An entire said too many info
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildSaidTooManyInfoMessagesError() {}

        /** This function is called after parsing a said too many define
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  said too many define error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a define CLang message and exceeded
         *       it's current quota
         *
         *  \pre An entire said too many define
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildSaidTooManyDefineMessagesError() {}

        /** This function is called after parsing a cound not parse say
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  cound not parse say error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client has sent a malformed CLang message
         *
         *  \pre An entire cound not parse say
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildCouldNotParseSayError() {}

        /** This function is called after parsing a said too many
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  said too many error messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client is using a pre version 7 protocol (this isn't
         *       actually supported with this parser)
         *
         *  \pre An entire said too many
         *       error message has been parsed.
         *
         */
        virtual
        void
        doBuildSaidTooManyMessagesError() {}

        /** This function is called after parsing an unknown
         *  error message.
         *
         *  Override this function in your subclass to handle 
         *  unknown error messages.
         *
         *  Unknown error message only occur when the parser is used with a
         *  newer version of the simulator and the parser has not been
         *  updated.  By using this function you can provide support in your
         *  teams for new error messages without having to wait for the parser
         *  to be upgraded.
         *
         *  \pre An entire unknown error
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildUnknownError( const std::string& error ) {}
        //@}


    
        //! \name Warning Parsing Functions
        //@{
        /** This function is called after parsing a \c no_team_found
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c no_team_found warning messages.
         *
         *  \pre The client is an offline coach
         *
         *  \pre The client sent a \c change_player_type 
         *       command and the team specified does not exist
         *
         *  \pre An entire \c no_team_found
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildNoTeamFoundWarning() {}

        /** This function is called after parsing a \c no_such_player
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c no_team_found warning messages.
         *
         *  \pre The client is an online or offline coach
         *
         *  \pre The client sent a \c change_player_type 
         *       command and the player specified does not exist
         *
         *  \pre An entire \c no_such_player
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildNoSuchPlayerWarning() {}

        /** This function is called after parsing a \c cannot_sub_while_playon
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c cannot_sub_while_playon warning messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client sent a \c change_player_type 
         *       command during playon
         *
         *  \pre An entire \c cannot_sub_while_playon
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildCannotSubWhilePlayOnWarning() {}

        /** This function is called after parsing a \c no_subs_left
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c no_subs_left warning messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client sent a \c change_player_type 
         *       command but the prescribed numbers of subs have
         *       already been used.
         *
         *  \pre An entire \c no_subs_left
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildNoSubsLeftWarning() {}

        /** This function is called after parsing a \c max_of_that_type_on_field
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c max_of_that_type_on_field warning messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client sent a \c change_player_type 
         *       command but there are already the prescribed maximum number of
         *       of that type already on the field.
         *
         *  \pre An entire \c max_of_that_type_on_field
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildMaxOfThatPlayerTypeOnFieldWarning() {}

        /** This function is called after parsing a \c cannot_change_goalie
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c cannot_change_goalie warning messages.
         *
         *  \pre The client is an online coach
         *
         *  \pre The client sent a \c change_player_type 
         *       but the player specified was a goalie.
         *
         *  \pre An entire \c max_of_that_type_on_field
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildCannotChangeGoalieWarning() {}
      
        /** This function is called after parsing a \c compression_unsupported
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  \c compression_unsupported warning messages.
         *
         *  \pre The client sent a \c compression command, 
         *       but the server was compiled without compression
         *       support.
         *
         *  \pre An entire \c compression command
         *       warning message has been parsed.
         *
         */
        virtual
        void
        doBuildCompressionWarning() {}

        /** This function is called after parsing an unknown
         *  warning message.
         *
         *  Override this function in your subclass to handle 
         *  unknown warning messages.
         *
         *  Unknown warning message only occur when the parser is used with a
         *  newer version of the simulator and the parser has not been
         *  updated.  By using this function you can provide support in your
         *  teams for new warning messages without having to wait for the parser
         *  to be upgraded.
         *
         *  \pre An entire unknown warning
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildUnknownWarning( const std::string& warning ) {}
        //@}


    
        //! \name OK Parsing Functions
        //@{
        /** This function is called after parsing an \c clang \c ver
         *  OK message.
         *
         *  Override this function in your subclass to handle 
         *  \c clang \v ver OK messages.
         *
         *  \pre The client is a field player
         *
         *  \pre The client sent a \c clang \c ver command
         *
         *  \pre An entire \c clang \c ver OK
         *       message has been parsed.
         *
         */       
        virtual
        void
        doBuildClangVerOK( int min, int max ) {}
          
        /** This function is called after parsing an \c eye OK
         *  message.
         *
         *  Override this function in your subclass to handle 
         *  \c eye OK messages.
         *
         *  \pre The client is an online or offline coach
         *
         *  \pre The client sent a \c eye command
         *
         *  \pre An entire \c eye OK 
         *       message has been parsed.
         *
         */       
        virtual
        void
        doBuildEyeOK( bool on ) {}

        /** This function is called after parsing an \c say OK
         *  message.
         *
         *  Override this function in your subclass to handle 
         *  \c say OK messages.
         *
         *  \pre The client is an online or offline coach
         *
         *  \pre The client sent a \c say command
         *
         *  \pre An entire \c say OK
         *       message has been parsed.
         *
         */       
        virtual
        void
        doBuildSayOK( ) {}

        /** This function is called after parsing an \c change_player_type
         *  OK message
         *
         *  Override this function in your subclass to handle 
         *  \c change_player_type OK messages.
         *
         *  \pre The client is an online or offline coach
         *
         *  \pre The client sent a \c change_player_type command
         *
         *  \pre An entire \c change_player_type OK
         *       message has been parsed.
         *
         */       
        virtual
        void
        doBuildChangePlayerTypeOK( int unum, int type ) {}

        /** This function is called after parsing an \c compression
         *  OK message
         *
         *  Override this function in your subclass to handle 
         *  \c compression OK messages.
         *
         *  \pre The client sent a \c compression command
         *
         *  \pre An entire \c compression OK
         *       message has been parsed.
         *
         */       
        virtual
        void
        doBuildCompressionOK( int level ) {}

        /** This function is called after parsing an unknown
         *  OK message.
         *
         *  Override this function in your subclass to handle 
         *  unknown OK messages.
         *
         *  Unknown OK message only occur when the parser is used with a
         *  newer version of the simulator and the parser has not been
         *  updated.  By using this function you can provide support in your
         *  teams for new OK messages without having to wait for the parser
         *  to be upgraded.
         *
         *  \pre An entire unknown OK
         *       message has been parsed.
         *
         */
        virtual
        void
        doBuildUnknownOK( const std::string& msg ) {} 
        //@}
 

    };
}

#endif
